program comp_cross_spectrum_3d
!! #############################################################################
!!
!! Copyright 2024 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_cross_spectrum_3d:
!
! ======================
! comp_cross_spectrum_3d
! ======================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 29/05/2024
!
! Purpose
! =======
!
! Compute Fast Fourier Transform (FFT) estimates of the power spectra and cross-power spectra between an index time series
! and a tridimensional variable readed from a NetCDF dataset. From the Power Spectral Density (PSD) and Cross-Spectral
! Density (CSD) estimates, this program also computes the amplitude and phase relationships between the signals, e.g.,
! the phases of the multi-channel cross-spectrum and the squared coherency estimates at each frequency. In output of
! comp_cross_spectrum_3d, both are given on the closed interval  [``0.``\ , ``1.``\ ], meaning that both are unitless.
!
! In general terms, the phase gives the lag (negative values) or lead (positive values) of the index time series with
! respect to the variable multi-channel time series at each frequency in radians.
! Here, the phase is given in fractions of a circle (e.g., on the closed interval [``0.``\ , ``1.``\ ] ) instead of being in radians.
! In other words, for phase values less than ``0.5`` the index time series leads the variable multi-channel time series and
! vice-versa for phase values greater than ``0.5``.
!
! The magnitude-squared-coherence can be used to identify significant frequency-domain correlation (at the Fourier frequencies)
! between the index time series and the multi-channel time series. Phase estimates in the cross-spectrum are only useful where
! significant frequency domain correlation exists and you can neglect the cross spectrum when the squared coherency is small.
!
! Spectral analysis amounts to estimate the power spectrum of a signal (e.g., a time series) from its time representation.
! The PSD estimates, which constitute the power spectrum, characterize the frequency content of the signal and decompose
! the variance of the signal into the different frequencies and allow to identify the periodicities present in the signal.
! In a similar way, cross-spectral analysis amounts to decompose the correlation between two signals into the different frequencies.
!
! comp_cross_spectrum_3d is based on a non-parametric method, which divides the time series into segments (see the **-sl=** argument),
! applies FFT on each segment, computes the PSD and CSD estimates from the FFTs and, finally, averages these estimates from each segment.
! Such non-parametric method is based on the observed time series and does not require prior knowledge about the data or a model generating
! the data. This method is known as the Welch's averaged, modified periodogram method for estimating the power and cross-power spectra of
! different time series [Welch]_ . This Welch's Overlapped Segment Averaging (WOSA) method also provides a consistent average coherence estimator,
! which prevents obtaining a squared coherency estimate that is identically ``1.`` for all frequencies.
!
! At the user option, the different time series can be detrended before computing the PSD and CSD estimates (see the **-tr=** and **-tr2=** arguments)
! in order to ensure that the analyzed time series remain stationary.
!
! In order to obtain stable PSD and CSD estimates (e.g., reduce the variance of the estimates) and minimize the loss of information, the real
! time series can also be segmented, and PSD and CSD estimates on, eventually overlapping, segments can be averaged at the user option (this is known
! as the Welch's method as already noted above). The stability of the PSD and CSD estimates is increased by this averaging process because the Welch's method
! uses PSD and CSD estimates from different segments of the time series, which represent approximately uncorrelated estimates of the true PSD and CSD
! estimates. That is, the greater the number of segments, the more stable, the resulting PSD and CSD estimates as their variance is reduced by the
! averaging.
!
! The user can also select different data windows for the computations of the PSD and CSD estimates (see the **-win=** argument) on each segment.
! The segments are multiplied by a data window (selected with the **-win=** argument) before performing the FFTs, so that the  Welch's method used in
! comp_cross_spectrum_3d computes a modified periodogram on each segment and then averaged these estimates to produce
! the final PSD and CSD estimates. Because the segments usually overlap (see the **-nooverlap** argument), data values at the beginning and end of
! each segment tapered by the data window on one segment, occur away from the ends of adjacent segments. In this way, the overlap guards against
! the loss of information caused by windowing, but allows to increase the number of segments and, finally, PSD and CSD estimates on each segment can
! still be considered as uncorrelated.
!
! Optionally, the PSD and CSD estimates may also been smoothed in the frequency
! domain by Daniell weights (e.g., a simple moving average; see the **-nsm=** argument).
!
! Additionally, the spectral and cross-spectral analysis can be done separately for different segments of equal length of
! the selected multi-channel time series if the time series are not continuous in time (see the **-p=** argument).
!
! The PSD estimates are returned in units, which are the square of the data
! if the **-nonormpsd** argument is used  (meaning that the sum of the PSD estimates is equal to the
! variance of the time series) or in PSD units otherwise (meaning that the total area under the power
! spectrum is equal to the variance of the time series).
!
! Finally, again at the user option, statistical confidence interval and equivalent number of degrees of freedom
! of the PSD estimates and critical values for testing the null hypothesis that the squared coherency
! between the time series is zero can be computed (see the **-conflim=** argument).
!
! All the final PSD and CSD estimates of the different time series and the associated statistics are stored in an output NetCDF dataset.
!
! If the NetCDF variable is unidimensional use :ref:`comp_cross_spectrum_1d` instead of comp_cross_spectrum_3d.
! If the multi-channel time series has a seasonal (or diurnal) cycle, use :ref:`comp_stl_3d` in order to estimate and
! remove the harmonic components of the time series before using comp_cross_spectrum_3d.
!
! comp_cross_spectrum_3d performs the same task as :ref:`comp_cross_spectrum2_3d`, but is more flexible (e.g., the **-tr=**
! and **-p=** arguments are not allowed in comp_cross_spectrum2_3d). However, for the analysis of very large NetCDF datasets,
! :ref:`comp_cross_spectrum2_3d` requires much less memory and will be faster if OpenMP parallelism is used.
!
! Finally, use :ref:`comp_spectvar_3d` to compute variance or PSD estimates in a selected frequency band for
! the real multi-channel time series after computing the power spectra of the real multi-channel time series.
!
! For more information on spectral and cross-spectral analysis, see the references below  [Welch]_  [Bloomfield]_ [Diggle]_.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_cross_spectrum_3d \
!              -f=input_netcdf_file \
!              -v=netcdf_variable \
!              -m=input_mesh_mask_netcdf_file           (optional) \
!              -g=grid_type                             (optional : n, t, u, v, w, f) \
!              -x=lon1,lon2                             (optional) \
!              -y=lat1,lat2                             (optional) \
!              -t=time1,time2                           (optional) \
!              -p=period_length                         (optional) \
!              -o=output_netcdf_file                    (optional) \
!              -vi=index_netcdf_variable                (optional) \
!              -ni=indice_for_2d_index_netcdf_variable  (optional) \
!              -fi=input_index_netcdf_file              (optional) \
!              -ti=itime1,itime2                        (optional) \
!              -pi=iperiodicity,istep                   (optional) \
!              -sl=segment_length                       (optional : 16,32,64,128,...) \
!              -tr=trend_removal                        (optional : 0,1,2,3) \
!              -tr2=trend_removal_on_segment            (optional : 0,1,2,3) \
!              -win=window_choice                       (optional : 1,2,3,4,5,6) \
!              -tap=tapered_percentage                  (optional : 0.0 > 1.) \
!              -nsm=Daniell_filter_length               (optional) \
!              -pro=critical_probability                (optional : 0.0 > 1.) \
!              -mi=missing_value                        (optional) \
!              -nonormpsd                               (optional) \
!              -nooverlap                               (optional) \
!              -conflim                                 (optional) \
!              -nospect                                 (optional) \
!              -double                                  (optional) \
!              -bigfile                                 (optional) \
!              -hdf5                                    (optional) \
!              -tlimited                                (optional)
!
! By default
! ----------
!
!        *-m=* 
!         an *input_mesh_mask_netcdf_file* is not used
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 2-D grid-mesh associated with
!         the input *netcdf_variable* is assumed to be regular or Gaussian
!        *-x=*
!         the whole longitude domain associated with the *netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *netcdf_variable*
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-p=*
!         the *period_length* is set to ``time2 - time1 + 1``, which means
!         that the time series is considered as continuous with only one time interval
!        *-o=*
!         *output_netcdf_file* name is set to ``cross_spectrum_``\ *netcdf_variable*\ ``.``\ *index_netcdf_variable*\ ``.nc``
!         if *index_netcdf_variable* is specified with the **-vi=** argument and to ``spectrum_``\ *netcdf_variable*\ ``.nc``
!         if the **-vi=** argument is not used
!        *-vi=*
!         a power spectrum analysis is performed                                                     
!        *-ni=*
!         if the *index_netcdf_variable* is bidimensional, the first time series is used
!        *-fi=*
!         the same as the *-f=* argument
!        *-ti=*
!         the whole time period associated with the *index_netcdf_variable*
!        *-pi=*
!         this parameter is not used, which is equivalent to use ``1,1`` for the **-pi=** argument
!        *-sl=*
!         the *segment_length* is set to ``0``, which means that the time series (or each
!         time interval if the **-p=** argument is used) is not segmented to stabilize the
!         the PSD and CSD estimates
!        *-tr=*
!         the *trend_removal* is set to ``1``, which means that the mean of the whole time series
!         is removed before the spectral computations
!        *-tr2=*
!         the *trend_removal_on_segment* is set to ``0``, which means that no trend processing is
!         done on each time segment before the spectral computations
!        *-win=*
!         the *window_choice* is set to ``1``, which means that the Bartlett window is used in the
!         spectral computations
!        *-tap=*
!         the *tapered_percentage* is set to ``0.2``, which means that if a split-cosine-bell window
!         is used for the spectral computations (e.g., if the *window_choice* is set to ``6`` with the
!         **-win=** argument), ``20``\ % of the data are tapered
!        *-nsm=*
!         the *Daniell_filter_length* is set to ``0``, which means that the PSD and CSD estimates are not
!         smoothed in the frequency domain by a moving average
!        *-pro=*
!         the *critical_probability* is set to ``0.05``, which means that if a confidence interval
!         for the power spectrum is requested with the **-conflim** argument, the lower and upper ``95``\ %
!         confidence limit factors are computed.
!         In addition, if the **-vi=** argument is used, the lower and upper ``(1-0.05)*100``\ % confidence
!         limit factors and the critical value for testing the null hypothesis that the squared coherency
!         is zero at the ``0.05*100``\ % significance level are computed
!        *-mi=*
!         the *missing_value* for the output NetCDF variables is set to ``1.e+20``
!        *-nonormpsd*
!         the PSD estimates are normalized in such a way that the total area under
!         the power spectrum is equal to the variance of the time series. If **-nonormpsd**
!         is specified, the PSD estimates are returned in units which are the square of the data
!        *-nooverlap*
!         normally, if the **-sl=** argument is used, the time segments are overlapped. However,
!         if **-nooverlap** is specified, the time segments are not overlapped
!        *-conflim*
!         normally, the equivalent number of degrees of freedom, the confidence limit factors
!         of the PSD estimates and the critical value for testing the null hypothesis that the
!         squared coherency is zero are not computed. However, if **-conflim** is specified,
!         these statistics are computed and stored in the output NetCDF dataset
!        *-nospect*
!         the PSD estimates are stored in the output NetCDF dataset
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension (e.g., the frequency axis) is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable (e.g., the multi-channel time series)
!      for which a  power or cross-power spectrum analysis must be performed and the **-f=**\ *input_netcdf_file* 
!      argument specifies that this NetCDF variable must be extracted from the NetCDF file *input_netcdf_file*.
!
!   #) The optional argument **-m=**\ *input_mesh_mask_netcdf_file* specifies the land-sea mask
!      to apply to *netcdf_variable* for transforming this tridimensional NetCDF variable as a
!      rectangular matrix of observed variables before computing the correlation analysis. By default,
!      it is assumed that each cell in the 2-D grid-mesh associated with the input tridimensional
!      NetCDF variable is a valid time series (e.g., missing values are not present).
!
!      The geographical shapes of the *netcdf_variable* (in the *input_netcdf_file*) and the mask
!      (in the *input_mesh_mask_netcdf_file*) must agree if an *input_mesh_mask_netcdf_file* is used.
!
!      Refer to :ref:`comp_clim_3d` or :ref:`comp_mask_3d` for creating a valid
!      *input_mesh_mask_netcdf_file* NetCDF file for regular or gaussian grids before
!      using comp_cross_spectrum_3d.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that
!      the input NetCDF variable is from an experiment with the NEMO model (ORCA configuration and R2, R4 or R05 resolutions).
!      This argument is also used to determine the name of the mesh_mask variable if
!      an *input_mesh_mask_netcdf_file* is used.
!
!   #) If the **-x=**\ *lon1,lon2* and **-y=**\ *lat1,lat2* arguments are missing the whole
!      geographical domain associated with the *netcdf_variable* is used.
!
!      The longitude or latitude range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_3d` for transforming geographical coordinates as indices or
!      generating an appropriate mesh-mask before using comp_cross_spectrum_3d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing the whole time period
!      associated with the *netcdf_variable* is used to estimate the power and cross-power spectrum.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ .
!
!      *time2* - *time1* + ``1`` must be an even integer if the **-sl=** argument is not used
!      and of any length if this argument is used.
!
!   #) If the **-p=** argument is specified, the PSD and CSD estimates are computed separately
!      for each time period of length *period_length* (as determined by the value of the **-p=** argument).
!
!      The whole selected time period (e.g., *time2* - *time1* + ``1`` ) must be a multiple of 
!      the *period_length*.
!      The *period_length* must also be of even length if the **-sl=** argument is not used.
!
!   #) The **-vi=**\ *index_netcdf_variable* specifies a time series for the cross-power
!      spectrum analysis.
!      If the **-vi=**\ *index_netcdf_variable* is present, the **-fi=** argument
!      must also be present and this argument specifies the NetCDF dataset
!      which contains the *index_netcdf_variable*. However, if the NetCDF dataset which
!      contains the *index_netcdf_variable* is the same as the NetCDF dataset specified
!      by the **-f=** argument, it is not necessary to specify the **-fi=** argument.
!
!      If the **-vi=**\ *index_netcdf_variable* is absent, a power spectrum analysis of
!      the NetCDF variable *netcdf_variable* specified by the **-v=** argument is performed.
!
!   #) The **-ni=** argument specifies the indice (e.g., an integer) for selecting the index time series if
!      the *index_netcdf_variable*  specified with the **-vi=** argument is a 2D NetCDF variable. By default,
!      the first time series is used, which is equivalent to set *indice_for_2d_netcdf_variable* to ``1``\ .
!
!   #) If the **-ti=**\ *itime1,itime2* argument is missing, data in the whole time period
!      associated with the *index_netcdf_variable* is taken into account.
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ .
!
!   #) The **-pi=** argument gives the periodicity and selects the time step for the
!      *index_netcdf_variable*. For example, to compute a cross-spectral analysis with the January
!      monthly time series extracted from the *index_netcdf_variable* which is assumed
!      to be sampled every month, **-pi=**\ ``12,1`` should be specified, with yearly data
!      **-pi=**\ ``1,1`` may be used, etc.
!
!   #) The selected time periods for the *netcdf_variable* and *index_netcdf_variable*
!      must agree. This means that the following equality must be verified
!
!      ``time2 - time1 + 1  = ceiling((itime2 - itime1 - istep + 2)/iperiodicity)``,
!
!      otherwise, an error message will be issued and the program will stop.
!
!   #) The **-sl=**\ *segment_length* argument specifies an integer used to segment the time series.
!      The *segment_length* must be a positive even integer less or equal to the length of the time series
!      (e.g., *time2* - *time1* + ``1`` ) or the *period_length* if the **-p=** argument is used.
!      Spectral computations are at ( *segment_length*\ /\ ``2``\  )  + ``1`` Fourier frequencies and suggested
!      values for *segment_length* are ``16``, ``32``, ``64``, ``128``, ...  (e.g., the spectral estimates are taken
!      at frequencies ``(i-1)``\  / *segment_length* for i=1,2, ... , (*segment_length*\ /\ ``2``\ ) + ``1``.
!      By default, *segment_length* is set to ``0``, e.g., the time series is not segmented and *segment_length* is
!      equal to *time2* - *time1* + ``1`` or *period_length* if the **-p=** argument is used.
!
!   #) The **-tr=** argument specifies pre-filtering processing of the real time series before
!      estimating the PSD and CSD estimates of the time series. If:
!
!      * **-tr=**\ ``+1``, the mean of the time series is removed before computing the PSD and CSD estimates
!
!      * **-tr=**\ ``+2``, the drift from the time series is removed before computing the PSD and CSD estimates.
!        The drift for each time series (say ``tseries``) is estimated using the formula : ``drift = ( tseries(ntime) - tseries(1) )/( ntime - 1 )``
!
!      * **-tr=**\ ``+3``, the least-squares line from the time series is removed before computing the PSD and CSD estimates.
!
!      For other values of the **-tr=** argument, nothing is done before estimating the PSD and CSD estimates.
!
!      If the **-p=** argument is present, the pre-filtering processing is applied to 
!      each time interval of length *period_length*, separately.
!
!      The **-tr=** argument must be an integer and the default value for the **-tr=** argument is ``1``,
!      e.g., the mean of the time series is removed before the spectrum computations.
!
!   #) The **-tr2=** argument specifies pre-filtering processing of the real time series before
!      estimating the PSD and CSD estimates on each time segment. If:
!
!      * **-tr2=**\ ``+1``, the mean of the time segment is removed before computing the PSD and CSD estimates
!        on the segment
!
!      * **-tr2=**\ ``+2``, the drift from the time segment is removed before computing the PSD and CSD estimates
!        on the segment
!
!      * **-tr2=**\ ``+3``, the least-squares line from the time segment is removed before computing the
!        PSD and CSD estimates on the segment.
!
!      For other values of the **-tr2=** argument, nothing is done before estimating the PSD and CSD estimates on each
!      time segment.
!
!      The **-tr2=** argument must be an integer and the default value for the **-tr2=** argument is ``0``,
!      e.g., nothing is done before the spectral computations on each time segment.
!
!   #) The **-win=**\ *window_choice* argument specifies the form of the data window used in the computations
!      of the modified periodogram on each segment. If:
!
!      * **-win=**\ ``1`` The Bartlett window is used
!
!      * **-win=**\ ``2`` The square window is used
!
!      * **-win=**\ ``3`` The Welch window is used
!
!      * **-win=**\ ``4`` The Hann window is used
!
!      * **-win=**\ ``5`` The Hamming window is used
!
!      * **-win=**\ ``6`` A split-cosine-bell window is used.
!
!      The *window_choice* must be an integer between ``1`` and ``6``.
!      The default value for the *window_choice* is ``1``, e.g., the Bartlett window is used for the spectral
!      computations on each segment.
!
!   #) The **-tap=**\ *tapered_percentage* argument specifies the total percentage of the data to be tapered if
!      the *window_choice* is set to ``6`` with the **-win=** argument.
!      The *tapered_percentage* is a real number which must be greater than ``0.`` and less or equal to ``1.``.
!
!      The default value for the *tapered_percentage* is ``0.2`` (e.g., ``20``\ % of the data are tapered, ``10``\ % at
!      the start of the series and ``10``\ % at the end).
!
!   #) The **-nsm=**\ *Daniell_filter_length* argument specified that the PSD and CSD estimates must be computed by smoothing
!      the periodogram with Daniell weights (e.g., a simple moving average).
!      On entry, the **-nsm=** argument gives the length of the Daniell filter to be applied.
!      The *Daniell_filter_length* must be an odd integer, greater than ``2`` and less or equal to ( *segment_length*\ /\ ``2``\  ) + ``1``
!      if the  **-sl=** argument is specified.
!
!      By default, the PSD and CSD estimates are not smoothed in the frequency domain.
!
!   #) If the **-nooverlap** argument is specified and the **-sl=** argument is also specified,
!      comp_cross_spectrum_3d segments the time series without any overlapping (in each time period 
!      if the **-p=** argument is used).
!
!      By default, comp_cross_spectrum_3d overlaps the segments by one half of their length (which is
!      equal to the value of the **-sl=** argument).
!
!      In both cases, each segment will be FFT'd, and the resulting (modified if the **-win=** argument is used)
!      periodograms and cross-periodograms will be averaged together to obtain PSD and CSD estimates at the
!      ( *segment_length*\ /\ ``2``\  ) + ``1`` frequencies.
!
!      If the **-sl=** argument is not specified, the **-nooverlap** argument has no effect.           
!
!   #) If the **-nonormpsd** argument is specified, the sum of the PSD estimates is equal to
!      the variance of the time series. By default, the PSD estimates are normalized in such
!      a way that the total area under the power spectrum is equal to the variance of the time
!      series.
!
!   #) If the **-conflim** argument is specified, the equivalent number of degrees of freedom,
!      the lower and upper (\  ``1``\  -\  *critical_probability*\  ) * ``100``\  % confidence limit factors of the PSD
!      estimates and the critical value for testing the null hypothesis that the squared coherency is zero at the
!      *critical_probability*\ * ``100``\  % significance level are computed.
!
!      Multiply the PSD estimates by the lower and upper confidence limit factors to get the lower and upper
!      limits of a (\  ``1``\  -\  *critical_probability*\  ) * ``100``\  % confidence
!      interval for the PSD estimates.
!
!      Squared coherency less than the computed critical value should be regarded as not
!      significantly different from zero at the *critical_probability*\ * ``100``\  % significance level.
!
!      By default, these statistics are not computed.
!
!   #) The **-pro=**\ *critical_probability* argument specifies the critical probability which
!      is used to determine the lower and upper confidence limit factors and the critical value
!      for testing the null hypothesis that the squared coherency is zero.
!
!      *critical_probability* must be greater than ``0.`` and less than ``1.``. The default value is ``0.05``.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator for
!      the output NetCDF variables in the *output_netcdf_file*. 
!
!      If the **-mi=** argument is not specified, the *missing_value* is set
!      to ``1.e+20``.
!       
!   #) If the **-nospect** argument is specified, the PSD estimates of the time series are not stored
!      in the output netCDF file. By default, the PSD estimates are stored.
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g., ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g., ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) It is assumed that the different real time series have no missing values.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence
!      of a parameter which will be used for the computations. Moreover, the
!      number of specified parameters must not be greater than the total number
!      of allowed parameters.
!
!   #) For more details on power and cross-power spectrum analysis and examples of use in the climate literature, see
!
!      * "The use of Fast Fourier Transform for the estimation of power spectra: A method based on time averaging over short, modified periodograms",
!        by P.D. Welch, IEEE trans. on audio and electroacoustics, 15:2, 70-73, 1967. DOI: https://doi.org/10.1109/TAU.1967.1161901
!
!      * "Fourier analysis of time series- An introduction",
!        by P. Bloomfield, John Wiley and Sons, New York, Chapter 6, 1976. http://as.wiley.com/WileyCDA/WileyTitle/productCd-0471889482.html
!
!      * "Time series: a biostatistical introduction",
!        by P.J. Diggle, Clarendon Press, Oxford, Chap. 4, 1990. http://www.oupcanada.com/catalog/9780198522263.html
!
!      * "Impact of intra-daily SST variability on ENSO characteristics in a coupled model"
!        by S. Masson et al., 2012, Climate Dynamics, Vol. 39:681-707, doi: `10.1007/s00382-011-1247-2`_
!
!      * "The role of the intra-daily SST variability in the Indian monsoon variability in a global coupled model"
!        by P. Terray et al., 2012, Climate Dynamics, Vol. 39:729-754, doi: `10.1007/s00382-011-1240-9`_
!
!      * "Southern Hemisphere extra-tropical forcing: A new paradigm for El Nino-Southern Oscillation"
!        by P. Terray, 2011, Climate Dynamics, Vol. 36:2171-2199, doi: `10.1007/s00382-010-0825-z`_
!
! .. _`10.1007/s00382-011-1247-2` : http://dx.doi.org/10.1007/s00382-011-1247-2
! .. _`10.1007/s00382-011-1240-9` : http://dx.doi.org/10.1007/s00382-011-1240-9
! .. _`10.1007/s00382-010-0825-z` : http://dx.doi.org/10.1007/s00382-010-0825-z
!
! Outputs
! -------
!
!   comp_cross_spectrum_3d creates an output NetCDF file that contains the PSD and CSD estimates of the
!   time series associated with the input NetCDF variables. The output NetCDF dataset contains the
!   following NetCDF variables (in the description below, nlat and nlon are the lengths of the spatial dimensions
!   of the input NetCDF variable *netcdf_variable*, nfreq is the number of frequencies where
!   the PSD and CSD estimates are computed (as determined by the **-t=**, **-p=** and **-sl=** arguments) and
!   nperiod is ``1`` if the **-p=** argument is not used or equal to ``ntime``\  / *period_length* if
!   it is used, where ``ntime``\  = *time2* - *time1* + ``1`` ) :
!
!   #) frequency\ ``(nfreq)`` : the frequencies at which the PSD estimates are computed.
!
!   #) periodicity\ ``(nfreq)`` : the corresponding periodicities in number of time observations.
!
!   #) *netcdf_variable*\ _spectrum\ ``(nperiod,nfreq,nlat,nlon)`` : the corresponding PSD estimates
!      for each time interval of the multi-channel time series associated with the input NetCDF variable.
!
!      This variable is stored only if the **-nospect** argument is absent when calling
!      comp_cross_spectrum_3d.
!
!   #) *index_netcdf_variable*\ _spectrum\ ``(nperiod,nfreq)`` : the corresponding PSD estimates
!      for each time interval of the time series associated with the input index NetCDF variable.
!
!      This variable is stored only if the **-vi=** argument has been specified when calling
!      comp_cross_spectrum_3d.
!
!   #) *index_netcdf_variable*\ _\ *netcdf_variable*\ _squaredcoherency\ ``(nperiod,nfreq,nlat,nlon)`` : the corresponding squared coherency
!      estimates for each time interval of the multi-channel time series associated with the input NetCDF variable and the time series
!      associated with the input index NetCDF variable. The square coherency is given on the closed interval [``0``\ , ``1``\ ].
!
!      This variable is stored only if the **-vi=** argument has been specified when calling
!      comp_cross_spectrum_3d.
!
!   #) *index_netcdf_variable*\ _\ *netcdf_variable*\ _phase\ ``(nperiod,nfreq,nlat,nlon)`` : the corresponding phase
!      estimates of the cross-spectrum for each time interval of the multi-channel time series associated with the input NetCDF
!      variable and the time series associated with the input index NetCDF variable. In general terms, the phase gives the
!      lag (negative values) or lead (positive values) of the *index_netcdf_variable* time series with respect to the
!      *netcdf_variable* multi-channel time series at each frequency in radians.
!      
!      Here, the phase is given in fractions of a circle (e.g., on the closed interval [``0.``\ , ``1.``\ ] ) instead of being in radians.
!      In other words, for phase values less than ``0.5`` the *index_netcdf_variable* leads the *netcdf_variable* and vice-versa for
!      phase values greater than ``0.5``.
!
!      This variable is stored only if the **-vi=** argument has been specified when calling
!      comp_cross_spectrum_3d.
!
!   #) edof\ ``(nfreq)`` : the equivalent number of degrees of freedom of the PSD and CSD estimates
!      of the time series associated with the input NetCDF variable.
!
!      This variable is stored only if the **-conflim** argument has been specified when calling
!      comp_cross_spectrum_3d.
!
!   #) lower_factor\ ``(nfreq)`` : the lower confidence limit factors of the PSD estimates.
!      Multiply the PSD estimates by the corresponding lower confidence limit factors to get the lower
!      limits of the (\  ``1``\  -\  *critical_probability*\  ) * ``100``\  % confidence
!      intervals for the PSD estimates.
!
!      This variable is stored only if the **-conflim** argument has been specified when calling
!      comp_cross_spectrum_3d.
!
!   #) upper_factor\ ``(nfreq)`` : the upper confidence limit factors of the PSD estimates.
!      Multiply the PSD estimates by the corresponding upper confidence limit factors to get the upper
!      limits of the (\  ``1``\  -\  *critical_probability*\  ) * ``100``\  % confidence
!      intervals for the PSD estimates.
!
!      This variable is stored only if the **-conflim** argument has been specified when calling
!      comp_cross_spectrum_3d.
!
!   #) coher_test\ ``(nfreq)`` : the critical values for the statistical test of the squared coherency estimates.
!      Squared coherency less than these computed critical values for a given frequency should be regarded as not
!      significantly different from zero at the *critical_probability*\ * ``100``\  % significance level.
!
!      This variable is stored only if the **-conflim** and  **-vi=** arguments have been specified when calling
!      comp_cross_spectrum_3d.
!
!   All the PSD and CSD estimates involving the input NetCDF variable *netcdf_variable*  are packed in tri- or fourdimensional
!   variables whose first and second dimensions are exactly the same as those associated with this input NetCDF variable
!   even if you restrict the geographical domain with the **-x=** and **-y=** arguments. However,
!   outside the selected domain, the output NetCDF variables are filled with missing values.
!
! Examples
! --------
!
!   #)   For computing the power and cross-power spectra of a real multi-channel monthly time series from a tridimensional NetCDF variable called ``sst``
!        extracted from the file ``sst.monthly.nc`` with a monthly Nino34 SST index in a NetCDF variable called ``sst_nino34`` extracted from
!        the file ``sst.monthly.nino34.nc`` and store the results in a NetCDF file named ``cross_spectrum_sst_nino34_sst.nc``, use the following command:
!
!        ::
!
!          $ comp_cross_spectrum_3d \
!            -f=sst.monthly.nc \
!            -v=sst \
!            -fi=sst.monthly.nino34.nc \
!            -vi=sst_nino34 \
!            -sl=128 \
!            -o=cross_spectrum_sst_nino34_sst.nc \
!            -nonormpsd \
!            -nooverlap
!
!-
! ================================================================================================
!
#ifndef _OPENMP
#undef _PARALLEL_READ
#endif
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!   
#ifdef _PARALLEL_READ
    use netcdf,           only : nf90_share
    use Statpack,         only : omp_limit
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!   
    use netcdf,   only : nf90_max_name, nf90_max_var_dims, nf90_clobber, nf90_real4, nf90_real8,      &
                         nf90_int, nf90_sizehint_default, nf90_open,                                  &
                         nf90_nowrite, nf90_noerr, nf90_get_att, nf90_get_var, nf90_strerror,         &
                         nf90_create, nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_unlimited,       &
                         nf90_def_var, nf90_inq_varid, nf90_inq_attname, nf90_put_att, nf90_copy_att, &
                         nf90_enddef, nf90_global, nf90_put_var, nf90_close
!   
    use Statpack, only : i1b, i4b, stnd, extd, lgl, true, false, zero, one, kfix, kexp, kint, tolower,       & 
                         case_change, string_to_val, val_to_string, arth, tseries_error58,                   &
                         allocate_error, merror, power_spctrm2, cross_spctrm2, init_fft, end_fft,            &
                         data_window, estim_dof2, comp_conflim 
!   
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing,      &
                                 get_listarg, get_variable_dims, get_variable_dimcoords
!   
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_cross_spectrum_3d : '
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncidi, ncid_spect, ncid_mask,                                 &
                     varid, maskid, varidi, spectid, varispectid, coherid, phaseid, periodid,          &
                     edofid, conlwrid, conuprid, testcoherid,                                          &
                     i, j, j1, j2, k, l, l1, l2, n, nt, nidim, nti, iindex,                            &
                     nlon, nlat, lon1, lon2, lon3, lat1, lat2, lat3, nvar, var_natts,                  &
                     time1, time2, time3, itime1, itime2, itime3, nf, nfm1, sl, nseg,                  &
                     trend, trend2, win, output_type, output_mode, period, nperiod, offset_rec, rec,   &
                     iperiod, istep, niperiod, nsmooth
    integer(i4b), dimension(3)                 :: startlec, countlec
    integer(i4b), dimension(4)                 :: coor_id2, startwri, countwri
    integer(i4b), dimension(5)                 :: id_var
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),     allocatable   :: veci, freq, conlwr, conupr, edof, testcoher, wk
    real(stnd), dimension(:,:),   allocatable   :: temp, mat, psveci
    real(stnd), dimension(:,:,:), allocatable   :: psmat, phase, coher
    real(stnd)                                  :: missing, tmp, scalfac, offset, taperp, probtest
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                               :: use_tlimited, use_x, use_y, use_mask, use_period, normpsd,      &
                                                  overlap, smooth, use_miss, comp_cross, use_scalfac, use_offset, &
                                                  save_spect, conflim, create_hdf5
    logical(lgl), dimension(nf90_max_var_dims) :: use_coor_variable
    logical(lgl), dimension(:,:), allocatable  :: mask
!
    character(len=nchar)                       :: name, name2, history
    character(len=nf90_max_name)               :: fic_name, var_name, mask_name, fici_name, vari_name,    &
                                                  spect_name, att_name
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                                          :: fmt
    character                                                  :: chr, grid_type
!
#ifdef _PARALLEL_READ
    integer(i4b)  :: ncid2, varid2
!
    logical  :: test_par
!
!   INTEGER OpenMP FUNCTIONS
!
    integer :: omp_get_num_procs, omp_get_max_threads
!    
!   OpenMP PROCEDURES
!
    external omp_get_num_procs, omp_get_max_threads
!
#endif
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
! 
    if ( narg==0 ) then
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute a Fast Fourier Transform (FFT) estimate of the cross-power'
        write (prtunit,*) ' spectra between an index time series and a tridimensional variable readed'
        write (prtunit,*) ' from a NetCDF dataset'
        write (prtunit,*) ' From the Cross-Spectral Density (CSD) estimates, this program computes' 
        write (prtunit,*) ' the amplitude and phase relationship between the signals.'
        write (prtunit,*) ' The Power Spectra Density (PSD) estimates of the time series are also stored'
        write (prtunit,*) ' in the output NetCDF file.'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_cross_spectrum_3d  '
        write (prtunit,*) '        -f=input_netcdf_file'
        write (prtunit,*) '        -v=netcdf_variable'
        write (prtunit,*) '        -m=input_mesh_mask_netcdf_file           (optional)'
        write (prtunit,*) '        -g=grid_type                             (optional : n, t, u, v, w, f)'
        write (prtunit,*) '        -x=lon1,lon2                             (optional)'
        write (prtunit,*) '        -y=lat1,lat2                             (optional)'
        write (prtunit,*) '        -t=time1,time2                           (optional)'
        write (prtunit,*) '        -p=period_length                         (optional)'
        write (prtunit,*) '        -o=output_netcdf_file                    (optional)'
        write (prtunit,*) '        -vi=index_netcdf_variable                (optional)'
        write (prtunit,*) '        -ni=indice_for_2d_index_netcdf_variable  (optional)'
        write (prtunit,*) '        -fi=input_index_netcdf_file              (optional)'
        write (prtunit,*) '        -ti=itime1,itime2                        (optional)'
        write (prtunit,*) '        -pi=iperiodicity,istep                   (optional)'
        write (prtunit,*) '        -sl=segment_length                       (optional : 16,32,64,128,...)'
        write (prtunit,*) '        -tr=trend_removal                        (optional : 0,1,2,3)'
        write (prtunit,*) '        -tr2=trend_removal_on_segment            (optional : 0,1,2,3)'
        write (prtunit,*) '        -win=window_choice                       (optional : 1,2,3,4,5,6)'
        write (prtunit,*) '        -tap=tapered_percentage                  (optional : 0. > 1.)'
        write (prtunit,*) '        -nsm=Daniell_filter_length               (optional)'
        write (prtunit,*) '        -pro=critical_probability                (optional : 0. > 1.)'
        write (prtunit,*) '        -mi=missing_value                        (optional)'
        write (prtunit,*) '        -nonormpsd                               (optional)'
        write (prtunit,*) '        -nooverlap                               (optional)'
        write (prtunit,*) '        -conflim                                 (optional)'
        write (prtunit,*) '        -nospect                                 (optional)'
        write (prtunit,*) '        -double                                  (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '        -bigfile                                 (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                                    (optional)'
#endif
        write (prtunit,*) '        -tlimited                                (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-m= an input_mesh_mask_netcdf_file is not used'
        write (prtunit,*) '-g=n : a regular or gaussian grid is assumed'
        write (prtunit,*) '-x= the whole longitude domain associated with the netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the netcdf_variable'
        write (prtunit,*) '-t= the whole time period associated with the netcdf_variable'
        write (prtunit,*) '-p=time2-time1+1 : the period_length argument is not used'
        write (prtunit,*) '-o=cross_spectrum_netcdf_variable.index_netcdf_variable.nc if -vi= argument is present'
        write (prtunit,*) '   spectrum_netcdf_variable.nc                             if -vi= argument is absent'
        write (prtunit,*) '-vi= a simple power spectrum analysis is performed'
        write (prtunit,*) '-ni=1 : if the index_netcdf_variable is bidimensional, the first time series is used'
        write (prtunit,*) '-fi= the same as the -f= argument'
        write (prtunit,*) '-ti= the whole time period  associated with the indice_netcdf_variable'
        write (prtunit,*) '-pi=1,1          : by default, this parameter is not used'
        write (prtunit,*) '-sl=0            : by default, the time series is not segmented'
        write (prtunit,*) '-tr=1            : by default, the mean of the time series is removed'
        write (prtunit,*) '-tr2=0           : by default, no preprocessing is done on the segments'
        write (prtunit,*) '-win=1           : by default, the Bartlett window is used'
        write (prtunit,*) '-tap=0.2         : by default, if a split-cosine-bell window is used, 20% of the data are tapered'
        write (prtunit,*) '-nsm=0           : by default, the PSD and CSD estimates are not smoothed in the frequency domain'
        write (prtunit,*) '-pro=0.05        : by default, the lower and upper 95% confidence limit factors and'
        write (prtunit,*) '                   the 5% significance level critical value for testing the squared coherency'
        write (prtunit,*) '                   are computed'
        write (prtunit,*) '-mi=1.e+20       : by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-nonormpsd       : by default, the PSD estimates are normalized'
        write (prtunit,*) '-nooverlap       : by default, the segments are overlapped'
        write (prtunit,*) '-conflim         : by default, the number of degrees of freedom, the confidence limit factors'
        write (prtunit,*) '                   and the critical value for testing the squared coherency are not computed'
        write (prtunit,*) '-nospect         : by default, the PSD estimates are stored in the output NetCDF file'
        write (prtunit,*) '-double          : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile         : by default, a NetCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5            : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited        : by default, the time (e.g., frequency) dimension is defined as unlimited'
#ifdef _OPENMP
        write (prtunit,*)
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<2 .or. narg>30 ) then
        call merror( name_proc//"Wrong number of line arguments" )
    end if
!
    fic_name    = " "
    var_name    = " "
    mask_name   = " "
    fici_name   = " "
    vari_name   = " "
    spect_name  = " "
!
    history = "comp_cross_spectrum_3d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
!
    sl            = 0
    trend         = 1
    trend2        = 0
    win           = 1
    nsmooth       = 0
    taperp        = 0.2_stnd
    probtest      = 0.05_stnd
!
    missing = def_missing
!
    grid_type     = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    time1     = 1
    time2     = huge( time2 )
!
    itime1    = 1
    itime2    = huge( itime2 )
    iperiod   = 1
    istep     = 1
    niperiod  = 0
    iindex    = 1
!
    period    = huge( period )
    nperiod   = 1
!
    normpsd      = true
    overlap      = true
    smooth       = false
    conflim      = false
    comp_cross   = false
    save_spect   = true
    use_period   = false
    use_mask     = false
    use_x        = false
    use_y        = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
!
            case ('-t=')
!
!               GET THE TIME DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong time definition",                &
                                  time1, time2 )
!
            case ('-p=')
!
!               GET THE PERIOD LENGTH.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode/=kint ) &
                call merror( name_proc//"The period length is an integer value" )
!
                read( name(4:), fmt ) period
!
                if ( period<=0 ) &
                call merror( name_proc//"The -p= parameter is a positive integer" )
!
                use_period = true
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
                use_mask = true
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                spect_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-fi=')
!
!                       GET THE NAME OF THE NETCDF DATASET FOR THE INDEX VARIABLE.
!
                        fici_name = name(5:)
!
                    case ('-vi=')
!
!                       GET THE NAME OF THE INDEX NETCDF VARIABLE IN THE DATASET.
!
                        vari_name = name(5:)
!
                        comp_cross = true
!
                    case ('-ti=')
!
!                       GET THE TIME DEFINITION FOR THE INDEX VARIABLE.
!
                        call get_listarg( name_proc, name(5:),                            &
                                          "itime1 and itime2 are integer values",         &
                                          "Wrong period definition for indice variable",  &
                                          itime1, itime2 )
!
                    case ('-pi=')
!
!                       GET THE PERIODICITY AND THE TIME STEP FOR THE INDEX VARIABLE.
!
                        call get_listarg( name_proc, name(5:),                                &
                                          "iperiod and istep are integer values",             &
                                          "Wrong periodicity definition for indice variable", &
                                          iperiod, istep )
!
                        if ( iperiod<=0 .or. istep<=0 .or. istep>iperiod )   &
                call merror( name_proc//"The -pi= parameters are positive integers with istep<=iperiod" )
!
                    case ('-ni=')
!
!                       GET THE INDICE FOR A 2D INDEX VARIABLE.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) iindex
                        end if
                        if ( kcode/=kint .or. iindex<=0  )                           &
                        call merror( name_proc//"The indice for a 2D index variable is a positive integer value" )
!
                    case ('-sl=')
!
!                       GET THE LENGTH OF THE SEGMENTS.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) sl
                        end if
!
                        if ( kcode/=kint .or. sl<0  )                           &
                        call merror( name_proc//"The -sl= argument is a positive integer value" )
!
!                       CHECK IF sl IS A POSITIVE EVEN INTEGER.
!
                        if ( sl/=2*(sl/2) )     &
                        call merror( name_proc//'The -sl= argument is an even integer !' )
!
                    case ('-tr=')
!
!                       DETERMINE PRE-FILTERING COMPUTATIONS.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) trend
                        else
                            call merror( name_proc//"The -tr= argument is an integer value" )
                        end if
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                        call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:5) )
!
                            case ('-tr2=')
!
!                               DETERMINE PRE-FILTERING COMPUTATIONS ON EACH SEGMENT.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) trend2
                                else
                                    call merror( name_proc//"The -tr2= argument is an integer value" )
                                end if
!
                            case ('-win=')
!
!                               DETERMINE THE FORM OF THE DATA WINDOW.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) win
                                end if
                                if ( kcode/=kint .or.      &
                                     win<1       .or.      &
                                     win>6 )               &
      call merror( name_proc//"The -win= argument must be an integer number and verify 1<=win<=6 " )
!
                            case ('-nsm=')
!
!                               DETERMINE THE LENGTH OF THE DANIELL FILTER TO BE APPLIED.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) nsmooth
                                end if
!
                                if ( kcode/=kint .or.             &
                                     nsmooth<3   .or.             &
                                     nsmooth==(nsmooth/2)*2   )   &
      call merror( name_proc//"The -nsm= argument must be an odd integer and verify 3<=nsm " )
!
                                smooth = true
!
                            case ('-tap=')
!
!                               DETERMINE THE TOTAL PERCENTAGE OF THE DATA TO BE TAPERED IF -WIN=6 .
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kfix .or. kcode==kexp ) then
                                    read( name(6:), fmt ) taperp
                                end if
                                if ( (kcode/=kfix .and. kcode/=kexp) .or.     &
                                     taperp<=zero                    .or.     &
                                     taperp>one )                             &
      call merror( name_proc//"The -tap= argument must be a real number and verify 0.<tap<=1. " )
!
                            case ('-pro=')
!
!                               DETERMINE THE CRITICAL PROBABILITY.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kfix .or. kcode==kexp ) then
                                    read( name(6:), fmt ) probtest
                                end if
                                if ( (kcode/=kfix .and. kcode/=kexp) .or.     &
                                     probtest<=zero                  .or.     &
                                     probtest>=one )                          &
      call merror( name_proc//"The -pro= argument must be a real number and verify 0.<pro<1. " )
!
                            case default
!
                                select case ( name(1:) )
!
                                    case ('-nonormpsd')
!
!                                       DO NOT NORMALIZE THE PSD ESTIMATES.
!
                                        normpsd = false
!
                                    case ('-nooverlap')
!
!                                       DO NOT OVERLAP THE SEGMENTS.
!
                                        overlap = false
!
                                    case ('-conflim')
!
!                                       COMPUTE THE CONFIDENCE LIMIT FACTORS AND THE
!                                       CRITICAL VALUE FOR TESTING THE COHERENCY.
!
                                        conflim = true
!
                                    case ('-nospect')
!
!                                       DO NOT SAVE THE PSD ESTIMATES.
!
                                        save_spect = false
!
#ifdef _USE_NETCDF36
                                    case ('-bigfile')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                                    case ('-hdf5')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_hdf5
                                        create_hdf5 = true
#endif
!
                                    case ('-tlimited')
!
!                                       DEFINE THE TIME DIMENSION AS LIMITED.
!
                                        use_tlimited = true
!
                                    case ('-double')
!
!                                       STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                        output_type = nf90_real8
!
                                    case default
!
!                                       WRONG ARGUMENT.
!
                                        call merror( name_proc//"Wrong argument" )
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( comp_cross ) then
!
        if ( fici_name==" " ) fici_name = fic_name
!
        if ( spect_name==" " )                                                    &
        spect_name="cross_spectrum_"//trim(var_name)//"."//trim(vari_name)//".nc"
!
    else
!
        if ( .not.save_spect )              &
        call merror( name_proc//"Nothing will be done if -nospect is specified and -vi= argument is absent" )
        
!
        if ( spect_name==" " ) spect_name="spectrum_"//trim(var_name)//".nc"
!
    end if
!
!   OPEN THE NETCDF DATASETS.
!
    i = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=i )
    if ( status/=nf90_noerr )                                                &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( use_mask ) then
        status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( comp_cross ) then
!
        if ( fici_name/=fic_name ) then
            status = nf90_open( trim(fici_name), nf90_nowrite, ncidi )
            if ( status/=nf90_noerr )                                        &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
        else
            ncidi = ncid
        end if
!
    end if
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    call get_variable_dimcoords(  name_proc, trim(var_name), ncid,                          &
                                  varid,  dimids_out=dimids, dim_names_out=dim_names,       &
                                  dimlens_out=dimlens, coor_id_out=coor_id,                 &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type,       &
                                  use_coor_variable_out=use_coor_variable, natts=var_natts  )
!
    if ( count(dimlens/=0)/=3 )           &
    call merror( name_proc//"Input NETCDF variable is not tridimensional" )
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nt     = dimlens(3)
!
    if ( use_mask ) then
!
        chr = merge( 't', grid_type, grid_type=='w' )
!
!       GET THE MASK ID FROM ITS NAME.
!       VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
        call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
        call merror( name_proc//"Shape disagreement with the mask" )
!
        if ( .not.use_x ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
            if ( status==nf90_noerr ) lon1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
            if ( status==nf90_noerr ) lon2 = i
!
        end if
!
        if ( .not.use_y ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
            if ( status==nf90_noerr ) lat1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
            if ( status==nf90_noerr ) lat2 = i
!
        end if
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    if ( lon1<0 ) then
        lon2 = max( lon2, 1 )
        lon3 = nlon + lon1 - lon2 + 1
    else
        lon1 = max( lon1, 1 )
        lon2 = min( lon2, nlon )
        lon3 = lon2 - lon1 + 1
    end if
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    if ( lon1<0 ) then
       lon3 = nlon
       j1   = 1
       j2   = nlon
    else
       j1 = lon1
       j2 = lon2
    end if
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE TIME AND PERIOD DEFINITIONS.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
!
    if ( time3<4  )       &
    call merror( name_proc//"Wrong time definition or not enough time observations" )
!
    if ( use_period ) then
!
        if ( period<4  )     &
        call merror( name_proc//"Not enough time observations in the periods" )
!
        nperiod = time3/period
        if ( nperiod*period/=time3 )    &
        call merror( name_proc//"The selected time interval is not a multiple of the period length" )
!
    else
!
        period = time3
!
    end if
!
!   CHECK SEGMENT LENGTH.
!
    if ( sl==0 ) then
!
        i = 2*(period/2)
        if ( period/=i )       &
    call merror( name_proc//"The period or the number of observations is not an even integer" )
!
        sl = period
!
    else
!
        if ( period<sl )       &
    call merror( name_proc//"Length of segments cannot be greater than the number of observations or the period length")
!
    end if
!
!   DETERMINE THE NUMBER OF FREQUENCIES.
!
    nf = ( sl/2 ) + 1
!
    if ( nsmooth>nf )    &
    call merror( name_proc//"The length of the Daniell filter must be less or equal to the number of frequencies")
!
    if ( use_mask ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( temp(nlon,nlat), mask(nlon,nlat), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET THE MASK.
!
        status = nf90_get_var( ncid_mask, maskid, temp )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       DETERMINE THE GRID POINTS USED IN THE SPECTRAL ANALYSIS FROM THE MASK.
!
        mask(:nlon,:nlat) = temp(:nlon,:nlat)==one
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
!
    else
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( mask(nlon,nlat), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       CONSTRUCT A DUMMY MASK.
!
        mask(:nlon,:nlat) = true
!
    end if
!
!   SELECT THE DOMAIN.
!
    if ( lon1>0 ) then
        mask(1:lon1-1,:nlat)    = false
        mask(lon2+1:nlon,:nlat) = false
    else
        mask(lon2+1:nlon+lon1,:nlat) = false
    end if
!
    mask(:nlon,1:lat1-1)    = false
    mask(:nlon,lat2+1:nlat) = false
!   
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   GET THE SCALE FACTOR IF ANY.
!
    status = nf90_get_att( ncid, varid, "scale_factor", scalfac )
    use_scalfac = status==nf90_noerr 
!
!   GET THE OFFSET IF ANY.
!
    status = nf90_get_att( ncid, varid, "add_offset", offset )
    use_offset = status==nf90_noerr 
!
    if ( comp_cross ) then
!
!       GET THE INDEX VARIABLE ID FROM ITS NAME.
!       FIND THE DIMENSION LENGTHS FOR THE NETCDF INDEX VARIABLE.
!
        call get_variable_dims( name_proc, trim(vari_name), ncidi,    &
                                varidi, dimids, dimlens )
!
        nidim = count(dimlens/=0)
        if ( nidim/=1 .and. nidim/=2 )     &
        call merror( name_proc//"Input index NETCDF variable is not uni- or bidimensional" )
!
        if ( nidim==2 ) then
            if ( iindex>dimlens(1) )     &
            call merror( name_proc//"Selected indice for the 2D index_variable is out of range" )
            nti = dimlens(2)
        else 
            nti = dimlens(1)
        end if
!
!       CHECK THE TIME DEFINITION FOR THE INDEX VARIABLE.
!
        itime1 = max( itime1, 1 )
        itime1 = itime1 + istep - 1
        itime2 = min( itime2, nti )
        itime3 = itime2 - itime1 + 1
!
        if ( itime3<=0  ) call merror( name_proc//"Wrong time definition for index NETCDF variable" )
!
!       CHECK THE TIME DEFINITION FOR THE INDEX VARIABLE.
!
        if ( iperiod>itime3 )  call merror( name_proc//"-ti=, -pi= options mismatch" )
!
        tmp      = real(itime3,stnd)/real(iperiod,stnd)
        niperiod = ceiling( tmp )
!
        if ( niperiod/=time3 )  call merror( name_proc//"-t=, -ti=, -pi= options mismatch" )
!
!       ALLOCATE WORK ARRAYS FOR THE TIME SERIES AND THE PSD ESTIMATES.
!
        allocate( veci(time3), psveci(nf,nperiod), psmat(nvar,nf,nperiod),      &
                  phase(nvar,nf,nperiod), coher(nvar,nf,nperiod), stat=status   )
!
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET INDEX VARIABLE.
!
        if ( nidim==1 ) then
            status = nf90_get_var( ncidi, varidi, veci(:time3),                               &   
                                   start=(/ itime1 /), count=(/ time3 /), stride=(/ iperiod /) )
        else 
            status = nf90_get_var( ncidi, varidi, veci(:time3),                               &   
                                   start=(/ iindex, itime1 /), count=(/ 1, time3 /), stride=(/ 1, iperiod /) )
        end if
!
        if ( status/=nf90_noerr )                                     &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       FIND THE MISSING VALUE ATTRIBUTE ASSOCIATED WITH THE INDEX NETCDF VARIABLE, IF ANY.
!
        status = nf90_get_att( ncidi, varidi, "missing_value", tmp )
        use_miss = status==nf90_noerr
!
        if ( .not.use_miss ) then
!
!           FIND THE _FillValue ATTRIBUTE ASSOCIATED WITH THE INDEX NETCDF VARIABLE,
!           IF NO MISSING VALUE ATTRIBUTE.
!
            status = nf90_get_att( ncidi, varidi, "_FillValue", tmp )
            use_miss = status==nf90_noerr
!
        end if
!
        if ( use_miss ) then
!
!           CHECK IF MISSING VALUES ARE PRESENT IN THE INDEX TIME SERIES.
!
            if ( any( veci(:time3)==tmp ) )      &
            call merror( name_proc//"Missing values are not allowed" )
!
        end if
!
!       MULTIPLY THE INDEX NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
        status = nf90_get_att( ncidi, varidi, "scale_factor", tmp )
        if ( status==nf90_noerr ) veci(:time3) = veci(:time3)*tmp
!
!       ADD OFFSET TO THE INDEX NETCDF VARIABLE IF REQUIRED.
!
        status = nf90_get_att( ncidi, varidi, "add_offset", tmp )
        if ( status==nf90_noerr ) veci(:time3) = veci(:time3) + tmp
!
    else
!
!       ALLOCATE WORK ARRAYS FOR THE TIME SERIES AND THE PSD ESTIMATES.
!
        allocate( psmat(nvar,nf,nperiod), stat=status   )
!
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
    end if
!
!   INITIALIZE THE FFT SUBROUTINE.
!
    nfm1 = nf - 1
    call init_fft( (/ nvar, nfm1 /), dim=2_i4b )
!
!   COMPUTE THE OFFSET FOR THE RECORD NUMBER.
!
    offset_rec = time1 - 1
!
    countlec(1:3)  = (/ lon3, lat3, 1_i4b /)
!
    if ( use_period ) then
!
!       COMPUTE THE CSD AND PSD ESTIMATES BY PERIODS.
!
#ifdef _PARALLEL_READ
!
        i = omp_get_num_procs()
        j = omp_get_max_threads()
        test_par = i>1 .and. j>1 .and.  nperiod>=j .and. time3*nvar>omp_limit
!
        if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,l,l1,l2,k,status,ncid2,varid2,startlec,temp,mat)     &
!$OMP         ,SHARED(nvar,nf,period,nperiod,lon3,lat3,j1,j2,lat1,lat2,       &
!$OMP                 varid,countlec,offset_rec,sl,trend,trend2,nsmooth,      &
!$OMP                 win,taperp,scalfac,offset,                              &
!$OMP                 use_scalfac,use_offset,comp_cross,                      &
!$OMP                 veci,mask,psveci,psmat,phase,coher)
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( temp(lon3,lat3), mat(nvar,period), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
            status = nf90_open( trim(fic_name), nf90_share, ncid2  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           GET THE VARIABLE ID FROM ITS NAME.
!
            status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
!           FIRST GET THE DATA.
!
            startlec(1)  = j1
            startlec(2)  = lat1
!
!$OMP DO SCHEDULE(STATIC)
            do i = 1, nperiod
!
                l = offset_rec + (i-1)*period
!
                do k = 1, period
!
                    l           = l + 1
                    startlec(3) = l
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE .
!
                    status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3), start=startlec, count=countlec )
                                           
                    if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE DATA .
!
                    mat(:nvar,k) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
                end do
!
!               MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
                if ( use_scalfac ) mat(:nvar,:period) = mat(:nvar,:period)*scalfac
!
!               ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
                if ( use_offset ) mat(:nvar,:period) = mat(:nvar,:period) + offset
!
!               ESTIMATE THE POWER AND CROSS SPECTRA OF THE REAL TIME SERIES.
!
                l  = (i-1)*period
                l1 = l + 1
                l2 = l + period
!
                if ( comp_cross ) then
!
                    call cross_spctrm2( veci(l1:l2), mat(:nvar,:period), sl, psveci(:nf,i), psmat(:nvar,:nf,i),     &
                                        phase(:nvar,:nf,i), coher(:nvar,:nf,i), overlap=overlap,                    &
                                        normpsd=normpsd, nsmooth=nsmooth, trend=trend, trend2=trend2,               &
                                        win=win, taperp=taperp, initfft=false                                       )
!
                else
!
                    call power_spctrm2(  mat(:nvar,:period), sl, psmat(:nvar,:nf,i), overlap=overlap,       &
                                         normpsd=normpsd, nsmooth=nsmooth, trend=trend, trend2=trend2,      &
                                         win=win, taperp=taperp, initfft=false                              )
!
                end if
!
            end do
!$OMP END DO NOWAIT
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( temp, mat )
!
!$OMP CRITICAL (closenc)
!
            status = nf90_close( ncid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
        else
!
#endif
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( temp(lon3,lat3), mat(nvar,period), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
            startlec(1:2)  = (/ j1, lat1 /)
!
            do i = 1, nperiod
!
                l = offset_rec + (i-1)*period
!
                do k = 1, period
!
                    l           = l + 1
                    startlec(3) = l
!
!                   FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE IN THE INPUT FILE .
!
                    status = nf90_get_var( ncid, varid, temp(:lon3,:lat3), start=startlec, count=countlec )
                                           
                    if ( status/=nf90_noerr )   call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!                   PACK THE DATA.
!
                    mat(:nvar,k) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
                end do
!
!               MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
                if ( use_scalfac ) mat(:nvar,:period) = mat(:nvar,:period)*scalfac
!
!               ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
                if ( use_offset ) mat(:nvar,:period) = mat(:nvar,:period) + offset
!
!               ESTIMATE THE POWER AND CROSS SPECTRA OF THE REAL TIME SERIES.
!
                l  = (i-1)*period
                l1 = l + 1
                l2 = l + period
!
                if ( comp_cross ) then
!
                    call cross_spctrm2( veci(l1:l2), mat(:nvar,:period), sl, psveci(:nf,i), psmat(:nvar,:nf,i),     &
                                        phase(:nvar,:nf,i), coher(:nvar,:nf,i), overlap=overlap,                    &
                                        normpsd=normpsd, nsmooth=nsmooth, trend=trend, trend2=trend2,               &
                                        win=win, taperp=taperp, initfft=false             )
!
                else
!
                    call power_spctrm2(  mat(:nvar,:period), sl, psmat(:nvar,:nf,i), overlap=overlap,       &
                                         normpsd=normpsd, nsmooth=nsmooth, trend=trend, trend2=trend2,      &
                                         win=win,  taperp=taperp, initfft=false                             )
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( temp, mat )
!
#ifdef _PARALLEL_READ
!
        end if
!
#endif
!
    else
!
!       COMPUTE THE PSD AND CSD ESTIMATES ON THE WHOLE TIME SERIES.
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( mat(nvar,time3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET THE DATA.
!
#ifdef _PARALLEL_READ
!
        i = omp_get_num_procs()
        j = omp_get_max_threads()
        test_par = i>1 .and. j>1 .and. time3*nvar>omp_limit .and. time3>=j
!
        if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(k,status,rec,startlec,ncid2,varid2,temp)        &
!$OMP         ,SHARED(mat,mask,varid,countlec,j1,j2,lon3,              &
!$OMP                 lat1,lat2,lat3,nvar,offset_rec,time1,time2)
!
!           ALLOCATE WORK ARRAY.
!
            allocate( temp(lon3,lat3), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
!$OMP CRITICAL (opennc)
!
            status = nf90_open( trim(fic_name), nf90_share, ncid2  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           GET THE VARIABLE ID FROM ITS NAME.
!
            status = nf90_inq_varid( ncid2, trim(var_name), varid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (opennc)
!
!           FIRST GET THE DATA.
!
            startlec(1)  = j1
            startlec(2)  = lat1
!
!$OMP DO SCHEDULE(STATIC)
!
            do k = time1, time2
!
                startlec(3) = k
!
!               FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
                status = nf90_get_var( ncid2, varid2, temp(:lon3,:lat3),                 &
                                       start=startlec, count=countlec )     
                if ( status/=nf90_noerr )                                                &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               PACK THE DATA.
!
                rec            = k - offset_rec
                mat(:nvar,rec) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
            end do
!
!$OMP END DO
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( temp )
!
!$OMP CRITICAL (closenc)
!
            status = nf90_close( ncid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!$OMP END CRITICAL (closenc)
!
!$OMP END PARALLEL
!
        else
!
#endif
!
!           ALLOCATE WORK ARRAY.
!
            allocate( temp(lon3,lat3), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
            startlec(1:2)  = (/ j1, lat1 /)
!
            do k = time1, time2
!
                startlec(3) = k
!
!               FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE .
!
                status = nf90_get_var( ncid, varid, temp(:lon3,:lat3),                      &
                                       start=startlec, count=countlec )     
                if ( status/=nf90_noerr )                                                   &
                call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!               PACK THE DATA.
!
                rec = k - offset_rec
                mat(:nvar,rec) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
            end do
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( temp )
!
#ifdef _PARALLEL_READ
!
        end if
!
#endif
!
!       MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
        if ( use_scalfac ) mat(:nvar,:time3) = mat(:nvar,:time3)*scalfac
!
!       ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
        if ( use_offset ) mat(:nvar,:time3) = mat(:nvar,:time3) + offset
!
!       ESTIMATE THE POWER AND CROSS SPECTRA OF THE REAL TIME SERIES.
!
        if ( comp_cross ) then
!
            call cross_spctrm2( veci(:time3), mat(:nvar,:time3), sl, psveci(:nf,1), psmat(:nvar,:nf,1),     &
                                phase(:nvar,:nf,1), coher(:nvar,:nf,1), overlap=overlap,                    &
                                normpsd=normpsd, nsmooth=nsmooth, trend=trend, trend2=trend2,               &
                                win=win, taperp=taperp, initfft=false                                       )
!
        else
!
            call power_spctrm2(  mat(:nvar,:time3), sl, psmat(:nvar,:nf,1), overlap=overlap,        &
                                 normpsd=normpsd, nsmooth=nsmooth, trend=trend, trend2=trend2,      &
                                 win=win, taperp=taperp, initfft=false                              )
!
       end if
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( mat )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    call end_fft( )
!
    if ( conflim ) then
!
!       COMPUTES DEGREES OF FREEDOM, CONFIDENCE LIMIT FACTORS 
!       AND CRITICAL VALUES FOR SQUARED COHERENCIES IF REQUIRED.
!
!       FIND THE FIRST INTEGER GREATER THAN OR EQUAL TO period THAT IS EVENLY
!       DIVISIBLE BY sl.
!
        i = mod( period, sl )
!
        if ( i==0 ) then
            n  = period
        else
            n  = period + sl - i
        end if
!
!       DETERMINE THE NUMBER OF SEGMENTS.
!
        if ( overlap ) then
            nseg = (2*n/sl) - 1
        else
            nseg = n/sl
        end if
!
        if ( nseg/=1 .and. overlap .and. win/=1 .and. win/=3 )       &
        call merror( name_proc//tseries_error58 )
!
!       ALLOCATE WORK ARRAYS.
!
        if ( comp_cross ) then
            allocate( edof(nf), conlwr(nf), conupr(nf), testcoher(nf), wk(sl), stat = status )
        else
            allocate( edof(nf), conlwr(nf), conupr(nf), wk(sl), stat = status )
        end if
!
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       CALCULATE DATA WINDOW.
!
        if ( win/=2 ) then
!
!           CALCULATE DATA WINDOW.
!
            wk(:sl)  = data_window( sl, win, taperp=taperp )
!
        else
!
!           CALCULATE RECTANGULAR WINDOW.
!
            wk(:sl)  = one
!
        end if
!
!       COMPUTES DEGREES OF FREEDOM.
!
        edof(:nf) = estim_dof2( wk(:sl), 0_i4b, win=win, nsmooth=nsmooth, nseg=nseg, overlap=overlap )
!
        if ( comp_cross ) then
            call comp_conflim( edof(:nf), probtest=probtest, conlwr=conlwr(:nf),     &
                               conupr=conupr(:nf), testcoher=testcoher(:nf)          )
        else
            call comp_conflim( edof(:nf), probtest=probtest, conlwr=conlwr(:nf),     &
                               conupr=conupr(:nf)                                    )
        end if
!
    end if
!
!   CREATE THE NEW NETCDF DATASET FOR THE FREQUENCY ANALYSIS.
!
    if ( create_hdf5 ) then
        status = nf90_create( trim(spect_name), cmode=output_mode, ncid=ncid_spect )
    else
        k = nf90_sizehint_default
        status = nf90_create( trim(spect_name), cmode=output_mode, ncid=ncid_spect, chunksize=k )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_spect, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = nlon
    dimlens(2)   = nlat
    dimlens(3)   = nf
!
    if ( use_tlimited ) then
        dimlens(4)   = nperiod
    else
        dimlens(4)   = nf90_unlimited
    end if
!
    coor_type(3) = output_type
    coor_type(4) = nf90_int
!
    dim_names(3) = "frequency"
    dim_names(4) = "time"
!
    do i = 1, 4
        status = nf90_def_dim( ncid_spect, name=trim(dim_names(i)),                       &
                               len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 4
        status = nf90_def_var( ncid_spect, trim(dim_names(i)), coor_type(i),              &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end do
!
    status = nf90_def_var( ncid_spect, "periodicity",  output_type,   &
                           dimids=dimids(3), varid=periodid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( save_spect ) then
!
        status = nf90_def_var( ncid_spect, trim(var_name)//"_spectrum",  output_type,   &
                               dimids=dimids(1:4), varid=spectid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    if ( comp_cross ) then
!
!       CHECK IF THE TWO NETCDF VARIABLES HAVE NOT THE SAME NAME BEFORE CREATING NEW NETCDF VARIABLES.
!
        if ( trim(vari_name)==trim(var_name) ) vari_name = trim(vari_name)//"_i"
!
        status = nf90_def_var( ncid_spect, trim(vari_name)//"_spectrum",  output_type,   &
                               dimids=dimids(3:4), varid=varispectid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        name = trim(vari_name)//"_"//trim(var_name)
!
        status = nf90_def_var( ncid_spect, trim(name)//"_squaredcoherency",  output_type,   &
                               dimids=dimids(1:4), varid=coherid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_def_var( ncid_spect, trim(name)//"_phase",  output_type,   &
                               dimids=dimids(1:4), varid=phaseid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    if ( conflim ) then
!
        status = nf90_def_var( ncid_spect, "edof",  output_type,   &
                               dimids=dimids(3), varid=edofid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_def_var( ncid_spect, "lower_factor",  output_type,   &
                               dimids=dimids(3), varid=conlwrid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_def_var( ncid_spect, "upper_factor",  output_type,   &
                               dimids=dimids(3), varid=conuprid )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!
        if ( comp_cross ) then
!
            status = nf90_def_var( ncid_spect, "coher_test",  output_type,   &
                                   dimids=dimids(3), varid=testcoherid )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end if
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 2
        if ( .not.use_coor_variable(j) ) cycle
        do i = 1, coor_natts(j)
            status = nf90_inq_attname( ncid, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid, coor_id(j), att_name, ncid_spect, coor_id2(j) )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
        end do
    end do
!
    status = nf90_put_att( ncid_spect, coor_id2(3), "long_name", "Frequency axis" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, coor_id2(3), "short_name", "Frequency" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, coor_id2(3), "units", "cycles per unit of time" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, periodid, "long_name", "Periodicity axis" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, periodid, "short_name", "Periodicity" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, periodid, "units", "number of time observations" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, coor_id2(4), "long_name", "Time axis" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, coor_id2(4), "short_name", "Time" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   SET missing_value, _FillValue, valid_min AND valid_max ATTRIBUTES.
!
    if ( comp_cross ) then
        id_var(1:4) = (/ periodid, varispectid, coherid, phaseid /)
        l1 = 4
    else
        id_var(1) = periodid
        l1 = 1
    end if
!
    if ( save_spect ) then
        l1         = l1 + 1
        id_var(l1) = spectid
    end if
!
    if (  output_type==nf90_real8 ) then
!
        do i=1, l1
            status = nf90_put_att( ncid_spect, id_var(i), "missing_value", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_spect, id_var(i), "_FillValue", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end do
!
        if ( comp_cross ) then
!
            status = nf90_put_att( ncid_spect, coherid, "valid_min", real( zero, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_spect, coherid, "valid_max", real( one, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_spect, phaseid, "valid_min", real( zero, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_spect, phaseid, "valid_max", real( one, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            if ( conflim ) then
!
                status = nf90_put_att( ncid_spect, testcoherid, "valid_min", real( zero, dp ) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_put_att( ncid_spect, testcoherid, "valid_max", real( one, dp ) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            end if
!
        end if
!
    else
!
        do i=1, l1
            status = nf90_put_att( ncid_spect, id_var(i), "missing_value", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_spect, id_var(i), "_FillValue", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end do
!
        if ( comp_cross ) then
!
            status = nf90_put_att( ncid_spect, coherid, "valid_min", real( zero, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_spect, coherid, "valid_max", real( one, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_spect, phaseid, "valid_min", real( zero, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_spect, phaseid, "valid_max", real( one, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            if ( conflim ) then
!
                status = nf90_put_att( ncid_spect, testcoherid, "valid_min", real( zero, sp ) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
                status = nf90_put_att( ncid_spect, testcoherid, "valid_max", real( one, sp ) )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            end if
!
        end if
!
    end if
!
    if ( save_spect ) then
!
        status = nf90_put_att( ncid_spect, spectid, "long_name", "Power Spectral Density of "//trim(var_name)  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, spectid, "short_name", "Power Spectral Density"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( normpsd ) then
!
            status = nf90_put_att( ncid_spect, spectid, "units", "spectral density units"  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        else
!
            name2 = " "
!
            status = nf90_get_att( ncid, varid, "units", name2 )
            if ( status==nf90_noerr ) then
                status = nf90_put_att( ncid_spect, spectid, "units", "("//trim(name2)//")**2 "  )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
            end if
!
        end if
!
    end if
!
    if ( conflim ) then
!
        status = nf90_put_att( ncid_spect, edofid, "long_name",                              &
                               "Equivalent number of degrees of freedom of PSD/CSD estimates"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, edofid, "short_name", "Number of degrees of freedom"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, conlwrid, "long_name",                              &
                               "Lower (1-critical_probability)*100% confidence limit factor"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, conlwrid, "short_name", "Lower limit factor"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, conuprid, "long_name",                              &
                               "Upper (1-critical_probability)*100% confidence limit factor"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, conuprid, "short_name", "Upper limit factor"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    if ( comp_cross ) then
!
        if ( nidim==1 ) then
            name = trim(vari_name)
        else
            call val_to_string( iindex, fmt, i )
            name = trim(vari_name)//"("//fmt(1:i)//")"
        end if
!
        status = nf90_put_att( ncid_spect, varispectid, "long_name", "Power Spectral Density of "//trim(name)  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, varispectid, "short_name", "Power Spectral Density"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( normpsd ) then
!
            status = nf90_put_att( ncid_spect, varispectid, "units", "spectral density units"  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        else
!
            name2 = " "
!
            status = nf90_get_att( ncidi, varidi, "units", name2 )
            if ( status==nf90_noerr ) then
                status = nf90_put_att( ncid_spect, varispectid, "units", "("//trim(name2)//")**2 "  )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
            end if
!
        end if
!
        name2 = trim(name)//" and "//trim(var_name)
!
        status = nf90_put_att( ncid_spect, coherid, "long_name", "Squared Coherency between "//trim(name2)  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, coherid, "short_name", "Squared Coherency"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, coherid, "units", "None" )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, phaseid, "long_name", "Phase of Cross Spectrum between "//trim(name2)  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, phaseid, "short_name", "Phase"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, phaseid, "units", "fractions of a circle" )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        if ( conflim ) then
!
            status = nf90_put_att( ncid_spect, testcoherid, "long_name",                                  &
                      "critical value for squared coherency at the critical_probability*100% significance level"  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_spect, testcoherid, "short_name", "critical value for squared coherency"  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
        name = " Cross Spectrum analysis between "//trim(name2)
!
    else
!
        name = " Spectrum analysis of "//trim(var_name)
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_spect, nf90_global, "title", trim(name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( comp_cross ) then
!
        status = nf90_put_att( ncid_spect, nf90_global, "iperiod", iperiod )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_spect, nf90_global, "istep", istep )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    status = nf90_put_att( ncid_spect, nf90_global, "period_length", period )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "segment_length", sl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "trend_removal", trend )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "trend_removal_segment", trend2 )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "window_choice", win )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( win==6 ) then
!
        status = nf90_put_att( ncid_spect, nf90_global, "tapered_percentage", taperp )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    if ( conflim ) then
!
        status = nf90_put_att( ncid_spect, nf90_global, "critical_probability", probtest )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    if ( normpsd ) then
        i = 1
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_spect, nf90_global, "normalization", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( smooth ) then
!
        status = nf90_put_att( ncid_spect, nf90_global, "Daniell_filter_length", nsmooth )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        i = 1
!
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_spect, nf90_global, "smoothing", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( overlap ) then
        i = 1
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_spect, nf90_global, "overlapping", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_spect, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_spect )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
    i = max( nlon, nlat )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( record(i), freq(nf), temp(nlon,nlat), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   WRITE GEOGRAPHICAL COORDINATE VARIABLES.
!
    do j = 1, 2
!
        j2 = dimlens(j)
!
        if ( use_coor_variable(j) ) then
!
            status = nf90_get_var( ncid, coor_id(j), record(:j2), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_spect, coor_id2(j), record(:j2), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        else
            status = nf90_put_var( ncid_spect, coor_id2(j), arth(1_i4b,1_i4b,j2 ) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end do
!
!   WRITE FREQUENCY AND PERIOD AXES.
!
    tmp = one/real( sl, stnd )
    freq(:nf) = arth( zero, tmp, nf )
!
    status = nf90_put_var( ncid_spect, coor_id2(3), freq(:nf) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_var( ncid_spect, coor_id2(4), arth(1_i4b,1_i4b,nperiod) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   BUILD UP PERIODICITY VECTOR.
!
    freq(1)    = missing
    freq(2:nf) = one/freq(2:nf)
!
    status = nf90_put_var( ncid_spect, periodid, freq(:nf) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    startwri(1:2) = 1
    countwri(1:4) = (/ nlon, nlat, 1_i4b, 1_i4b /)
!
    if ( save_spect ) then
!
!       WRITE THE PSD ESTIMATES.
!
        do i = 1, nperiod
!
            startwri(4) = i
!
            do l = 1, nf
!
                startwri(3) = l
!
                temp(:nlon,:nlat) = unpack( psmat(:nvar,l,i), mask(:nlon,:nlat), missing )
!
                status = nf90_put_var( ncid_spect, spectid, temp, start=startwri, count=countwri )
                if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            end do
!
        end do
!
    end if
!
    if ( conflim ) then
!
!       WRITE THE NUMBER OF DEGREES OF FREEDOM AND THE CONFIDENCE LIMIT FACTORS.
!
        status = nf90_put_var( ncid_spect, edofid, edof(:nf) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_var( ncid_spect, conlwrid, conlwr(:nf) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_var( ncid_spect, conuprid, conupr(:nf) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    if ( comp_cross ) then
!
!       WRITE THE PSD ESTIMATES FOR THE INDEX TIME SERIES.
!
        status = nf90_put_var( ncid_spect, varispectid, psveci(:nf,:nperiod) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       WRITE THE SQUARED COHERENCY ESTIMATES.
!
        do i = 1, nperiod
!
            startwri(4) = i
!
            do l = 1, nf
!
                startwri(3) = l
!
                temp(:nlon,:nlat) = unpack( coher(:nvar,l,i), mask(:nlon,:nlat), missing )
!
                status = nf90_put_var( ncid_spect, coherid, temp, start=startwri, count=countwri )
                if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            end do
!
        end do
!
!       WRITE THE PHASE ESTIMATES.
!
        do i = 1, nperiod
!
            startwri(4) = i
!
            do l = 1, nf
!
                startwri(3) = l
!
                temp(:nlon,:nlat) = unpack( phase(:nvar,l,i), mask(:nlon,:nlat), missing )
!
                status = nf90_put_var( ncid_spect, phaseid, temp, start=startwri, count=countwri )
                if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            end do
!
        end do
!
        if ( conflim ) then
!
!           WRITE THE CRITICAL VALUES FOR THE SQUARED COHERENCY.
!
            status = nf90_put_var( ncid_spect, testcoherid, testcoher(:nf) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( record, temp, freq, mask, veci, psmat, psveci, phase, coher,   &
                        edof, conlwr, conupr, testcoher, wk      )
!
        else
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( record, temp, freq, mask, veci, psmat, psveci, phase, coher )
!
        end if
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        if ( conflim ) then
            deallocate( record, temp, freq, mask, psmat, edof, conlwr, conupr, wk )
        else
            deallocate( record, temp, freq, mask, psmat )
        end if
!
    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( comp_cross ) then
!
        if ( ncidi/=ncid ) then
!
            status = nf90_close( ncidi )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end if
!
    if ( use_mask ) then
        status = nf90_close( ncid_mask )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_spect )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_cross_spectrum_3d
! =====================================
!
end program comp_cross_spectrum_3d
