program comp_fftfilter_1d
!! #############################################################################
!!
!! Copyright 2024 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_fftfilter_1d:
!
! =================
! comp_fftfilter_1d
! =================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 28/05/2024
!
! Purpose
! =======
!
! Filter a real time series in a selected frequency band by windowed filtering [Iacobucci_Noullez]_ .
! The time series is extracted from a uni- or bidimensional variable readed from
! a NetCDF dataset and can also be detrended before windowed filtering at the user option.
!
! The windowed filter is applied to the time series in the frequency domain [Iacobucci_Noullez]_ .
! The filter is obtained by convolving a raised-cosine window with the ideal rectangular filter
! response function. This filter is stationary and symmetric, therefore, it induces no phase-shift
! and is a good candidate for extracting frequency-defined series components from short-length
! time series.
!
! Applying the filter in the frequency domain assumes implicitly that the time series is part
! of a periodic infinite series whose period is exactly equal to the length of the analyzed time series.
! The selected ideal filter gain vector (in the frequency domain) is convolved with the Fourier transform
! of a raised-cosine window to get the final filter frequency response.
! This filter frequency response is then convolved with the Fourier transform of the time series and the resulting
! sequence is inverted back in the time domain to obtain the final filtered time series. The specific form of the
! raised-cosine window can be controlled by the user with the help of the **-win=** argument, described below.
!
! Additionally, the filtering can be done separately for different segments of equal length of the selected
! time series if this time series is not continuous in time (see the **-p=** argument).
!
! This procedure returns the filtered real time series in a NetCDF dataset. If the NetCDF variable
! is trid or fourdimensional use :ref:`comp_fftfilter_3d` or :ref:`comp_fftfilter_4d`, respectively,
! instead of comp_fftfilter_1d. If the time series has a seasonal (or diurnal) cycle,
! use :ref:`comp_stl_1d` in order to estimate and remove the harmonic components of the time series
! before using comp_fftfilter_1d.
!
! Finally, use :ref:`comp_spectvar_1d` to compute variance estimates in the selected frequency band for
! the real time series before or after filtering the real time series.
!                                                                              
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!                                                                              
!       $ comp_fftfilter_1d \
!         -f=input_netcdf_file \
!         -v=netcdf_variable \
!         -t=time1,time2                           (optional) \
!         -o=output_netcdf_file                    (optional) \
!         -ni=index_for_2d_netcdf_variable         (optional) \
!         -p=period_length                         (optional) \
!         -pl=minimum_period                       (optional) \
!         -ph=maximum_period                       (optional) \
!         -tr=trend_removal                        (optional : 0, 1, 2, 3, -1, -2, -3) \
!         -win=window_choice                       (optional : 0.5 > 1.) \
!         -mi=missing_value                        (optional) \
!         -double                                  (optional) \
!         -hdf5                                    (optional) \
!         -tlimited                                (optional)
!
! By default
! ----------
!
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-o=*
!         the *output_netcdf_file* is named ``filt_``\ *netcdf_variable*\ ``.nc``
!        *-ni=*
!         if the *netcdf_variable* is bidimensional, the first time series is used
!        *-p=*
!         the *period_length* is set to ``time2 - time1 + 1`` , which means
!         that the time series is considered as continuous with only one time segment
!        *-pl=*
!         the *minimum_period* is set to ``0``, which means that no filtering is done
!         for the shorter periods
!        *-ph=*
!         the *maximum_period* is set to ``0``, which means that no filtering is done
!         for the longer periods
!        *-tr=0*
!         *trend_removal* is set to ``0``, which means that no detrending is done before
!         filtering
!        *-win=0.54*
!         the Hamming window is convolved with the ideal filter response
!        *-mi=*
!         the *missing_value* for the output variable is equal to ``1.e+20``
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable for which a
!      windowed filtering operation must be computed and the **-f=**\ *input_netcdf_file* 
!      argument specifies that this NetCDF variable must be extracted from the NetCDF file
!      *input_netcdf_file*.
!
!   #) If the **-t=**\ *time1,time2* argument is missing the whole time period
!      associated with the *netcdf_variable* is used to decompose the time series.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` time observations.
!
!   #) The **-ni=**\ *index_for_2d_netcdf_variable* argument specifies the index for selecting the
!      time series if the *netcdf_variable* is a 2D NetCDF variable. By default, the first time series
!      is used, which is equivalent to set *index_for_2d_netcdf_variable* to ``1``\ .
!
!   #) If the **-p=** argument is specified, the filtering is applied separately for each time segment
!      of length *period_length* (as determined by the value of the **-p=** argument).
!      The whole selected time period (e.g., *time2* - *time1* + ``1`` ) must also be a multiple of 
!      the *period_length*.
!
!   #) The **-pl=** argument specifies the minimum period of oscillation of the filtered time series. 
!      The *minimum_period* is expressed in number of time observations. 
!
!      Do not use the **-pl=** argument or use **-pl=**\ ``0`` for high-pass filtering 
!      frequencies corresponding to periods shorter than **-ph=**\ ``PH`` .
!
!      The **-pl=** argument is a positive integer equal to ``0`` or greater than ``2``.
!
!   #) The **-ph=** argument specifies the maximum period of oscillation of the filtered time series. 
!      The *maximum_period* is expressed in number of time observations. 
!      Do not use the **-ph=** argument or use **-ph=**\ ``0`` for low-pass filtering 
!      frequencies corresponding to periods longer than **-pl=**\ ``PL`` .
!      For example, **-pl=**\ ``6``  (or ``18``) and **-ph=**\ ``32``  (or ``96``) select periods
!      between ``1.5`` and ``8``  years for quarterly (monthly) time series.
!
!      The **-ph=** argument is a positive integer equal to ``0`` or greater than ``2`` and 
!      less than the length of the time series or the *period_length* if the **-p=**
!      argument is used.
!
!      The **-ph=** argument must also be greater or equal to the **-pl=** argument if 
!      both are specified.
!
!   #) Setting **-pl=** and **-ph=** to the same value ``P`` is allowed. 
!      In this case, an -ideal- band-pass filter with peak response near one
!      at the single period ``P`` is computed and applied to the time series.
!                                                                              
!   #) Setting **-pl=**\ ``PL``, **-ph=**\ ``PH`` and ``PH``<``PL`` is also allowed and
!      performs band rejection of periods between ``PH`` and ``PL``. In that case, the meaning
!      of the **-pl=** and **-ph=** arguments reversed.                  
!
!   #) Setting both **-pl=**\ ``0`` and **-ph=**\ ``0`` is also allowed. In that case,
!      no frequencies filtering is done, but the data may be detrended if the **-tr=** argument 
!      is used with a value of ``1``, ``2`` or ``3``.
!
!   #) The **-tr=** argument specifies pre- and post-filtering processing of the time series. If:
!
!      * **-tr=**\ ``+/-1``, the mean of the time series is removed before time filtering
!
!      * **-tr=**\ ``+/-2``, the drift from the time series is removed before time 
!        filtering. The drift for the time series is estimated using the 
!        formula : ``drift = ( tseries(ntime) - tseries(1) )/( ntime - 1 )``
!
!      * **-tr=**\ ``+/-3``, the least-squares line from the time series is removed before
!        time filtering.
!
!      If **-tr=**\ ``-1``, ``-2`` or ``-3``, the mean, drift or least-squares line are reintroduced
!      post-filtering, respectively.
!
!      For other values of the **-tr=** argument, nothing is done before or after 
!      filtering.
!
!      If the **-p=** argument is present, the pre-filtering and post-filtering processing is applied to 
!      each time segment, separately.
!
!      The **-tr=** argument must be an integer and the default value for the **-tr=** argument is ``0``.
!
!   #) The **-win=** argument controls the form of the window, which will be convolved with the ideal filter response.
!      By default, a Hamming window is used (e.g., **-win=**\ ``0.54``).
!
!      Set **-win=**\ ``0.5`` for using a Hanning window or **-win=**\ ``1.`` for a rectangular
!      window  (e.g., the "ideal" filter).
!                                                        
!      The **-win=** argument is a real number greater or equal to ``O.5`` and less or equal to ``1.``.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator for
!      the output variable in the *output_netcdf_file*. 
!
!      If the **-mi=** argument is not specified, the *missing_value* is set
!      to ``1.e+20``.
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g., ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) It the time series has a seasonal or diurnal cycle, use :ref:`comp_stl_1d`
!      to remove the pure harmonic components from the time series before filtering.
!
!   #) It is assumed that the data has no missing values.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence
!      of a parameter which will be used for the computations. Moreover, the
!      number of specified parameters must not be greater than the total number
!      of allowed parameters.
!                                                                              
!   #) For more details on windowed filtering, see
!                                                                              
!      * "A Frequency Selective Filter for Short-Length Time Series", 
!         by Iacobucci, A., and Noullez, A., Computational Economics, Vol. 25, 75-102, 2005. https://link.springer.com/article/10.1007/s10614-005-6276-7
!
! Outputs
! -------
!
!   comp_fftfilter_1d creates an output NetCDF file that contains the filtered time series estimated from the
!   time series associated with the input NetCDF variable. The output NetCDF dataset contains the
!   following NetCDF variable (in the description below, ntime is the selected number of time observations) :
!
!   #) *netcdf_variable*\ _filt\ ``(ntime)`` : the filtered time series for the time series 
!      associated with the input NetCDF variable.
!                                                                              
! Examples
! --------
!
!   #)   For windowed filtering a real monthly time series between ``18`` and ``30``
!        months (e.g., biennial time scale) from a NetCDF variable called ``sst``
!        extracted from the file ``sst.monthly.nino34.nc``, which includes a monthly time
!        series, and store the results in the NetCDF file ``qbo_sst_nino34.nc``, use the following
!        command :
!
!        ::
!
!          $ comp_fftfilter_1d  \
!            -f=sst.monthly.nino34.nc \
!            -v=sst \
!            -pl=18 \
!            -ph=30 \
!            -o=qbo_sst_nino34.nc
!                                                                              
!-
! ================================================================================================
!
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!   
    use netcdf,   only : nf90_max_name, nf90_max_var_dims, nf90_clobber, nf90_real4, nf90_real8,   &
                         nf90_sizehint_default, nf90_clobber, nf90_open,                           &
                         nf90_nowrite, nf90_noerr, nf90_get_att, nf90_get_var, nf90_strerror,      &
                         nf90_create, nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_unlimited,    &
                         nf90_def_var, nf90_inq_attname, nf90_copy_att, nf90_put_att, nf90_enddef, &
                         nf90_global, nf90_put_var, nf90_close
!   
    use Statpack, only : i1b, i4b, stnd, extd, lgl, true, false, one, half, kfix, kexp, kint, tolower,     & 
                         tseries_error13, case_change, string_to_val, val_to_string,                       &
                         allocate_error, merror, arth, detrend, hwfilter, init_fft, end_fft
!   
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing,                              &
                                 get_listarg, get_variable_dimcoords
!   
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_fftfilter_1d : '
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_filt,                           &
                     varid, filtid, coor_id2,                                 &
                     i, i1, i2, iindex, ndim, nt, var_natts,                  &
                     time1, time2, time3,                                     &
                     pl, ph, trend, output_type, output_mode, period, nperiod
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),   allocatable    :: vec
    real(stnd)                                 :: missing, win, tmp
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                               :: use_tlimited, detrend_only, use_miss, use_period
    logical(lgl), dimension(nf90_max_var_dims) :: use_coor_variable
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: fic_name, var_name, filt_name, att_name 
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                                          :: fmt
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
! 
    if ( narg==0 ) then
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Filter a real time series from a uni or bidimensional variable extracted'
        write (prtunit,*) ' from a NetCDF dataset in a selected frequency band by windowed filtering'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_fftfilter_1d  '
        write (prtunit,*) '        -f=input_netcdf_file'
        write (prtunit,*) '        -v=netcdf_variable'
        write (prtunit,*) '        -t=time1,time2                   (optional)'
        write (prtunit,*) '        -o=output_netcdf_file            (optional)'
        write (prtunit,*) '        -ni=index_for_2d_netcdf_variable (optional)'
        write (prtunit,*) '        -p=period_length                 (optional)'
        write (prtunit,*) '        -pl=minimum_period               (optional)'
        write (prtunit,*) '        -ph=maximum_period               (optional)'
        write (prtunit,*) '        -tr=trend_removal                (optional : 0,1,2,3,-1,-2,-3)'
        write (prtunit,*) '        -win=window_choice               (optional : 0.5 > 1.)'
        write (prtunit,*) '        -mi=missing_value                (optional)'
        write (prtunit,*) '        -double                          (optional)'
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                            (optional)'
#endif
        write (prtunit,*) '        -tlimited                        (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-t= the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-o=filt_netcdf_variable.nc'
        write (prtunit,*) '-ni=1            : by default if the netcdf_variable is bidimensional,'
        write (prtunit,*) '                   the first time series is used'
        write (prtunit,*) '-p=time2-time1+1 : by default, the period_length argument is not used'
        write (prtunit,*) '-pl=0            : by default, no filtering is done'
        write (prtunit,*) '-ph=0            : by default, no filtering is done'
        write (prtunit,*) '-tr=0            : by default, no trend removal is done'
        write (prtunit,*) '-win=0.54        : by default, the Hamming window is convolved with the ideal filter response'
        write (prtunit,*) '-mi=1.e+20       : by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-double          : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5            : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited        : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<2 .or. narg>14 ) then
        call merror( name_proc//"Wrong number of line arguments" )
    end if
!
    fic_name   = " "
    var_name   = " "
    filt_name  = " "
!
    history = "comp_fftfilter_1d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
!
    iindex        = 1
    pl            = 0
    ph            = 0
    trend         = 0
    win           = 0.54_stnd
!
    missing = def_missing
!
    time1     = 1
    time2     = huge( time2 )
!
    period    = huge( period )
    nperiod   = 1
!
    detrend_only = false
    use_tlimited = false
    use_period   = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-p=')
!
!               GET THE PERIOD LENGTH.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode/=kint ) &
                call merror( name_proc//"The period length is an integer value" )
!
                read( name(4:), fmt ) period
!
                if ( period<=0 ) &
                call merror( name_proc//"The -p= parameter is a positive integer" )
!
                use_period = true
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                filt_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-ni=')
!
!                       GET THE INDEX FOR A 2D VARIABLE.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) iindex
                        end if
                        if ( kcode/=kint .or. iindex<=0  )                           &
                        call merror( name_proc//"The index for a 2D variable is a positive integer value" )
!
                    case ('-pl=')
!
!                       GET THE MINIMUM PERIOD OF OSCILLATION.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) pl
                        end if
                        if ( kcode/=kint .or. pl<0 .or. pl==1  )                           &
                        call merror( name_proc//"The -pl= argument is an integer equal to 0 or greater than 1" )
!
                    case ('-ph=')
!
!                       GET THE MAXIMUM PERIOD OF OSCILLATION.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) ph
                        end if
                        if ( kcode/=kint .or. ph<0 .or. ph==1  )                           &
                        call merror( name_proc//"The -ph= argument is an integer equal to 0 or greater than 1" )
!
                    case ('-tr=')
!
!                       DETERMINE PRE- AND POST-FILTERING COMPUTATIONS.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) trend
                        else
                            call merror( name_proc//"The -tr= argument is an integer value" )
                        end if
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                        call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:5) )
!
                            case ('-win=')
!
!                               DETERMINE THE FORM OF THE WINDOW.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kfix .or. kcode==kexp ) then
                                    read( name(6:), fmt ) win
                                end if
                                if ( (kcode/=kfix .and. kcode/=kexp) .or.     &
                                     win<half                        .or.     &
                                     win>one )                                &
      call merror( name_proc//"The -win= argument must be a real number and verify 0.5<=win<=1. " )
!
                            case default
!
                                select case ( name(1:) )
!
                                    case ('-double')
!
!                                       STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                        output_type = nf90_real8
!
#ifdef _USE_NETCDF4
                                    case ('-hdf5')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_hdf5
#endif
!
                                    case ('-tlimited')
!
!                                       DEFINE THE TIME DIMENSION AS LIMITED.
!
                                        use_tlimited = true
!
                                    case default
!
!                                       WRONG ARGUMENT.
!
                                        call merror( name_proc//"Wrong argument" )
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( filt_name==" " ) filt_name="filt_"//trim(var_name)//".nc"
!
!   CHECK PERIODS ORDERING TO SELECT FREQUENCIES.
!
    if ( max(pl,ph)==0_i4b ) then
        if ( trend>=1 .and. trend<=3 ) then
            detrend_only = true
        else
            call merror( name_proc//"Nothing will be done on input NETCDF variable" )
        end if
    end if
!
!   OPEN THE NETCDF DATASET.
!
    i = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=i )
    if ( status/=nf90_noerr )                                                &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    call get_variable_dimcoords(  name_proc, trim(var_name), ncid,                          &
                                  varid,  dimids_out=dimids, dim_names_out=dim_names,       &
                                  dimlens_out=dimlens, coor_id_out=coor_id,                 &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type,       &
                                  use_coor_variable_out=use_coor_variable, natts=var_natts  )
!
    ndim = count(dimlens/=0)
    if ( ndim/=1 .and. ndim/=2 )     &
    call merror( name_proc//"Input NETCDF variable is not uni or bidimensional" )
!
    if ( ndim==2 ) then
        if ( iindex>dimlens(1) )     &
        call merror( name_proc//"Selected index for 2D input NETCDF variable is out of range " )
        nt = dimlens(2)
    else 
        nt = dimlens(1)
    end if
!
!   CHECK THE TIME DEFINITION AND PERIOD LENGTH DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<4  ) call merror( name_proc//"Wrong time definition or not enough time observations" )
!
    if ( use_period ) then
!
        if ( period<4  ) call merror( name_proc//"Not enough time observations in the periods" )
!
        nperiod = time3/period
        if ( nperiod*period/=time3 )    &
        call merror( name_proc//"The selected time period is not a multiple of the period length" )
!
    else
!
        period = time3
!
    end if
!
    if ( period<max(pl,ph) )   call merror( name_proc//tseries_error13 )
!
!   ALLOCATE WORK ARRAY FOR THE TIME SERIES.
!
    allocate( vec(time3), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE DATA.
!
    if ( ndim==1 ) then
        status = nf90_get_var( ncid, varid, vec(:time3),              &   
                               start=(/ time1 /), count=(/ time3 /)   )
    else 
        status = nf90_get_var( ncid, varid, vec(:time3),                          &   
                               start=(/ iindex, time1 /), count=(/ 1, time3 /)    )
    end if
!
    if ( status/=nf90_noerr )                                     &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE MISSING VALUE ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "missing_value", tmp )
    use_miss = status==nf90_noerr
!
    if ( .not.use_miss ) then
        status = nf90_get_att( ncid, varid, "_FillValue", tmp )
        use_miss = status==nf90_noerr
    end if
!
    if ( use_miss ) then
        if ( any( vec(:time3)==tmp ) )      &
        call merror( name_proc//"Missing values are not allowed" )
    end if
!
!   MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
    status = nf90_get_att( ncid, varid, "scale_factor", tmp )
    if ( status==nf90_noerr ) vec(:time3) = vec(:time3)*tmp
!
!   ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
    status = nf90_get_att( ncid, varid, "add_offset", tmp )
    if ( status==nf90_noerr ) vec(:time3) = vec(:time3) + tmp
!
!   INITIALIZE THE FFT SUBROUTINE.
!
    if ( .not.detrend_only ) then
        call init_fft( period  )
    end if
!
!   FILTER THE TIME SERIES BY PERIODS.
!
    do i = 1, nperiod
!
        i1 = (i-1)*period +1
        i2 = i*period
!
        if ( detrend_only ) then
!
!           DETREND THE TIME SERIES, ONLY.
!
           call detrend( vec(i1:i2), trend )
!
        else
!
!           FREQUENCIES FILTERING OF THE REAL TIME SERIES IS REQUIRED.
!
            call hwfilter( vec(i1:i2), pl, ph, initfft=false, trend=trend, win=win )
!
        end if
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( .not.detrend_only ) then
        call end_fft( )
    end if
!
!   CREATE THE NEW NETCDF DATASET FOR THE FREQUENCY ANALYSIS.
!
    status = nf90_create( filt_name, cmode=output_mode, ncid=ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_filt, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    if ( use_tlimited ) then
        dimlens(ndim)   = time3
    else
        dimlens(ndim)   = nf90_unlimited
    end if
!
    status = nf90_def_dim( ncid_filt, name=trim(dim_names(ndim)),                         &
                           len=dimlens(ndim), dimid=dimids(ndim) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    status = nf90_def_var( ncid_filt, trim(dim_names(ndim)), coor_type(ndim),         &
                           dimids=dimids(ndim), varid=coor_id2 )
    if ( status/=nf90_noerr )                                                         &
    call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_filt, trim(var_name)//"_filt",  output_type,              &
                           dimids=dimids(ndim), varid=filtid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    if ( use_coor_variable(ndim) ) then
!
        do i = 1, coor_natts(ndim)
!
            status = nf90_inq_attname( ncid, coor_id(ndim), i, att_name )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLE.
!
            status = nf90_copy_att(  ncid, coor_id(ndim), att_name, ncid_filt, coor_id2 )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    else
!
        status = nf90_put_att( ncid_filt, coor_id2, "long_name", "Time axis" )
        if ( status/=nf90_noerr )                                                          &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, coor_id2, "short_name", "Time" )
        if ( status/=nf90_noerr )                                                          &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   COPY ATTRIBUTES FOR THE FILTERED VARIABLE.
!
    do i = 1, var_natts
!
        status = nf90_inq_attname( ncid, varid, i, att_name )
        if ( status/=nf90_noerr )  call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( att_name=="add_offset"         .or.  &
             att_name=="scale_factor"       .or.  &
             att_name=="missing_value"      .or.  &
             att_name=="_FillValue"               )   cycle
!
!       COPY USEFUL ATTRIBUTES.
!
        status = nf90_copy_att(  ncid, varid, att_name, ncid_filt, filtid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_filt, filtid, "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, filtid, "_FillValue", real( missing, dp ) )
!
    else
!
        status = nf90_put_att( ncid_filt, filtid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, filtid, "_FillValue", real( missing, sp ) )
!
    end if
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( ndim==1 ) then
        name = trim(var_name)
    else
        call val_to_string( iindex, fmt, i )
        name = trim(var_name)//"("//fmt(1:i)//")"
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    status = nf90_put_att( ncid_filt, nf90_global, "title", " Windowed filtering of "//trim(name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "minimum_period", pl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "maximum_period", ph )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "trend_removal", trend )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "window_parameter", win )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "number_of_time_segments", nperiod )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
    if ( use_coor_variable(ndim) ) then
!
        allocate( record(time3), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        status = nf90_get_var( ncid, coor_id(ndim), record, start=(/ time1 /), count=(/ time3 /) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_var( ncid_filt, coor_id2, record )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        deallocate( record )
    else
        status = nf90_put_var( ncid_filt, coor_id2, arth(time1,1_i4b,time3 ) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
!   WRITE FILTERED VARIABLE.
!
    status = nf90_put_var( ncid_filt, filtid, vec(:time3) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( vec )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_fftfilter_1d
! ================================
!
end program comp_fftfilter_1d
