program comp_loess_rot_eof_4d
!! #############################################################################
!!
!! Copyright 2024 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_loess_rot_eof_4d:
!
! =====================
! comp_loess_rot_eof_4d
! =====================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 29/04/2024
!
! Purpose
! =======
!
! Performs an orthogonal rotation of selected standardized Principal Components (PC) time series
! extracted from a previous Empirical Orthogonal Function (EOF) analysis or Principal Component Analysis
! (PCA) of a fourdimensional NetCDF variable towards low-frequency or high-frequency components using a LOESS
! smoother [Wills_etal]_  [Cleveland]_  [Cleveland_Devlin]_. The associated factor loading matrix (e.g., the associated
! scaled EOF patterns corresponding to the selected rotated PC time series) are also rotated and the new rotated EOF
! or PCA model is stored in a NetCDF file.
!
! Spatial EOF patterns (e.g., scaled EOFs) and PC time series derived from EOF analysis or PCA
! are constrained to be both spatially and temporally orthogonal to one another so that they maximize
! the variance they described over the entire analysis domain and selected time period [Jolliffe]_ 
! [vonStorch_Zwiers]_ . However, the spatial orthogonality of the EOFs is a strong and undesirable constraint
! imposed on the EOFs and, as a result, EOFs are subject to problems such as domain dependence and inaccurate
! representation of physical relationships embedded in the data [Jolliffe]_  [Richman]_.
!
! Orthogonal rotation of selected standardized PC time series can be used to obtain spatial or temporal patterns (e.g.,
! rotated scaled EOFs and PC time series) that are hopefully more physically meaningful by relaxing the spatial orthogonality
! of the rotated scaled EOFs, while preserving the temporal orthogonality of the rotated standardized PC time series.
! Preserving the orthogonality of the (rotated) PC time series implies that these rotated PCs describe the same amount
! of variance as the original (unrotated) PC time series, which is a very useful property, but this variance is more
! equally (or differently) distributed among the rotated PC time series [Jolliffe]_  [Richman]_.
!
! In classical orthogonal rotation PCA methods (e.g., like varimax rotation, see :ref:`comp_ortho_rot_eof_4d` for details),
! the objective is to make the rotated standardized PC time series as simple as possible to interpret by attempting to constrain
! the spatial coefficients (e.g., the scaled EOFs) associated with each of them to be large or very small in absolute magnitude, but
! avoiding as much as possible spatial loadings of moderate size [Jolliffe]_  [Jackson]_. On the other hand, here in comp_loess_rot_eof_4d,
! the focus of the orthogonal rotation is to obtain a cleaner representation of the different time scales embedded in the selected PC
! time series such that these time scales will not be mixed on the same PC time series after the orthogonal rotation [Wills_etal]_ .
!
! It can be shown that a rotation matrix, which will maximize such clean representation of the time scales of interest, can be 
! estimated as the eigenvectors of the covariance matrix between the selected standardized PC time series filtered in the
! frequency bands of interest. This type of method is called Low-Frequency Components Analysis (LFCA) [Wills_etal]_ .
!
! Here, LFCA is considered as an orthogonal rotation method of the original standardized PC time series of a (partial) EOF or PCA model.
! The ``nrot`` by ``nrot`` orthogonal matrix (where ``nrot`` is the number of rotated PCs) used to rotate the original PCs is
! computed as the eigenvectors of the covariance (e.g., symmetric positive-definite) matrix between the filtered selected standardized
! PC time series [Wills_etal]_ .
!
! In comp_loess_rot_eof_4d, the filtering of the original PC time series is performed with the help of a LOESS smoother specified
! by the values of the **-nt=**, **-itdeg=**, **--ntjump=** and **-a=** arguments. See [Cleveland]_ [Cleveland_Devlin]_  for more
! details on the LOESS smoother used here. In the LOESS procedure, the analyzed PC time series are decomposed into two terms:
!
!           **PC**\ (*t*\ ) = **T**\ (*t*\ ) + **R**\ (*t*\ )
!
! where *t* refers to a time index, the **T** term is used to quantify the trend and low-frequency variations in the PC time series,
! and, finally, the **R** term contains the residual (high-frequency) component of the PC time series. The trend is estimated through
! a sequence of applications of locally weighted regression or low-order polynomial (e.g., LOESS) to data windows whose length is chosen
! by the user (see description of the **-nt=** and **-itdeg=** arguments below for details) and the residual component is estimated by
! difference with the original PC time series. This LOESS smoother is a good candidate for extracting frequency-defined series components
! from time series. This filter is also implemented in :ref:`comp_trend_4d`, where more details are provided.
!
! This LOESS smoother is applied to the selected PC time series before their covariance matrix and its associated eigenvectors are
! computed. At the user option, the covariance matrix can be computed from the trend or residual components of the PC time series (see
! description of the **-a=** argument below for details). In the first case, the rotated standardized PC time series will be ordered
! from low- to high-frequency modes and, in the second case, they will be ordered from high- to low-frequency modes.
!
! The purpose here is exactly the same as in :ref:`comp_filt_rot_eof_4d` excepted that comp_loess_rot_eof_4d uses a LOESS smoother
! [Cleveland]_ [Cleveland_Devlin]_ instead of a windowed filtering [Iacobucci_Noullez]_ to filter the original PC time series.
!
! In summary, using as input an EOF NetCDF file produced by :ref:`comp_eof_4d`, comp_loess_rot_eof_4d 
! extracts selected PC time series from this input file (see the description of **-se=** argument below for details) and
! performs the orthogonal rotation of the selected standardized PC time series towards modes that have more energy in the
! frequency bands of interest selected with the help of the **-nt=**, **-itdeg=** and  **-a=** arguments.
! comp_loess_rot_eof_4d also reads the associated scaled EOFs from the input EOF NetCDF file and computes the regression
! spatial patterns associated with the rotated standardized PC time series (e.g., the orthogonal projection onto the orthonormal
! set formed by the new rotated standardized PC time series).
! The percentages of variance described by the new rotated standardized PC time series are also computed and stored in the output
! NetCDF file, as are the ratios of filtered variance to total variance for the rotated standardized PC time series.
!
! The selected standardized PC time series, which will be rotated, are determined on entry by the value of
! the **-se=**\  argument, which is a list of PC (or EOF) numbers.
!
! An output NetCDF dataset containing these rotated PC standardized time series, their associated
! regression spatial patterns and the explained variance and ratio statistics is created.
!
! If the NetCDF variable used in the original EOF analysis is tridimensional use :ref:`comp_loess_rot_eof_3d` instead
! of comp_loess_rot_eof_4d.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_loess_rot_eof_4d\
!         -f=eof_netcdf_file \
!         -v=eof_netcdf_variable \
!         -m=input_mesh_mask_netcdf_file \
!         -nt=trend_smoother_length        (nt) \
!         -se=selected_eofs                          (optional) \
!         -g=grid_type                               (optional : n, t, u, v, w, f) \
!         -r=resolution                              (optional : r2, r4) \
!         -b=nlon_orca, nlat_orca, nlevel_orca       (optional) \
!         -x=lon1,lon2                               (optional) \
!         -y=lat1,lat2                               (optional) \
!         -z=level1,level2                           (optional) \
!         -t=time1,time2                             (optional) \
!         -itdeg=trend_smoother_degree     (itdeg)   (optional : 0, 1, 2 ) \
!         -ntjump=trend_skipping_value     (ntjump)  (optional) \
!         -a=type_of_analysis                        (optional : trend, residual)  \
!         -o=output_rot_eof_netcdf_file              (optional) \
!         -mi=missing_value                          (optional) \
!         -double                                    (optional) \
!         -bigfile                                   (optional) \
!         -hdf5                                      (optional) \
!         -tlimited                                  (optional)
!
! By default
! ----------
!
!        *-se=*
!         all the PC time series stored in the *eof_netcdf_file*
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 2-D grid-mesh associated with
!         the input NetCDF EOF variable, *eof_netcdf_variable*, is assumed to be regular or Gaussian
!        *-r=*
!         if the input *eof_netcdf_variable* is from the NEMO ocean model (e.g., if **-g=** argument
!         is not set to ``n``) the resolution is assumed to be ``r2``
!        *-b=*
!         if **-n=** is not set to ``n``\ , the dimensions of the 3-D grid-mesh, *nlon_orca*, 
!         *nlat_orca* and *nlevel_orca* are determined from the **-r=** argument. However, you
!         may override this choice by default with the **-b=** argument
!        *-x=*
!         the whole longitude domain associated with the *eof_netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *eof_netcdf_variable*
!        *-z=*
!         the whole vertical resolution associated with the *eof_netcdf_variable*
!        *-t=*
!         the whole time period associated with the input PC time series
!        *-itdeg=*
!         the *trend_smoother_degree* is set to ``1``
!        *-ntjump=*
!         the *trend_skipping_value* is set to ``nt/10`` where ``nt`` is the value of
!         the *trend_smoother_length* argument
!        *-a=*
!         the *type_of_analysis* is set to ``trend``. This means that the trend components of the PC
!         times series are used to compute the orthogonal rotation matrix
!        *-o=*
!         the *output_rot_eof_netcdf_file* is named ``loess_rot_``\ *eof_netcdf_variable*\ ``.nc``
!        *-mi=*
!         the *missing_value* attribute for the rotated EOFs in the output NetCDF file is set to ``1.e+20``
!        *-double*
!         the results of the rotated EOF analysis are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *eof_netcdf_variable* argument specifies the NetCDF variable for which an
!      EOF analysis was originally computed by :ref:`comp_eof_4d`\ )
!      and the **-f=**\ *input_eof_netcdf_file* argument specifies that the resulting EOF patterns
!      and PC time series must be extracted from the NetCDF file, *eof_netcdf_file*\ . 
!      This NetCDF file must have exactly the same format as the files produced by :ref:`comp_eof_4d`\ .
!
!      These EOF patterns and PC time series are used to compute the orthogonal rotation matrix which will
!      be applied to the selected EOF model.
!
!   #) The **-m=**\ *input_mesh_mask_netcdf_file* argument specifies the land-sea mask
!      to apply to the *eof_netcdf_variable* (which contains the scaled eigenvectors of the EOF
!      analysis) for transforming this fourdimensional NetCDF variable as a rectangular matrix
!      before performing the rotated EOF analysis. It is assumed that this land-sea mask is exactly
!      the same as the one used in the orignal EOF analysis, whose results are stored in the input
!      NetCDF file *eof_netcdf_file*\ .
!
!   #) The **-se=**\ *selected_eofs* argument allows the user to select the standardized PC time series
!      which must be included in the orthogonal rotation. The list of selected PCs may
!      be given in two formats:
!
!      * **-se=**\ *1,3,...,nn* allows to include *pc1,pc3,...* and *pcnn* in the orthogonal rotation
!
!      * **-se=**\ ``1:4`` allows to include from *pc1* to *pc4* in the orthogonal rotation.
!
!      The two forms of the **-se=** argument may be combined and repeated any number of times,
!      but duplicate PC numbers are not allowed.
!      If the **-se=** argument is not specified, all the PC time series
!      stored in the *input_eof_netcdf_file* are used in the orthogonal rotation.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that the
!      NetCDF variable, *eof_netcdf_variable*, is from an experiment with the NEMO ocean model.
!      In this case, the duplicate points from the ORCA grid are removed 
!      before the rotated EOF analysis, as far as possible, and, in particular,
!      if the 3-D grid-mesh of the input NetCDF variable covers the whole globe. 
!
!      If **-g=** is set to ``n``, it is assumed that the 2-D grid-mesh is regular or Gaussian
!      and as such has no duplicate points.
!
!      The **-g=** argument is also used to determine the name of the NetCDF variables which
!      contain the 3-D mesh-mask in the *input_mesh_mask_netcdf_file* (e.g., the variable named *grid_type*\ mask).
!      This *input_mesh_mask_netcdf_file* may be created by :ref:`comp_clim_4d` if the 3-D grid-mesh is regular or gaussian.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` (e.g., if the input NetCDF variable,
!      *eof_netcdf_variable*, is from an experiment with the NEMO ocean model), the **-r=** argument gives the 
!      resolution used. If:
!
!      * **-r=**\ ``r2`` the NetCDF variable is from an experiment with the ORCA R2 configuration of the NEMO ocean model
!
!      * **-r=**\ ``r4`` the NetCDF variable is from an experiment with the ORCA R4 configuration of the NEMO ocean model.
!
!   #) If the NetCDF variable, *eof_netcdf_variable*, is from an experiment with the NEMO model,
!      but the resolution is not R2 or R4, the dimensions of the ORCA grid must be specified
!      explicitly with the **-b=** argument.
!
!   #) If the **-x=**\ *lon1,lon2*, **-y=**\ *lat1,lat2* and **-z=**\ *level1,level2* arguments
!      are not specified, the geographical domain used in the rotated EOF analysis is determined from the attributes
!      of the input mesh-mask NetCDF variable named *grid_type*\ mask (e.g., lon1_Eastern_limit, lon2_Western_limit,
!      lat1_Southern_limit, lat2_Northern_limit, level1_First_level and level2_Last_level), which is read from the
!      input NetCDF file *input_mesh_mask_netcdf_file*\ .
!      If these attributes are missing, the whole geographical domain associated with the *netcdf_variable*
!      is used in the EOF projection. These arguments must be set normally to the same values as used in the
!      original EOF analysis, whose results are stored in the input NetCDF file, *eof_netcdf_file*\ .
!
!      The longitude, latitude or level range must be a vector of two integers specifying
!      the first and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_4d` for transforming geographical coordinates as indices or generating an
!      appropriate mesh-mask before using comp_loess_rot_eof_4d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing, the whole time period
!      associated with the original PC time series is used to compute the orthogonal rotation matrix.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` observations.
!
!   #) The geographical shapes of the *eof_netcdf_variable* (in the *eof_netcdf_file*) and the land-sea mask
!      (in the *input_mesh_mask_netcdf_file*) must agree.
!
!   #) The **-nt=** argument specifies the length of the trend smoother, ``nt``. The value of ``nt``
!      should be an odd integer greater than or equal to ``3``. As ``nt`` increases the values of
!      the trend components of the PC time series, which are used to compute the orthogonal rotation
!      matrix (if **-a=**\ ``trend``\ ) become smoother.
!
!   #) The **-itdeg=** argument specifies the degree of the locally-fitted polynomial in trend
!      smoothing of the PC time series.  The value is ``0``, ``1`` or ``2``.
!
!   #) The **-ntjump=** argument specifies the skipping value for trend smoothing of
!      the PC time series. The trend smoother skips ahead  ``ntjump``  points and then
!      linearly interpolates in between.  The  value  of  ``ntjump``
!      should  be  a  positive  integer; if ``ntjump`` is set to ``1``, a trend
!      smooth is calculated  at all points in the PC time series. To make the
!      procedure run faster, a reasonable choice for ``ntjump`` is
!      ``10``\ % or ``20``\ % of ``nt``.
!
!   #) The **-a=** argument specifies if the residuals from the trend components or
!      the trend components are used to compute the orthogonal rotation matrix. If:
!
!      * **-a=**\ ``trend``, the trend components are used.
!
!      * **-a=**\ ``residual``, the residuals from the trends are used.
!
!      The default is **-a=**\ ``trend`` , e.g., the trend components are used to estimate
!      the orthogonal rotation matrix. More precisely, the covariance matrix between the trend components of
!      the PC time series is then computed and the eigenvectors of this symmetric positive-definite matrix, ranked
!      in decreasing order of the associated (positive) eigenvalues are calculated in order to obtain the
!      orthogonal rotation matrix to apply to the selected EOFs and PC time series.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator associated with
!      the rotated EOF NetCDF variable in the *output_rot_eof_netcdf_file*.
!      If the **-mi=** argument is not specified *missing_value* is set to ``1.e+20``.
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file, *output_rot_eof_netcdf_file*.
!
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g., ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!
!      If this argument is specified, the *output_rot_eof_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g., ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) It is assumed that the original EOF variable has no missing values excepted those associated with
!      a constant land-sea mask.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
!   #) For more details on EOF analysis, PCA and orthogonal rotation of PC time series to
!      selected-frequency components and LOESS smoothing of time series, see
!
!      * "A manual for EOF and SVD analyses of climate data",
!        by Bjornsson, H., and Venegas, S.A., McGill University, CCGCR Report No. 97-1,
!        Montral, Qubec, 52pp, 1997. https://www.jsg.utexas.edu/fu/files/EOFSVD.pdf
!
!      * "Statistical Analysis in Climate Research",
!        by von Storch, H., and Zwiers, F.W., Cambridge University press,
!        Cambridge, UK, Chapter 13, 484 pp., 2002. ISBN: `9780521012300`_
!
!      * "Principal Component Analysis",
!        by Jolliffe, I.T., Springer-Verlag, New York, USA, 487 pp., 2nd Ed, 2002. ISBN: 978-0-387-22440-4. https://www.springer.com/gp/book/9780387954424
!
!      * "A user's guide to principal components",
!        by Jackson, J.E., John Wiley and Sons, New York, USA, 592 pp., 2003. ISBN: `978-0-471-47134-9`_
!
!      * "Rotation of principal components"
!        by Richman, M.B., International Journal of Climatology, Vol. 6, 293-335, 1986.  https://doi.org/10.1002/joc.3370060305
!
!      * "Disentangling Global Warming, Multidecadal Variability, and El Nino in Pacific Temperatures",
!        by Wills, R.C., Schneider, T., Wallace, J.M., Battisti, D.S., and Hartmann, D.L.,
!        Geophysical Research Letters, Vol. 45, 2487-2496, 2018.  https://doi.org/10.1002/2017GL076327
!
!      * "Robust Locally Weighted Regression and Smoothing Scatterplots",
!        by Cleveland, W.S., Journal of the American Statistical Association, Vol. 74, 829-836, 1979. doi: `10.1080/01621459.1979.10481038`_
!
!      * "Locally Weighted Regression: An Approach to Regression Analysis by Local Fitting",
!        by Cleveland, W.S., and Devlin, S.J., Journal of the American Statistical Association, Vol. 83, 596-610, 1988. doi: `10.1080/01621459.1988.10478639`_
!
!      * "A Seasonal-Trend Decomposition Procedure Based on Loess",
!        by Cleveland, R.B.,  Cleveland, W.S., McRae, J.E., and Terpennings, I.,
!        Journal of Official Statistics, 6, 3-73, 1990. http://www.jos.nu/Articles/abstract.asp?article=613
!                                                                              
!      * "A Frequency Selective Filter for Short-Length Time Series", 
!        by Iacobucci, A., and Noullez, A., Computational Economics, Vol. 25, 75-102, 2005. https://link.springer.com/article/10.1007/s10614-005-6276-7
!
! .. _9780521012300: http://www.cambridge.org/gb/academic/subjects/earth-and-environmental-science/climatology-and-climate-change/statistical-analysis-climate-research?format=PB
! .. _978-0-471-47134-9: https://www.wiley.com/en-us/A+User%27s+Guide+to+Principal+Components-p-9780471471349
! .. _10.1080/01621459.1979.10481038 : http://dx.doi.org/10.1080/01621459.1979.10481038
! .. _10.1080/01621459.1988.10478639 : http://dx.doi.org/10.1080/01621459.1988.10478639
!
! Outputs
! -------
!
!   comp_loess_rot_eof_4d creates an output NetCDF file that contains the rotated standardized PC 
!   time series, their associated rotated EOF patterns and the variances described by the computed
!   rotated PC time series. The ratios of filtered variance to total variance for the rotated
!   standardized PC time series are also stored in the output NetCDF file.
!
!   The number of rotated PCs and EOFs, ``nrot``, stored in the output NetCDF dataset is determined by
!   the **-se=**\ *selected_eofs* argument. The number of observations in the output NetCDF dataset
!   is determined from the **-t=**\ *time1,time2* argument. The output NetCDF dataset contains the following
!   NetCDF variables (in the description below, nlev, nlat and nlon are the lengths of the vertical and spatial dimensions
!   of the input NetCDF variable *eof_netcdf_variable* and ``nrot`` is the number of selected PC time series, which have been
!   rotated as determined by the list specified with the **-se=**\ *selected_eofs* argument) :
!
!   #) pc_number\ ``(nrot)`` : the list of rotated PC time series as specified with the
!      **-se=**\ *selected_eofs* argument.
!
!   #) *eof_netcdf_variable*\ _loess_rot_eof\ ``(nrot,nlev,nlat,nlon)`` : the rotated EOF patterns.
!      These rotated EOFs  are scaled such that they give the scalar products (if **-a=**\ ``scp`` in the
!      original EOF analysis), covariances (**-a=**\ ``cov`` in the original EOF analysis) or correlations
!      (**-a=**\ ``cor`` in the original EOF analysis) between the original observed variables
!      and the rotated PC time series.
!
!      The rotated EOF patterns are packed in a fourdimensional variable whose first, second and third dimensions
!      are exactly the same as those associated with the input NetCDF variable *eof_netcdf_variable*
!      even if you restrict the geographical domain with the **-x=**, **-y=** and **-z=** arguments.
!      However, outside the selected domain, this output NetCDF variable is filled with missing values.
!
!   #) *eof_netcdf_variable*\ _loess_rot_pc\ ``(ntime,nrot)`` : the rotated standardized PC time series.
!
!      The rotated PC time series are always standardized to unit variance.
!
!   #) *eof_netcdf_variable*\ _loess_rot_std\ ``(nrot)`` : the standard-deviations of the rotated  PC time series.
!
!      The squares of these statistics are equal to the raw variance described by the rotated PC
!      time series.
!
!   #) *eof_netcdf_variable*\ _loess_rot_var\ ``(nrot)`` : the proportion  of variance explained by each rotated
!      PC time series (given between 0. and 1. with 1. corresponding to 100%).
!
!   #) *eof_netcdf_variable*\ _loess_rot_ratio\ ``(nrot)`` : the ratio of filtered variance to total variance for
!      each rotated PC time series (given between 0. and 1. with 1. corresponding to 100%).
!
! Examples
! --------
!
!   #)   For computing a rotated ``10``\ -EOF analysis towards low-frequency modesfrom a previous EOF analysis (stored
!        in the file ``eof_1m_197902_200501_votemper_oi.nc``) of a fourdimensional NetCDF variable
!        named ``votemper`` and store the results of the rotated EOF analysis in a NetCDF file named
!        ``loess_rot_eof_1m_197902_200501_votemper_oi.nc``, use the following command :
!
!        ::
!
!          $ comp_loess_rot_eof_4d \
!            -f=eof_1m_197902_200501_votemper_oi.nc  \
!            -v=votemper \
!            -se=1:10 \
!            -nt=48 \
!            -m=oi_mask.nc \
!            -o=loess_rot_eof_1m_197902_200501_votemper_oi.nc
!
!-
! ================================================================================================
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _BLAS
    use BLAS_interfaces,  only : gemm
#else
    use Statpack,         only : matmul=>matmul2
#endif
!
#ifdef _TRANSPOSE
    use Utilities,        only : transpose=>transpose2
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,    only : nf90_clobber, nf90_nowrite, nf90_int, nf90_real4, nf90_real8,                &
                          nf90_sizehint_default, nf90_max_name, nf90_max_var_dims, nf90_global,        &
                          nf90_noerr, nf90_unlimited, nf90_nofill,                                     &
                          nf90_open, nf90_strerror, nf90_get_att, nf90_get_var,                        &
                          nf90_close, nf90_create, nf90_set_fill, nf90_def_dim, nf90_def_var,          &
                          nf90_inq_attname, nf90_copy_att, nf90_put_att, nf90_enddef, nf90_put_var,    &
                          nf90_close
!
    use Statpack,         only : i1b, i4b, stnd, extd, lgl, true, false, zero, one, ten, c0_9, & 
                                 kint, kfix, kexp, tolower, case_change, string_to_val,        &
                                 allocate_error, merror, arth, comp_lfc_rot_pc
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, nlon_def2, nlat_def2, ndepth_def2,            &
                                 nlon_def4, nlat_def4, ndepth_def4, def_missing,                       &
                                 get_listarg, get_int_list, get_variable_dims, get_variable_dimcoords, &
                                 check_lon_domain, select_grid_points_3d, duplicate_orca_points_3d
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_loess_rot_eof_4d : ',     &
!
                                   cmethod(2)=(/"trend   ","residual"/),     &
!
                                   cdistance(3)=(/"dist2","dist3","ident"/), &   
!
                                   cresolution(2)=(/"r2","r4"/)
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid_mask, ncid_eof, ncid_rot,                      &
                     maskid, eofid, roteofid, pcid, stdid, varid, ratioid,             &
                     i, j, j1, j2, nlon, nlat, nlevel, ntime, nvar, nt, itdeg, ntjump, &
                     lon1, lon2, lon3, lat1, lat2, lat3, time1, time2, time3,          &
                     level1, level2, level3, distance, method, nlon_def, nlat_def,     &
                     nlevel_def, nrot, neof, output_type, output_mode, var_natts

    integer(i4b), dimension(5)                 :: coor_id2
    integer(i4b), dimension(4)                 :: startlec, countlec
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type, &
                                                  coor_natts_pc, coor_id_pc, coor_type_pc
    integer(i4b), dimension(:),    pointer     :: eof_list
!
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),     allocatable  :: std_pc, std_rot_pc, var_rot_pc, ratio_rot_pc
    real(stnd), dimension(:,:),   allocatable  :: eof, rot_eof, pc, rot_pc, rot
    real(stnd), dimension(:,:,:), allocatable  :: temp
    real(stnd)                                 :: tmp, ssqmat, missing
!
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                               :: use_tlimited, use_x, use_y, use_z, failure, comp_resid,  &
                                                  create_hdf5, missing_att
    logical(lgl), dimension(nf90_max_var_dims)  :: use_coor_variable, use_coor_variable_pc
    logical(lgl), dimension(:,:,:), allocatable :: mask
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: mask_name, eof_fic_name, eof_name, rot_name, att_name 
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names, dim_names_pc
    character(len=14)                                          :: fmt
    character                                                  :: chr, grid_type
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Perform an orthogonal rotation towards specific time scales of a (partial)'
        write (prtunit,*) ' EOF or PCA model of a fourdimensional NetCDF variable extracted from a NetCDF'
        write (prtunit,*) ' dataset using LOESS smoothing.'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_loess_rot_eof_4d -f=eof_netcdf_file'
        write (prtunit,*) '                       -v=eof_netcdf_variable (without _eof extension)'
        write (prtunit,*) '                       -m=input_mesh_mask_netcdf_file'
        write (prtunit,*) '                       -nt=trend_smoother_length'
        write (prtunit,*) '                       -se=selected_eofs                    (optional)'
        write (prtunit,*) '                       -g=grid_type                         (optional : n, t, u, v, w, f)'
        write (prtunit,*) '                       -r=resolution                        (optional : r2, r4)'
        write (prtunit,*) '                       -b=nlon_orca, nlat_orca, nlevel_orca (optional)'
        write (prtunit,*) '                       -x=lon1,lon2                         (optional)'
        write (prtunit,*) '                       -y=lat1,lat2                         (optional)'
        write (prtunit,*) '                       -z=level1,level2                     (optional)'
        write (prtunit,*) '                       -t=time1,time2                       (optional)'
        write (prtunit,*) '                       -itdeg=trend_smoother_degree         (optional : 0, 1, 2)'
        write (prtunit,*) '                       -ntjump=trend_skipping_value         (optional)'
        write (prtunit,*) '                       -a=type_of_analysis                  (optional : trend, residual)'
        write (prtunit,*) '                       -o=output_rot_eof_netcdf_file        (optional)'
        write (prtunit,*) '                       -mi=missing_value                    (optional)'
        write (prtunit,*) '                       -double                              (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '                       -bigfile                             (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '                       -hdf5                                (optional)'
#endif
        write (prtunit,*) '                       -tlimited                            (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-se= all the PC time series stored in eof_netcdf_file'
        write (prtunit,*) '-g=n  : a regular or gaussian grid is assumed'
        write (prtunit,*) '-r=r2 : if the grid_type is not n, orca2 configuration is assumed'
        write (prtunit,*) '-b= nlon_orca, nlat_orca and nlevel_orca are determined from the -r= argument'
        write (prtunit,*) '-x= the whole longitude domain associated with the eof_netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the eof_netcdf_variable'
        write (prtunit,*) '-z= the whole vertical resolution associated with the netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the eof_netcdf_variable'
        write (prtunit,*) '-itdeg=1'
        write (prtunit,*) '-ntjump=nt/10 where nt is the value of trend_smoother_length'
        write (prtunit,*) '-a=trend'
        write (prtunit,*) '-o=loess_rot_eof_netcdf_variable.nc'
        write (prtunit,*) '-mi= the missing_value is equal to 1.e+20'
        write (prtunit,*) '-double  : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile : by default, a NetCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
#ifdef _OPENMP
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<4 .or. narg>21 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    eof_fic_name = " "
    eof_name     = " "
    mask_name    = " "
    rot_name     = " "
!
    history = "comp_loess_rot_eof_4d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode = nf90_clobber
    output_type = nf90_real4
    nlat_def    = nlat_def2
    nlon_def    = nlon_def2
    nlevel_def  = ndepth_def2
!
    missing = def_missing
!
    nt      = -1
    itdeg   = 1
    ntjump  = -1
    method  = 1
!
    nrot = 0
!
    nullify( eof_list )
!
    grid_type  = "n"
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    level1    = 1
    level2    = huge( level2 )
    time1     = 1
    time2     = huge( time2 )
!
    use_x        = false
    use_y        = false
    use_z        = false
    comp_resid   = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE EOF NETCDF DATASET.
!
                eof_fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE EOF NETCDF VARIABLE IN THE DATASET.
!
                eof_name = name(4:)
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-r=')
!
!               DETERMINE THE RESOLUTION FOR THE ORCA GRID.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cresolution(1) )
                        nlat_def   = nlat_def2
                        nlon_def   = nlon_def2
                        nlevel_def = ndepth_def2
                    case ( cresolution(2) )
                        nlat_def   = nlat_def4
                        nlon_def   = nlon_def4
                        nlevel_def = ndepth_def4
                    case default
                        call merror( name_proc//"Unknown resolution option" )
                end select
!
            case ('-b=')
!
!               GET THE DIMENSIONS OF THE ORCA GRID.
!
                call get_listarg( name_proc, name(4:),                                        &
                                  "nlon_orca, nlat_orca and nlevel_orca are integer values",  &
                                  "Wrong ORCA grid definition ",                              &
                                  nlon_def, nlat_def, nlevel_def )
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-z=')
!
!               GET THE LEVEL DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "level1 and level2 are integer values", &
                                  "Wrong level definition",               &
                                  level1, level2 )
!
                use_z = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
            case ('-a=')
!
!               DETERMINE IF THE RESIDUALS ARE REQUIRED.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                    case ( cmethod(2) )
                        method     = 2
                        comp_resid = true
                    case default
                        call merror( name_proc//"Unknown analysis option" )
                end select
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE ROTATED EOF ANALYSIS.
!
                rot_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-se=')
!
!                       GET THE LIST OF PCs TO ROTATE.
!
                        if ( nrot/=0 ) deallocate( eof_list )
!
                        call get_int_list(  name_proc, name, 5_i4b, nrot , eof_list )
!
                    case ('-nt=')
!
!                       GET THE LENGTH OF THE TREND SMOOTHER.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) nt
                        end if
!
                        if ( kcode/=kint .or. nt<=2  )                           &
                        call merror( name_proc//"The -nt= argument is an integer > 2" )
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
!
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                call merror( name_proc//"The -mi= argument must be a real number" )
!
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(5:7), tolower )
!
                        select case ( name(1:7) )
!
                            case ('-itdeg=')
!
!                               GET THE DEGREE OF LOCALLY-FITTED POLYNOMIAL IN TREND SMOOTHING.
!
                                call string_to_val( name(8:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(8:), fmt ) itdeg
                                end if
!
                                if ( kcode/=kint .or. (itdeg<0 .or. itdeg>2)  )                &
                                call merror( name_proc//"The -itdeg= argument must be 0, 1 or 2." )
!
                            case default
!
                                select case ( name(1:8) )
!
                                    case ('-ntjump=')
!
!                                       GET THE TREND SKIPPING VALUE.
!
                                        call string_to_val( name(9:), kcode, fmt )
                                        if ( kcode==kint ) then
                                            read( name(9:), fmt ) ntjump
                                        end if
!
                                        if ( kcode/=kint .or. ntjump<=0  )                &
                            call merror( name_proc//"The -ntjump= argument is a stricly positive integer" )
!
                                    case default
!
                                        call case_change( name(8:), tolower )
!
                                        select case ( name(1:) )
!
                                            case ('-double')
!
!                                               STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                                output_type = nf90_real8
!
#ifdef _USE_NETCDF36
                                            case ('-bigfile')
!
!                                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                                output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                                            case ('-hdf5')
!
!                                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                                output_mode = nf90_hdf5
                                                create_hdf5 = true
#endif
!
                                            case ('-tlimited')
!
!                                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                                use_tlimited = true
!
                                            case default
!
!                                               WRONG ARGUMENT.
!
                                                call merror( name_proc//"Wrong argument" )
!
                                        end select
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( eof_fic_name==" " )                                              &
    call merror( name_proc//"Input EOF NETCDF file is missing" )
!
    if ( eof_name==" " )                                                  &
    call merror( name_proc//"Input EOF NETCDF variable is missing" )
!
    if ( mask_name==" " )                                                 &
    call merror( name_proc//"Input mesh-mask NETCDF file is missing" )
!
    if ( rot_name==" " )  rot_name="loess_rot_eof_"//trim(eof_name)//".nc"
!
!   CHECK LOESS PARAMETERS.
!
    if ( nt==-1_i4b )                                                    &
    call merror( name_proc//"The -nt= argument is missing" )
!
    if ( mod(nt,2_i4b)==0_i4b) nt = nt + 1_i4b
!
    if ( ntjump==-1_i4b ) then
        ntjump = max( 1_i4b, int(real(nt,stnd)/ten + c0_9,i4b) )
    end if
!
!   OPEN THE NETCDF DATASETS.
!
    status = nf90_open( trim(eof_fic_name), nf90_nowrite, ncid_eof )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE EOF VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    name = trim(eof_name)//"_eof"
!
!    call get_variable_dims( name_proc, trim(name), ncid_eof, eofid, dimids, dimlens )
!
    call get_variable_dimcoords(  name_proc, trim(name), ncid_eof,                         &
                                  eofid,  dimids_out=dimids, dim_names_out=dim_names,      &
                                  dimlens_out=dimlens, coor_id_out=coor_id,                &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type,      &
                                  use_coor_variable_out=use_coor_variable, natts=var_natts )
!
    if ( count(dimlens/=0)/=4 )           &
    call merror( name_proc//"Input EOF NETCDF variable is not fourdimensional" )
!
    nlon   = dimlens(1)
    nlat   = dimlens(2)
    nlevel = dimlens(3)
    neof   = dimlens(4)
!
!   GET THE PC VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!
    name = trim(eof_name)//"_pc"
!
!    call get_variable_dims( name_proc, trim(name), ncid_eof, pcid, dimids, dimlens )
!
    call get_variable_dimcoords(  name_proc, trim(name), ncid_eof,                          &
                                  pcid,  dimids_out=dimids, dim_names_out=dim_names_pc,     &
                                  dimlens_out=dimlens, coor_id_out=coor_id_pc,              &
                                  coor_natts_out=coor_natts_pc, coor_type_out=coor_type_pc, &
                                  use_coor_variable_out=use_coor_variable_pc                )
!
    if ( count(dimlens/=0)/=2 )   &
    call merror( name_proc//"Input PC time series NETCDF variable is not bidimensional" )
!
    if ( dimlens(1)/=neof )           &
    call merror( name_proc//"Shape disagreement between input PC and EOF NETCDF variables" )
!
    ntime = dimlens(2)
!
    dim_names(5)         = dim_names_pc(2)
    coor_id(5)           = coor_id_pc(2)
    coor_natts(5)        = coor_natts_pc(2)
    coor_type(5)         = coor_type_pc(2)
    use_coor_variable(5) = use_coor_variable_pc(2)
!
!   GET THE VARIABLE ID OF THE STANDARD_DEVIATIONS OF THE PC TIME SERIES FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!
    name = trim(eof_name)//"_sing"
!
    call get_variable_dims( name_proc, trim(name), ncid_eof, stdid, dimids, dimlens )
!
    if ( count(dimlens/=0)/=1 )   &
    call merror( name_proc//"Input Singular Values NETCDF variable is not unidimensional" )
!
    if ( dimlens(1)/=neof )           &
    call merror( name_proc//"Incorrect size of the Singular Values NETCDF variable" )
!
!   GET THE MASK ID FROM ITS NAME.
!   VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
    chr = merge( 't', grid_type, grid_type=='w' )
!
    call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
    if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat .or. dimlens(3)/=nlevel )    &
    call merror( name_proc//"Shape disagreement with the mask" )
!
!   GET THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN FROM THE MESH-MASK DATASET.
!
    if ( .not.use_x ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
        if ( status==nf90_noerr ) lon1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
        if ( status==nf90_noerr ) lon2 = i
!
    end if
!
    if ( .not.use_y ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
        if ( status==nf90_noerr ) lat1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
        if ( status==nf90_noerr ) lat2 = i
!
    end if
!
    if ( .not.use_z ) then
!
        status = nf90_get_att( ncid_mask, maskid, "level1_First_level", i )
        if ( status==nf90_noerr ) level1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "level2_Last_level", i )
        if ( status==nf90_noerr ) level2 = i
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!   CHECK THE GEOGRAPHICAL DOMAIN.
!   REMOVE DUPLICATE POINTS FROM THE ORCA_R2 GRID.
!
    if ( grid_type/='n' ) then
!
        if ( nlon>nlon_def )                                                     &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (longitude)" )
!
        if ( nlat>nlat_def )                                                     &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (latitude)" )
!
        if ( nlevel>nlevel_def )                                                 &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (level)" )
!
!       A TRICK TO DISABLE THE ELIMINATION AND DUPLICATION IN THE HIGH LATITUDES OF THE ORCA GRID
!       IF THE RESOLUTION IS NOT R2 OR R4.
!
        if ( ( nlat_def/=nlat_def2 .or. nlon_def/=nlon_def2 ) .and. ( nlat_def/=nlat_def4 .or. nlon_def/=nlon_def4 ) )    &
        nlat_def = nlat_def + 1
!
    end if
!
    call check_lon_domain( grid_type, nlon, nlon_def, lon1, lon2, lon3 )
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE LEVEL DEFINITION.
!
    level1 = max( level1, 1 )
    level2 = min( level2, nlevel )
    level3 = level2 - level1 + 1
    if ( level3<=0  ) call merror( name_proc//"Wrong vertical definition" )
!
!   CHECK THE PERIOD DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, ntime )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( temp(nlon,nlat,nlevel), mask(nlon,nlat,nlevel), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE MASK.
!
    status = nf90_get_var( ncid_mask, maskid, temp )     
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DETERMINE THE GRID POINTS USED IN THE EOF ANALYSIS FROM THE MASK.
!
    mask(:nlon,:nlat,:nlevel) = temp(:nlon,:nlat,:nlevel)==one
!
!   SELECT THE DOMAIN.
!
    call select_grid_points_3d( grid_type, lon1, lon2, lat1, lat2, level1, level2, &
                                nlon_def, nlat_def, mask, j1, j2  )
!
    if ( lon1<0 ) lon3 = nlon
!
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   CHECK THE EOF LIST.
!
    if ( nrot==0 ) then
!
        nrot = neof
!
!       ALLOCATE VARIABLE FOR THE EOF LIST IF NOT DONE BEFORE.
!
        allocate( eof_list(nrot), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        eof_list(1:nrot) = arth( 1_i4b, 1_i4b, nrot )
!
    else
!
        if ( any( eof_list(1:nrot)>neof ) .or. any( eof_list(1:nrot)<1 ) )           &
        call merror( name_proc//"Wrong -se= argument list for the selection of EOFs" )
!
    end if
!
    if ( size(eof_list)<=1 )          &
    call merror( name_proc//"At least two EOFs must be selected for orthogonal rotation" )
!
!   ALLOCATE WORK ARRAYS FOR THE ORIGINAL AND ROTATED SPATIAL LOADINGS AND PC TIME SERIES
!   AND STATISTICS ASSOCIATED WITH THE NEW ROTATED EOF MODEL.
!
    allocate( eof(nvar,nrot), pc(time3,nrot), rot(nrot,nrot),   &
              rot_eof(nvar,nrot), rot_pc(time3,nrot),           &
              std_pc(nrot), std_rot_pc(nrot), var_rot_pc(nrot), &
              ratio_rot_pc(nrot), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE ORIGINAL SPATIAL LOADINGS OF THE EOF ANALYSIS.
!
    startlec(1:3)  = (/ j1, lat1, level1 /)
    countlec(1:4)  = (/ lon3, lat3, level3, 1_i4b /)
!
    do i = 1, nrot
!
        startlec(4) = eof_list(i)
!
        status = nf90_get_var( ncid_eof, eofid, temp(:lon3,:lat3,:level3), &
                               start=startlec, count=countlec      )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        eof(:nvar,i) = pack( temp(:lon3,:lat3,:level3), mask(j1:j2,lat1:lat2,level1:level2) )
!
    end do
!
!   FIND THE MISSING VALUE ATTRIBUTE ASSOCIATED WITH THE EOF NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid_eof, eofid, "missing_value", tmp )
    missing_att = status==nf90_noerr
!
    if ( .not.missing_att ) then
        status = nf90_get_att( ncid_eof, eofid, "_FillValue", tmp )
        missing_att = status==nf90_noerr
    end if
!
    if ( missing_att ) then
        if ( any( eof(:nvar,:nrot)==tmp ) )    &
        call merror( name_proc//"Missing values in the EOF spatial patterns" )
    end if
!
    status = nf90_get_att( ncid_eof, nf90_global, "distance", name )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( trim(name)==trim(cdistance(1)) ) then
        distance = 1
    else
        distance = 2
    end if
!
!   GET THE ORIGINAL PC TIME SERIES OF THE EOF ANALYSIS.
!
    startlec(1:2)  = (/ 1_i4b, time1 /)
    countlec(1:2)  = (/ 1_i4b, time3 /)
!
    do i = 1, nrot
!
        startlec(1) = eof_list(i)
!
        status = nf90_get_var( ncid_eof, pcid, pc(:time3,i), &
                               start=startlec(1:2), count=countlec(1:2)      )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   GET THE SINGULAR VALUES VARIABLE ID FROM ITS NAME.
!
    status = nf90_get_var( ncid_eof, stdid, std_pc(:nrot) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE SQUARE ROOT OF THE TOTAL VARIANCE.
!
    status = nf90_get_att( ncid_eof, nf90_global, "sqrt_ssqmat", ssqmat )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   PERFORM THE ORTHOGONAL ROTATION OF THE SELECTED STANDARDIZED PC TIME SERIES.
!
    call comp_lfc_rot_pc( pc(:time3,:nrot), std_pc(:nrot), nt, rot_pc(:time3,:nrot), rot(:nrot,:nrot), &
                          std_rot_pc(:nrot), failure, itdeg=itdeg, ntjump=ntjump, residual=comp_resid, &
                          smooth=ratio_rot_pc  )
!
!   NOW ROTATE THE SELECTED EOFs.
!
#ifdef _BLAS
    call gemm('N', 'N', nvar, nrot, nrot, one, eof, nvar, rot, nrot, zero, rot_eof, nvar )
#else
    rot_eof(:nvar,:nrot) = matmul( eof(:nvar,:nrot), rot(:nrot,:nrot) )
#endif
!
!   COMPUTE EXPLAINED VARIANCES BY THE ROTATED PC TIME SERIES.
!
    var_rot_pc(:nrot) = (std_rot_pc(:nrot)/ssqmat)**2
!
!   CREATE THE NEW NETCDF DATASET FOR THE ROTATED EOF MODEL.
!
    if ( create_hdf5 ) then
        status = nf90_create( rot_name, cmode=output_mode, ncid=ncid_rot )
    else
        i = nf90_sizehint_default
        status = nf90_create( rot_name, cmode=output_mode, ncid=ncid_rot, chunksize=i )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_rot, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = nlon
    dimlens(2)   = nlat
    dimlens(3)   = nlevel
    dimlens(4)   = nrot
!
    if ( use_tlimited ) then
        dimlens(5)   = time3
    else
        dimlens(5)   = nf90_unlimited
    end if
!
    dim_names(4) = "pc_number"
!
    coor_type(4) = nf90_int
!
    do i = 1, 5
!
        status = nf90_def_dim( ncid_rot, name=trim(dim_names(i)), len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
!   DEFINE COORDINATE VARIABLES.
!
    do i = 1, 5
!
        status = nf90_def_var( ncid_rot, trim(dim_names(i)), coor_type(i),                &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEFINE VARIABLES RELATED TO THE ROTATED EOF MODEL.
!
    status = nf90_def_var( ncid_rot, trim(eof_name)//"_loess_rot_eof", output_type,       &
                           dimids=dimids(1:4), varid=roteofid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_rot, trim(eof_name)//"_loess_rot_pc", output_type,        &
                           dimids=dimids(4:5), varid=pcid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_rot, trim(eof_name)//"_loess_rot_std", nf90_real8,        &
                           dimids=dimids(4), varid=stdid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_rot, trim(eof_name)//"_loess_rot_var", output_type,       &
                           dimids=dimids(4), varid=varid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_rot, trim(eof_name)//"_loess_rot_ratio", output_type,       &
                           dimids=dimids(4), varid=ratioid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
!   ASSIGN ATTRIBUTES FOR COORDINATE VARIABLES.
!
    do j = 1, 5
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid_eof, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid_eof, coor_id(j), att_name, ncid_rot, coor_id2(j) )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
!   ASSIGN ATTRIBUTES FOR VARIABLES OF THE ROTATED EOF MODEL.
!
    if ( .not.use_coor_variable(4) ) then
!
        status = nf90_put_att( ncid_rot, coor_id2(4), "long_name", "Empirical Orthogonal Function numbers" )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_rot, coor_id2(4), "short_name", "EOF numbers" )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    do i = 1, var_natts
!
        status = nf90_inq_attname( ncid_eof, eofid, i, att_name )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        select case ( att_name )
!
            case ( '_FillValue', 'missing_value', 'long_name', 'short_name'  )
!
!
            case default
!
                status = nf90_copy_att(  ncid_eof, eofid, att_name, ncid_rot, roteofid )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end select
!
    end do
!
    status = nf90_put_att( ncid_rot, roteofid, "long_name", "Rotated Empirical Orthogonal Functions" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, roteofid, "short_name", "Rotated EOFs" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_rot, roteofid, "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_rot, roteofid, "_FillValue", real( missing, dp ) )
!
    else
!
        status = nf90_put_att( ncid_rot, roteofid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_rot, roteofid, "_FillValue", real( missing, sp ) )
!
    end if
!
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, pcid, "long_name", "Rotated Standardized Principal Components" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, pcid, "short_name", "Rotated PCs" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, stdid,     &
             "long_name", "Standard-Deviations of Rotated Principal Components" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, stdid, "short_name", "STDs of Rotated PCs" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, varid, "long_name", "Explained Variances" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, varid, "short_name", "ExpVars" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, ratioid, "long_name", "Filter Ratios" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, ratioid, "short_name", "FiltRatios" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   SET GLOBAL ATTRIBUTES.
!
    name = " LOESS rotated EOF analysis"
!
    status = nf90_put_att( ncid_rot, nf90_global, "title", trim(eof_name)//trim(name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "trend_smoother_length", nt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "trend_smoother_degree", itdeg )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "trend_skipping_value", ntjump )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "method", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "distance", cdistance(distance) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "sqrt_ssqmat", ssqmat )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_rot )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 5
!
        if ( j==4 ) cycle
!
        select case ( j )
            case ( 1 )
                j1 = 1
                j2 = nlon
            case ( 2 )
                j1 = 1
                j2 = nlat
            case ( 3 )
                j1 = 1
                j2 = nlevel
            case ( 5 )
                j1 = time1
                j2 = time3
        end select
!
        if ( use_coor_variable(j) ) then
!
            allocate( record(j2), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
            status = nf90_get_var( ncid_eof, coor_id(j), record, start=(/ j1 /), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_rot, coor_id2(j), record )
!
            deallocate( record )
!
        else
!
            status = nf90_put_var( ncid_rot, coor_id2(j), arth(j1,1_i4b,j2 ) )
!
        end if
!
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   WRITE THE LIST OF ROTATED PC TIME SERIES.
!
    status = nf90_put_var( ncid_rot, coor_id2(4), eof_list(1:nrot) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE STANDARD-DEVIATIONS OF THE ROTATED PC TIME SERIES.
!
    status = nf90_put_var( ncid_rot, stdid, std_rot_pc(:nrot) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE EXPLAINED VARIANCES BY THE ROTATED PC TIME SERIES.
!
    status = nf90_put_var( ncid_rot, varid, var_rot_pc(:nrot) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE FILTER RATIOS OF THE ROTATED PC TIME SERIES.
!
    status = nf90_put_var( ncid_rot, ratioid, ratio_rot_pc(:nrot) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE THE ROTATED EOFS.
!
    startlec(1:3) = 1
!
    do i = 1, nrot
!
        temp(:nlon,:nlat,:nlevel) = unpack( rot_eof(:nvar,i), mask, missing )
!
        if ( grid_type/='n' .and. nlon==nlon_def ) then
!
            if ( lat2>=nlat_def-1 .and. nlat==nlat_def ) then
!
!               ADD DUPLICATE POINTS TO THE ORCA GRID IF THE WHOLE GLOBE IS USED.
!
                call duplicate_orca_points_3d( grid_type, mask, temp )
!
            else
!
!               ADD DUPLICATE LONGITUDES LINES TO THE ORCA GRID IF THE WHOLE LONGITUDE
!               DOMAIN IS USED.
!
                temp(1,:nlat,:nlevel) = temp(nlon-1,:nlat,:nlevel)
                temp(nlon,:nlat,:nlevel) = temp(2,:nlat,:nlevel)
!
            end if
!
        end if
!
        startlec(4) = i
!
        status = nf90_put_var( ncid_rot, roteofid, temp, start=startlec )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( temp, mask, eof_list, eof, rot_eof, pc, rot, std_pc, &
                std_rot_pc, var_rot_pc, ratio_rot_pc )
!
!   ALLOCATE WORK ARRAY.
!
    allocate( temp(nrot,time3,1),  stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   WRITE ROTATED STANDARDIZED PC TIME SERIES.
!
    temp(:nrot,:time3,1) = transpose( rot_pc(:time3,:nrot) )
!
    status = nf90_put_var( ncid_rot, pcid, temp(:nrot,:time3,1) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( rot_pc, temp )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid_eof )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_mask )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_rot )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_loess_rot_eof_4d
! ====================================
!
end program comp_loess_rot_eof_4d
