program comp_ortho_rot_eof_3d
!! #############################################################################
!!
!! Copyright 2024 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_ortho_rot_eof_3d:
!
! =====================
! comp_ortho_rot_eof_3d
! =====================
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 29/04/2024
!
! Purpose
! =======
!
! Performs an orthogonal rotation of a (partial) Empirical Orthogonal Function (EOF)
! or Principal Component Analysis (PCA) model (e.g., a factor loading matrix and the
! associated standardized amplitude time series) computed from a previous EOF or PCA
! analysis of a tridimensional NetCDF variable. The factor loading matrix (e.g.,
! the associated scaled EOF patterns corresponding to the selected rotated PC time series)
! are also rotated and the new rotated EOF or PCA model is stored in a NetCDF file.
!
! Spatial EOF patterns (e.g., scaled EOFs) and Principal Component (PC) time series derived from
! EOF analysis or PCA are constrained to be both spatially and temporally orthogonal to one another
! so that they maximize the variance they described over the entire analysis domain and selected
! time period [Jolliffe]_  [vonStorch_Zwiers]_ . However, the spatial orthogonality of the EOFs
! is a strong and undesirable constraint imposed on the EOFs and as a result, EOFs are subject
! to problems such as domain dependence and inaccurate representation of physical relationships embedded
! in the data [Jolliffe]_  [Richman]_ .
!
! Orthogonal rotation of selected standardized PC time series can be used to obtain spatial and time patterns (e.g.,
! rotated scaled EOFs and PC time series) that are hopefully more physically meaningful by relaxing the spatial orthogonality
! of the rotated scaled EOFs, while preserving the temporal orthogonality of the rotated standardized PC time series.
! Preserving the orthogonality of the (rotated) PC time series implies that these rotated PCs describe the same amount
! of variance as the original (unrotated) PC time series, which is a very useful statistical property, but this variance
! is more equally (or differently) distributed among the rotated PC time series [Jolliffe]_  [Richman]_ .
!
! More precisely, the orthogonal rotation in comp_ortho_rot_eof_3d has the objective to make the rotated standardized
! PC time series as simple as possible to interpret by attempting to constrain the spatial coefficients (or loadings)
! associated with each of them to be large or very small in absolute magnitude, but avoiding as much as possible spatial
! loadings of moderate size [Jolliffe]_  [Jackson]_ .
!
! The orthogonal rotation matrix, which will perform this linear transformation is obtained by maximizing a generalized
! orthomax criterion, including quartimax, varimax and equamax rotation methods [Jackson]_ at the user option (see description
! of the **-w=**\  argument below for more details about the selection of the rotation method which will be used).
! This generalized orthomax criterion is described in details in [Clarkson_Jennrich]_   [Jackson]_ and amounts to obtain
! rotated standardized PC time series which are more spatially localized by relaxing the spatial orthogonality of the
! original EOF (or PCA) modes, while preserving the orthogonality of the (rotated) standardized PC time series as described above.
!
! The **-w=**\ *rotation_factor*  argument determines the kind of orthogonal rotation to be computed [Clarkson_Jennrich]_  [Jackson]_ and
! may be set as follows :
!
!   * **-w=**\ ``0.`` is the quartimax method [Jackson]_ , which attempts to get each variable (e.g., grid-point) to load highly
!     on only one (or a few) EOFs or factors.
!
!   * **-w=**\ ``1.`` is the varimax method [Jackson]_ , which attempts to load highly a relatively low number of variables
!     on each EOF or factor. Varimax is the most widely used method of orthogonal rotation.
!
!   * **-w=**\ ``nrot``\ /\ ``2.``\ , where ``nrot`` is the number of rotated PCs, is the equamax method  [Jackson]_ , which is
!     a compromise of the above two.
!
! More generally, the value of the **-w=**\  argument can be any positive real number, but best values lie in the closed
! interval [``0.``\ , ``5.``\  * ``nrot``\ ]. Generally, the larger the value is, the more equal is the dispersion of the variance
! accounted for across the rotated EOFs or factors.
!
! As an illustration, by default, comp_ortho_rot_eof_3d uses the varimax rotation method, which simplifies the spatial
! structure of the rotated (scaled) EOFs by maximizing the sum of the variances of the squared columns loadings of the
! selected rotated (scaled) EOFs [Jolliffe]_  [Jackson]_  [vonStorch_Zwiers]_ . As a result, the rotated EOFs tend to have
! coefficients biased towards either large or zero values. Hence these rotated EOFs should be easier to interpret than
! their unrotated counterparts.
!
! Note that, other methods of rotation exist to obtain more physically based EOFs and PC time series. For example,
! the orthogonal rotation matrix can be designed to maximize the energy in a selected frequency band for the rotated PC
! time series, see :ref:`comp_filt_rot_eof_3d` for more details.
!
! Using as input an EOF NetCDF file produced by :ref:`comp_eof_3d` or :ref:`comp_eof_miss_3d`,
! comp_ortho_rot_eof_3d performs an orthogonal rotation of selected standardized PC time series readed
! from this input EOF NetCDF file and computes the regression spatial patterns associated with these
! rotated standardized PC time series (e.g., the orthogonal projection onto the orthonormal set formed
! by the new rotated standardized PC time series).
! The raw variances and percentages of variance described by the new rotated standardized PC time series
! are also computed.
!
! The selected standardized PC time series, which will be rotated is determined on entry by the value of
! the **-se=**\  argument, which is a list of PC (or EOF) numbers.
!
! Optionally, the rotated PC time series can be estimated with the help of a metric such that the results are
! weighted by the surface (or volume) associated with each cell in the grid associated with the input
! tridimensional NetCDF variable (which contains the original scaled EOFs) so that equal areas (or volumes)
! carry equal weights in the results of the rotated EOF analysis (see the **-d=**\  argument description).
!
! An output NetCDF dataset containing these rotated PC standardized time series, their associated
! regression spatial patterns and the explained variance statistics is created.
!
! If the NetCDF variable used in the original EOF analysis is fourdimensional use :ref:`comp_ortho_rot_eof_4d` instead
! of comp_ortho_rot_eof_3d.
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_ortho_rot_eof_3d\
!         -f=eof_netcdf_file \
!         -v=eof_netcdf_variable \
!         -m=input_mesh_mask_netcdf_file \
!         -se=selected_eofs                 (optional) \
!         -g=grid_type                      (optional : n, t, u, v, w, f) \
!         -r=resolution                     (optional : r2, r4) \
!         -b=nlon_orca, nlat_orca           (optional) \
!         -x=lon1,lon2                      (optional) \
!         -y=lat1,lat2                      (optional) \
!         -t=time1,time2                    (optional) \
!         -d=distance                       (optional : dist2, ident) \
!         -w=rotation_factor                (optional :  0. > +inf) \
!         -o=output_rot_eof_netcdf_file     (optional) \
!         -mi=missing_value                 (optional) \
!         -knorm                            (optional) \
!         -double                           (optional) \
!         -bigfile                          (optional) \
!         -hdf5                             (optional) \
!         -tlimited                         (optional)
!
! By default
! ----------
!
!        *-se=*
!         all the PC time series stored in the *eof_netcdf_file*
!        *-g=*
!         the *grid_type* is set to ``n`` which means that the 2-D grid-mesh associated with
!         the input NetCDF EOF variable, *eof_netcdf_variable*, is assumed to be regular or Gaussian
!        *-r=*
!         if the input *eof_netcdf_variable* is from the NEMO ocean model (e.g., if **-g=** argument
!         is not set to ``n``) the resolution is assumed to be ``r2``
!        *-b=*
!         if **-g=** is not set to ``n``\ , the dimensions of the 2-D grid-mesh, *nlon_orca* and
!         *nlat_orca*, are determined from the **-r=** argument. However, you may override
!         this default choice with the **-b=** argument
!        *-x=*
!         the whole longitude domain associated with the *eof_netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *eof_netcdf_variable*
!        *-t=*
!         the whole time period associated with the input PC time series
!        *-d=*
!         the *distance* is set to the distance used in the original EOF analysis.
!        *-w=*
!         the *rotation_factor* is set to ``0.``, which means that a varimax orthogonal rotation is performed.
!        *-o=*
!         the *output_rot_eof_netcdf_file* is named ``ortho_rot_``\ *eof_netcdf_variable*\ ``.nc``
!        *-mi=*
!         the *missing_value* attribute for the rotated EOFs in the output NetCDF file is set to ``1.e+20``
!        *-knorm*
!         the orthogonal rotation of the standardized PC time series is performed without Kaiser's row normalization.
!         If **-knorm** is activated, Kaiser's row normalization is performed before the orthogonal rotation of
!         the PC time series and the spatial loadings (e.g., the rotated EOFs) associated with the rotated PC time series
!         are rescaled after the orthogonal rotation
!        *-double*
!         the results of the rotated EOF analysis are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. If **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *eof_netcdf_variable* argument specifies the NetCDF variable for which an
!      EOF analysis was originally computed by :ref:`comp_eof_3d` (or :ref:`comp_eof_miss_3d`\ )
!      and the **-f=**\ *input_eof_netcdf_file* argument specifies that the resulting EOF patterns
!      and PC time series must be extracted from the NetCDF file, *eof_netcdf_file*\ . 
!      This NetCDF file must have exactly the same format as the files produced
!      by :ref:`comp_eof_3d` or :ref:`comp_eof_miss_3d`\ .
!
!      These EOF patterns and PC time series are used to compute the orthogonal rotation matrix which will
!      be applied to the selected EOF model.
!
!   #) The **-m=**\ *input_mesh_mask_netcdf_file* argument specifies the land-sea mask
!      to apply to the *eof_netcdf_variable* (which contains the scaled eigenvectors of the EOF
!      analysis) for transforming this tridimensional NetCDF variable as a rectangular matrix
!      before performing the rotated EOF analysis. It is assumed that this land-sea mask is exactly
!      the same as the one used in the orignal EOF analysis, whose results are stored in the input
!      NetCDF file *eof_netcdf_file*\ .
!
!      The scale factors associated with the 2-D grid-mesh of the input NetCDF variable (needed if
!      **-d=**\ ``dist2``\  is specified when calling the procedure) are also read from the
!      *input_mesh_mask_netcdf_file*\ .
!
!   #) The **-se=**\ *selected_eofs* argument allows the user to select the standardized PC time series
!      which must be included in the orthogonal rotation. The list of selected PCs may
!      be given in two formats:
!
!      * **-se=**\ *1,3,...,nn* allows to include *pc1,pc3,...* and *pcnn* in the orthogonal rotation
!
!      * **-se=**\ ``1:4`` allows to include from *pc1* to *pc4* in the orthogonal rotation.
!
!      The two forms of the **-se=** argument may be combined and repeated any number of times.
!      Duplicate PC numbers are not allowed.
!      If the **-se=** argument is not specified, all the PC time series
!      stored in the *input_eof_netcdf_file* are used in the orthogonal rotation.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that the
!      NetCDF variable, *eof_netcdf_variable*, is from an experiment with the NEMO ocean model.
!      In this case, the duplicate points from the ORCA grid are removed 
!      before the rotated EOF analysis, as far as possible, and, in particular,
!      if the 2-D grid-mesh of the input NetCDF variable covers the whole globe. 
!
!      If **-g=** is set to ``n``, it is assumed that the 2-D grid-mesh is regular or Gaussian
!      and as such has no duplicate points.
!
!      The **-g=** argument is also used to determine the name of the NetCDF variables which
!      contain the 2-D mesh-mask and the scale factors in the *input_mesh_mask_netcdf_file* (e.g.,
!      these variables are named *grid_type*\ mask, e1\ *grid_type* and e2\ *grid_type*\ ,
!      respectively). This *input_mesh_mask_netcdf_file* may be created by :ref:`comp_clim_3d`
!      if the 2-D grid-mesh is regular or gaussian.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` (e.g., if the input NetCDF variable,
!      *eof_netcdf_variable*, is from an experiment with the NEMO ocean model), the **-r=** argument gives the 
!      resolution used. If:
!
!      * **-r=**\ ``r2`` the NetCDF variable is from an experiment with the ORCA R2 configuration of the NEMO ocean model
!
!      * **-r=**\ ``r4`` the NetCDF variable is from an experiment with the ORCA R4 configuration of the NEMO ocean model.
!
!   #) If the NetCDF variable, *eof_netcdf_variable*, is from an experiment with the NEMO model,
!      but the resolution is not R2 or R4, the dimensions of the ORCA grid
!      must be specified explicitly with the **-b=** argument.
!
!   #) If the **-x=**\ *lon1,lon2* and **-y=**\ *lat1,lat2* arguments are not specified, the geographical
!      domain used in the rotated EOF analysis is determined from the attributes of the input mesh mask
!      NetCDF variable named *grid_type*\ mask (e.g,. lon1_Eastern_limit, lon2_Western_limit, lat1_Southern_limit
!      and lat2_Northern_limit), which is read from the input NetCDF file *input_mesh_mask_netcdf_file*\ .
!      If these attributes are missing, the whole geographical domain associated with the *netcdf_variable*
!      is used in the EOF projection. These arguments must be set normally to the same values as used in the
!      original EOF analysis, whose results are stored in the input NetCDF file, *eof_netcdf_file*\ .
!
!      The longitude or latitude range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_3d` for transforming geographical coordinates as indices or generating an
!      appropriate mesh-mask before using comp_ortho_rot_eof_3d.
!
!   #) If the **-t=**\ *time1,time2* argument is missing, the whole time period
!      associated with the original PC time series is used in the orthogonal rotation.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ . Note that the
!      output NetCDF file will have ``ntime`` = *time2* - *time1* + ``1`` observations.
!
!   #) The geographical shapes of the *eof_netcdf_variable* (in the *eof_netcdf_file*), the land-sea mask
!      and scale factors (in the *input_mesh_mask_netcdf_file*) must agree.
!
!   #) The **-d=** argument specifies the metric and scalar product used in the rotated EOF analysis:
!
!      * **-d=**\ ``dist2`` means that the orthogonal rotation is done with the diagonal distance
!        associated with the horizontal 2-D grid-mesh (e.g., each grid-point is weighted accordingly to
!        the surface associated with it)
!
!      * **-d=**\ ``ident`` means that the orthogonal rotation is done with the identity metric, e.g., the usual
!        Euclidean distance and scalar product are used in the rotated EOF analysis.
!
!      By default, the **-d=** argument is set to the distance used in the original EOF analysis.
!
!   #) The **-w=**\ *rotation_factor* argument specifies the exact form of the function criterion, which must be
!      maximized over the set of ``nrot`` by ``nrot`` orthogonal rotation matrices (where ``nrot`` is the number
!      of selected PCs as specified with the **-se=** argument). *rotation_factor* is a positive real scalar and
!      may be set as follows:
!
!      * **-w=**\ ``0.`` is the quartimax method, which attempts to get each variable (e.g., grid-point) to load highly
!        on only one (or a few) EOFs or factors.
!
!      * **-w=**\ ``1.`` is the varimax method, which attempts to load highly a relatively low number of variables
!        on each EOF or factor. Varimax is the most widely used method of orthogonal rotation.
!
!      * **-w=**\ ``nrot``/``2.``\ , where ``nrot`` is the number of rotated PCs, is the equamax method, which is a compromise
!        of the above two.
!
!      More generally, *rotation_factor* can be any positive real number, but best values lie in the closed
!      interval [``0.``\ , ``5.``\  * ``nrot``\ ]. Generally, the larger the value of *rotation_factor* is, the more
!      equal is the dispersion of the variance accounted for across the rotated EOFs or factors.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator associated with
!      the NetCDF variables in the *output_rot_eof_netcdf_file*.
!
!      If the **-mi=** argument is not specified *missing_value* is set to ``1.e+20``.
!
!   #) The **-knorm** argument specifies that orthogonal rotation of the original PC time series
!      must be performed  using Kaiser's row normalization of the associated spatial loadings (e.g., the
!      original scaled EOFs) and the computed spatial loadings (e.g., the rotated EOFs) associated with the
!      rotated PC time series are rescaled after the orthogonal rotation.
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file, *output_rot_eof_netcdf_file*.
!
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g., ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!
!      If this argument is specified, the *output_rot_eof_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g., ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) It is assumed that the data has no missing values excepted those associated with
!      a constant land-sea mask.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
!   #) For more details on EOF analysis, PCA and orthogonal rotation methods, see
!
!      * "A manual for EOF and SVD analyses of climate data",
!        by Bjornsson, H., and Venegas, S.A., McGill University, CCGCR Report No. 97-1,
!        Montral, Qubec, 52pp, 1997. https://www.jsg.utexas.edu/fu/files/EOFSVD.pdf
!
!      * "Statistical Analysis in Climate Research",
!        by von Storch, H., and Zwiers, F.W., Cambridge University press,
!        Cambridge, UK, Chapter 13, 484 pp., 2002. ISBN: `9780521012300`_
!
!      * "Principal Component Analysis",
!        by Jolliffe, I.T., Springer-Verlag, New York, USA, 487 pp., 2nd Ed, 2002. ISBN: 978-0-387-22440-4. https://www.springer.com/gp/book/9780387954424
!
!      * "A user's guide to principal components",
!        by Jackson, J.E., John Wiley and Sons, New York, USA, 592 pp., 2003. ISBN: `978-0-471-47134-9`_
!
!      * "Quartic rotation criteria and algorithms"
!        by Clarkson, D.B., and Jennrich, R.I., Psychometrika, Vol. 53, 251-259, 1988. https://link.springer.com/article/10.1007/BF02294136
!
!      * "Rotation of principal components"
!        by Richman, M.B., International Journal of Climatology, Vol. 6, 293-335, 1986.  https://doi.org/10.1002/joc.3370060305
!
! .. _9780521012300: http://www.cambridge.org/gb/academic/subjects/earth-and-environmental-science/climatology-and-climate-change/statistical-analysis-climate-research?format=PB
! .. _978-0-471-47134-9: https://www.wiley.com/en-us/A+User%27s+Guide+to+Principal+Components-p-9780471471349
!
! Outputs
! -------
!
!   comp_ortho_rot_eof_3d creates an output NetCDF file that contains the rotated standardized PC 
!   time series, their associated rotated EOF patterns and the variances described by the computed
!   rotated PC time series.
!
!   The number of rotated PCs and EOFs, ``nrot``, stored in the output NetCDF dataset is determined by
!   the **-se=**\ *selected_eofs* argument. The number of observations in the output NetCDF dataset
!   is determined from the **-t=**\ *time1,time2* argument. The output NetCDF dataset contains the following
!   NetCDF variables (in the description below, nlat and nlon are the lengths of the spatial dimensions of the input
!   NetCDF variable *eof_netcdf_variable* and ``nrot`` is the number of selected PC time series, which have been
!   rotated as determined by the list specified with the **-se=**\ *selected_eofs* argument) :
!
!   #) pc_number\ ``(nrot)`` : the list of rotated PC time series as specified with the
!      **-se=**\ *selected_eofs* argument.
!
!   #) *eof_netcdf_variable*\ _ortho_rot_eof\ ``(nrot,nlat,nlon)`` : the rotated EOF patterns.
!      These rotated EOFs  are scaled such that they give the scalar products (if **-a=**\ ``scp`` in the
!      original EOF analysis), covariances (**-a=**\ ``cov`` in the original EOF analysis) or correlations
!      (**-a=**\ ``cor`` in the original EOF analysis)  between the original observed variables
!      and the rotated PC time series.
!
!      The rotated EOF patterns are packed in a tridimensional variable whose first and second dimensions
!      are exactly the same as those associated with the input NetCDF variable *eof_netcdf_variable*
!      even if you restrict the geographical domain with the **-x=** and **-y=** arguments. However,
!      outside the selected domain, this output NetCDF variable is filled with missing values.
!
!   #) *eof_netcdf_variable*\ _ortho_rot_pc\ ``(ntime,nrot)`` : the rotated standardized PC time series.
!
!      The rotated PC time series are always standardized to unit variance.
!
!   #) *eof_netcdf_variable*\ _ortho_rot_std\ ``(nrot)`` : the standard-deviations of the rotated  PC time series.
!
!      The squares of these statistics are equal to the raw variance described by the rotated PC
!      time series.
!
!   #) *eof_netcdf_variable*\ _ortho_rot_var\ ``(nrot)`` : the proportion of variance explained by each rotated
!      PC time series (given between 0. and 1. with 1. corresponding to 100%).
!
! Examples
! --------
!
!   #)   For computing a varimax rotated ``10``\ -EOF analysis from a previous EOF analysis (stored
!        in the file ``eof_HadISST1_1m_197902_200501_sst_oi.nc``) of a tridimensional NetCDF variable
!        named ``sst`` and store the results of the rotated EOF analysis in a NetCDF file named
!        ``rot_eof_HadISST1_1m_197902_200501_sst_oi.nc``, use the following command :
!
!        ::
!
!          $ comp_ortho_rot_eof_3d \
!            -f=eof_HadISST1_1m_197902_200501_sst_oi.nc  \
!            -v=sst \
!            -se=1:10 \
!            -m=HadISST1_mask.nc \
!            -o=rot_eof_HadISST1_1m_197902_200501_sst_oi.nc
!
!-
! ================================================================================================
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _BLAS
    use BLAS_interfaces,  only : gemm
#else
    use Statpack,         only : matmul=>matmul2
#endif
!
#ifdef _TRANSPOSE
    use Utilities,        only : transpose=>transpose2
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,    only : nf90_hdf5
#endif
!
    use netcdf,    only : nf90_clobber, nf90_nowrite, nf90_int, nf90_real4, nf90_real8,                &
                          nf90_sizehint_default, nf90_max_name, nf90_max_var_dims, nf90_global,        &
                          nf90_noerr, nf90_unlimited, nf90_nofill,                                     &
                          nf90_open, nf90_strerror, nf90_get_att, nf90_get_var,                        &
                          nf90_close, nf90_create, nf90_set_fill, nf90_def_dim, nf90_def_var,          &
                          nf90_inq_attname, nf90_copy_att, nf90_put_att, nf90_enddef, nf90_put_var,    &
                          nf90_close
!
    use Statpack,         only : i1b, i4b, stnd, extd, lgl, true, false, zero, one,       & 
                                 kfix, kexp, tolower, case_change, string_to_val,         &
                                 allocate_error, merror, arth, comp_ortho_rot_eof
!
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, nlon_def2, nlat_def2, nlon_def4, nlat_def4, def_missing, &
                                 get_listarg, get_int_list, get_variable_dims, get_variable_dimcoords,            &
                                 check_lon_domain, select_grid_points_2d, duplicate_orca_points_2d
!
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
!
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_ortho_rot_eof_3d : ',  &
!
                                   cdistance(2)=(/"dist2","ident"/),      &   
!
                                   cresolution(2)=(/"r2","r4"/)
!
!   USER-DEFINED CONVERGENCE CRITERION FOR THE ROTATION.
!
    real(stnd), parameter :: delta=0.00001_stnd
!
!   maxiter CONTROLS THE MAXIMUM NUMBER OF ALLOWED ITERATIONS FOR ROTATION.
!
    integer(i4b), parameter :: maxiter=60
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid_mask, ncid_eof, ncid_rot,             &
                     maskid, e1id, e2id, eofid, roteofid, pcid, stdid, varid, &
                     i, j, j1, j2, nlon, nlat, nt, nvar, var_natts,           &
                     lon1, lon2, lon3, lat1, lat2, lat3, time1, time2, time3, &
                     distance, nlon_def, nlat_def, nrot, neof, output_type,   &
                     output_mode

    integer(i4b), dimension(4)                 :: coor_id2
    integer(i4b), dimension(3)                 :: startlec, countlec
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type, &
                                                  coor_natts_pc, coor_id_pc, coor_type_pc
    integer(i1b)                               :: kcode
    integer(i4b), dimension(:),    pointer     :: eof_list
!
    real(stnd), dimension(:),     allocatable  :: dat, scal, std_rot_pc, var_rot_pc
    real(stnd), dimension(:,:),   allocatable  :: temp, eof, rot_eof, pc, rot_pc, rot
    real(stnd)                                 :: tmp, ssqmat, w, missing
!
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                               :: use_tlimited, use_x, use_y, knorm, failure,  &
                                                  create_hdf5, missing_att
    logical(lgl), dimension(nf90_max_var_dims) :: use_coor_variable, use_coor_variable_pc
    logical(lgl), dimension(:,:), allocatable  :: mask
!
    character(len=nchar)                       :: name, history
    character(len=nf90_max_name)               :: mask_name, eof_fic_name, eof_name, rot_name, att_name 
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names, dim_names_pc
    character(len=14)                                          :: fmt
    character                                                  :: chr, grid_type
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    if ( narg==0 ) then
!
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Perform an orthogonal rotation of a (partial) EOF or PCA model extracted from'
        write (prtunit,*) ' a NetCDF dataset and computed from a tridimensional NetCDF variable, using a'
        write (prtunit,*) ' generalized orthomax criterion.'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_ortho_rot_eof_3d  -f=eof_netcdf_file'
        write (prtunit,*) '                        -v=eof_netcdf_variable (without _eof extension)'
        write (prtunit,*) '                        -m=input_mesh_mask_netcdf_file'
        write (prtunit,*) '                        -se=selected_eofs                (optional)'
        write (prtunit,*) '                        -g=grid_type                     (optional : n, t, u, v, w, f)'
        write (prtunit,*) '                        -r=resolution                    (optional : r2, r4)'
        write (prtunit,*) '                        -b=nlon_orca, nlat_orca          (optional)'
        write (prtunit,*) '                        -x=lon1,lon2                     (optional)'
        write (prtunit,*) '                        -y=lat1,lat2                     (optional)'
        write (prtunit,*) '                        -t=time1,time2                   (optional)'
        write (prtunit,*) '                        -d=distance                      (optional : dist2, ident)'
        write (prtunit,*) '                        -w=rotation_factor               (optional : 0. > +inf)'
        write (prtunit,*) '                        -o=output_rot_eof_netcdf_file    (optional)'
        write (prtunit,*) '                        -mi=missing_value                (optional)'
        write (prtunit,*) '                        -knorm                           (optional)'
        write (prtunit,*) '                        -double                          (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '                        -bigfile                         (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '                        -hdf5                            (optional)'
#endif
        write (prtunit,*) '                        -tlimited                        (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-se= all the PC time series stored in eof_netcdf_file'
        write (prtunit,*) '-g=n  : a regular or gaussian grid is assumed'
        write (prtunit,*) '-r=r2 : if the grid_type is not n, orca2 configuration is assumed'
        write (prtunit,*) '-b= nlon_orca and nlat_orca are determined from the -r= argument'
        write (prtunit,*) '-x= the whole longitude domain associated with the eof_netcdf_variable'
        write (prtunit,*) '-y= the whole latitude domain associated with the eof_netcdf_variable'
        write (prtunit,*) '-t= the whole time period  associated with the eof_netcdf_variable'
        write (prtunit,*) '-d= the distance used in the original EOF analysis'
        write (prtunit,*) '-w=0. a varimax orthogonal rotation is performed'
        write (prtunit,*) '-o=ortho_rot_eof_netcdf_variable.nc'
        write (prtunit,*) '-mi= the missing_value is equal to 1.e+20'
        write (prtunit,*) '-knorm   : by default, Kaiser row normalization is not used in the rotation'
        write (prtunit,*) '-double  : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile : by default, a NetCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5     : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
#ifdef _OPENMP
        write (prtunit,*) 'This procedure is parallelized with OpenMP'
#endif
        write (prtunit,*)
!
        stop
!
    else if ( narg<3 .or. narg>19 ) then
!
        call merror( name_proc//"Wrong number of line arguments" )
!
    end if
!
    eof_fic_name = " "
    eof_name     = " "
    mask_name    = " "
    rot_name     = " "
!
    history = "comp_ortho_rot_eof_3d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode = nf90_clobber
    output_type = nf90_real4
    nlat_def    = nlat_def2
    nlon_def    = nlon_def2
    distance    = 0
!
    missing = def_missing
!
    w           = zero
    nrot        = 0
!
    grid_type  = "n"
!
    nullify( eof_list )
!
    lon1      = 1
    lon2      = huge( lon2 )
    lat1      = 1
    lat2      = huge( lat2 )
    time1     = 1
    time2     = huge( time2 )
!
    use_x        = false
    use_y        = false
    knorm        = false
    create_hdf5  = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE EOF NETCDF DATASET.
!
                eof_fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE EOF NETCDF VARIABLE IN THE DATASET.
!
                eof_name = name(4:)
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-r=')
!
!               DETERMINE THE RESOLUTION FOR THE ORCA GRID.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cresolution(1) )
                        nlat_def  = nlat_def2
                        nlon_def  = nlon_def2
                    case ( cresolution(2) )
                        nlat_def  = nlat_def4
                        nlon_def  = nlon_def4
                    case default
                        call merror( name_proc//"Unknown resolution option" )
                end select
!
            case ('-b=')
!
!               GET THE DIMENSIONS OF THE ORCA GRID.
!
                call get_listarg( name_proc, name(4:),                           &
                                  "nlon_orca and nlat_orca are integer values",  &
                                  "Wrong ORCA grid definition ",                 &
                                  nlon_def, nlat_def )
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-t=')
!
!               GET THE PERIOD DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
            case ('-d=')
!
!               DETERMINE THE DISTANCE AND SCALAR PRODUCT TO USE IN THE ORTHOGONAL ROTATION.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cdistance(1) )
                        distance = 1
                    case ( cdistance(2) )
                        distance = 2
                    case default
                        call merror( name_proc//"Unknown distance option" )
                end select
!
            case ('-w=')
!
!               DETERMINE THE FORM OF THE ORTHOGONAL ROTATION.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode==kfix .or. kcode==kexp ) then
                    read( name(4:), fmt ) w
                end if
                if ( (kcode/=kfix .and. kcode/=kexp) .or. w<zero )  &
      call merror( name_proc//"The -w= argument must be a positive real number" )
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE ROTATED EOF ANALYSIS.
!
                rot_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-se=')
!
!                       GET THE LIST OF EOFs TO ROTATE.
!
                        if ( nrot/=0 ) deallocate( eof_list )
!
                        call get_int_list(  name_proc, name, 5_i4b, nrot , eof_list )
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
!
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                call merror( name_proc//"The -mi= argument must be a real number" )
!
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:) )
!
                            case ('-knorm')
!
!                               USE KAISER ROW NORMALIZATION IN THE ROTATION.
!
                                knorm = true
!
                            case ('-double')
!
!                               STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                output_type = nf90_real8
!
#ifdef _USE_NETCDF36
                            case ('-bigfile')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                            case ('-hdf5')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_hdf5
                                create_hdf5 = true
#endif
!
                            case ('-tlimited')
!
!                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                use_tlimited = true
!
                            case default
!
!                               WRONG ARGUMENT.
!
                                call merror( name_proc//"Wrong argument" )
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( eof_fic_name==" " )                                              &
    call merror( name_proc//"Input EOF NETCDF file is missing" )
!
    if ( eof_name==" " )                                                  &
    call merror( name_proc//"Input EOF NETCDF variable is missing" )
!
    if ( mask_name==" " )                                                 &
    call merror( name_proc//"Input mesh-mask NETCDF file is missing" )
!
    if ( rot_name==" " )  rot_name="ortho_rot_eof_"//trim(eof_name)//".nc"
!
!   OPEN THE NETCDF DATASETS.
!
    status = nf90_open( trim(eof_fic_name), nf90_nowrite, ncid_eof )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
    if ( status/=nf90_noerr )                                            &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE EOF VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!   GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!   FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!   FOR THE COORDINATE VARIABLES.
!
    name = trim(eof_name)//"_eof"
!
!    call get_variable_dims( name_proc, trim(name), ncid_eof, eofid, dimids, dimlens )
!
    call get_variable_dimcoords(  name_proc, trim(name), ncid_eof,                         &
                                  eofid,  dimids_out=dimids, dim_names_out=dim_names,      &
                                  dimlens_out=dimlens, coor_id_out=coor_id,                &
                                  coor_natts_out=coor_natts, coor_type_out=coor_type,      &
                                  use_coor_variable_out=use_coor_variable, natts=var_natts )
!
    if ( count(dimlens/=0)/=3 )           &
    call merror( name_proc//"Input EOF NETCDF variable is not tridimensional" )
!
    nlon = dimlens(1)
    nlat = dimlens(2)
    neof = dimlens(3)
!
!   GET THE PC VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!
    name = trim(eof_name)//"_pc"
!
!    call get_variable_dims( name_proc, trim(name), ncid_eof, pcid, dimids, dimlens )
!
    call get_variable_dimcoords(  name_proc, trim(name), ncid_eof,                          &
                                  pcid,  dimids_out=dimids, dim_names_out=dim_names_pc,     &
                                  dimlens_out=dimlens, coor_id_out=coor_id_pc,              &
                                  coor_natts_out=coor_natts_pc, coor_type_out=coor_type_pc, &
                                  use_coor_variable_out=use_coor_variable_pc                )
!
    if ( count(dimlens/=0)/=2 )   &
    call merror( name_proc//"Input PC time series NETCDF variable is not bidimensional" )
!
    if ( dimlens(1)/=neof )           &
    call merror( name_proc//"Shape disagreement between input PC and EOF NETCDF variables" )
!
    nt = dimlens(2)
!
    dim_names(4)         = dim_names_pc(2)
    coor_id(4)           = coor_id_pc(2)
    coor_natts(4)        = coor_natts_pc(2)
    coor_type(4)         = coor_type_pc(2)
    use_coor_variable(4) = use_coor_variable_pc(2)
!
!   GET THE MASK ID FROM ITS NAME.
!   VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
    chr = merge( 't', grid_type, grid_type=='w' )
!
    call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
    if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
    call merror( name_proc//"Shape disagreement with the mask" )
!
!   GET THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN FROM THE MESH-MASK DATASET.
!
    if ( .not.use_x ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
        if ( status==nf90_noerr ) lon1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
        if ( status==nf90_noerr ) lon2 = i
!
    end if
!
    if ( .not.use_y ) then
!
        status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
        if ( status==nf90_noerr ) lat1 = i
!
        status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
        if ( status==nf90_noerr ) lat2 = i
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!   CHECK THE GEOGRAPHICAL DOMAIN.
!   REMOVE DUPLICATE POINTS FROM THE ORCA_R2 GRID.
!
    if ( grid_type/='n' ) then
!
        if ( nlon>nlon_def )                                                     &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (longitude)" )
!
        if ( nlat>nlat_def )                                                     &
        call merror( name_proc//"Input NETCDF variable is not on the ORCA GRID (latitude)" )
!
!       A TRICK TO DISABLE THE ELIMINATION AND DUPLICATION IN THE HIGH LATITUDES OF THE ORCA GRID
!       IF THE RESOLUTION IS NOT R2 OR R4.
!
        if ( ( nlat_def/=nlat_def2 .or. nlon_def/=nlon_def2 ) .and. ( nlat_def/=nlat_def4 .or. nlon_def/=nlon_def4 ) )    &
        nlat_def = nlat_def + 1
!
    end if
!
    call check_lon_domain( grid_type, nlon, nlon_def, lon1, lon2, lon3 )
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE PERIOD DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
    if ( time3<=0  ) call merror( name_proc//"Wrong time definition" )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( temp(nlon,nlat), mask(nlon,nlat), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE MASK.
!
    status = nf90_get_var( ncid_mask, maskid, temp )     
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   DETERMINE THE GRID POINTS USED IN THE EOF ANALYSIS FROM THE MASK.
!
    mask(:nlon,:nlat) = temp(:nlon,:nlat)==one
!
!   SELECT THE DOMAIN.
!
    call select_grid_points_2d( grid_type, lon1, lon2, lat1, lat2,  &
                                nlon_def, nlat_def, mask, j1, j2  )
!
    if ( lon1<0 ) lon3 = nlon
!
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   CHECK THE EOF LIST.
!
    if ( nrot==0 ) then
!
        nrot = neof
!
!       ALLOCATE VARIABLE FOR THE EOF LIST IF NOT DONE BEFORE.
!
        allocate( eof_list(nrot), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
        eof_list(1:nrot) = arth( 1_i4b, 1_i4b, nrot )
!
    else
!
        if ( any( eof_list(1:nrot)>neof ) .or. any( eof_list(1:nrot)<1 ) )           &
        call merror( name_proc//"Wrong -se= argument list for the selection of EOFs" )
!
    end if
!
    if ( size(eof_list)<=1 )          &
    call merror( name_proc//"At least two EOFs must be selected for orthogonal rotation" )
!
!   ALLOCATE WORK ARRAYS FOR THE ORIGINAL AND ROTATED SPATIAL LOADINGS AND PC TIME SERIES
!   AND STATISTICS ASSOCIATED WITH THE ROTATED NEW EOF MODEL.
!
    allocate( eof(nvar,nrot), pc(time3,nrot), rot(nrot,nrot), &
              rot_eof(nvar,nrot), rot_pc(time3,nrot),         &
              std_rot_pc(nrot), var_rot_pc(nrot),             &
              stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE ORIGINAL SPATIAL LOADINGS OF THE EOF ANALYSIS.
!
    startlec(1:2)  = (/ j1, lat1 /)
    countlec(1:3)  = (/ lon3, lat3, 1_i4b /)
!
    do i = 1, nrot
!
        startlec(3) = eof_list(i)
!
        status = nf90_get_var( ncid_eof, eofid, temp(:lon3,:lat3), &
                               start=startlec, count=countlec      )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        eof(:nvar,i) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
    end do
!
!   FIND THE MISSING VALUE ATTRIBUTE ASSOCIATED WITH THE EOF NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid_eof, eofid, "missing_value", tmp )
    missing_att = status==nf90_noerr
!
    if ( .not.missing_att ) then
        status = nf90_get_att( ncid_eof, eofid, "_FillValue", tmp )
        missing_att = status==nf90_noerr
    end if
!
    if ( missing_att ) then
        if ( any( eof(:nvar,:nrot)==tmp ) )    &
        call merror( name_proc//"Missing values in the EOF spatial patterns" )
    end if
!
    if ( distance==0 ) then
!
        status = nf90_get_att( ncid_eof, nf90_global, "distance", name )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        if ( trim(name)==trim(cdistance(1)) ) then
            distance = 1
        else
            distance = 2
        end if
!
    end if
!
!   COMPUTE THE DISTANCE AND THE SCALAR PRODUCT TO BE USED IN THE ROTATION
!   IF REQUIRED.
!
    if ( distance==1 .and. .not.knorm ) then
!
!       THE DISTANCE IS ASSOCIATED WITH THE (HORIZONTAL) GRID.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( scal(nvar), dat(nvar), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET THE SCALE FACTORS IDs FROM THEIR NAMES.
!       VERIFY THE DIMENSION LENGTHS FOR THE SCALE FACTOR VARIABLES.
!
        call get_variable_dims( name_proc, "e1"//chr, ncid_mask,    &
                                e1id, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
        call merror( name_proc//"Shape disagreement with the first scale factor" )
!
        call get_variable_dims( name_proc, "e2"//chr, ncid_mask,    &
                                e2id, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )               &
        call merror( name_proc//"Shape disagreement with the second scale factor" )
!
!       GET THE SCALE FACTORS.
!
        status = nf90_get_var( ncid_mask, e1id, temp(:lon3,:lat3), start=startlec(1:2), &  
                               count=countlec(1:2)  )     
        if ( status/=nf90_noerr )                                                     &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
        dat(:nvar) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
        status = nf90_get_var( ncid_mask, e2id, temp(:lon3,:lat3), start=startlec(1:2), &
                               count=countlec(1:2)  )     
        if ( status/=nf90_noerr )                                                     &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        scal(:nvar) = dat(:nvar)*pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
!       COMPUTE THE DIAGONAL METRIC ASSOCIATED WITH THE HORIZONTAL GRID.
!
        scal(:nvar) = sqrt( scal(:nvar)*(one/maxval(scal(:nvar))) )
!
!       SCALE THE ORIGINAL SPATIAL LOADINGS BEFORE ROTATION IF REQUIRED.
!
        do i = 1, nrot
!
            eof(:nvar,i) = eof(:nvar,i)*scal(:nvar)
!
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( dat )
!
    end if
!
!   GET THE ORIGINAL PC TIME SERIES OF THE EOF ANALYSIS.
!
    startlec(1:2)  = (/ 1_i4b, time1 /)
    countlec(1:2)  = (/ 1_i4b, time3 /)
!
    do i = 1, nrot
!
        startlec(1) = eof_list(i)
!
        status = nf90_get_var( ncid_eof, pcid, pc(:time3,i), &
                               start=startlec(1:2), count=countlec(1:2)      )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   GET THE SINGULAR VALUES VARIABLE ID FROM ITS NAME.
!
!    name = trim(eof_name)//"_sing"
!
!    status = nf90_inq_varid( ncid_eof, trim(name), stdid )
!    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!    status = nf90_get_var( ncid_eof, stdid, std_pc(:neof) )
!    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE EXPLAINED VARIANCES VARIABLE ID FROM ITS NAME.
!
!    name = trim(eof_name)//"_var"
!
!    status = nf90_inq_varid( ncid_eof, trim(name), varid )
!    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!    status = nf90_get_var( ncid_eof, varid, var_pc(:neof) )
!    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE SQUARE ROOT OF THE TOTAL VARIANCE.
!
    status = nf90_get_att( ncid_eof, nf90_global, "sqrt_ssqmat", ssqmat )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   PERFORM THE ORTHOGONAL ROTATION OF THE SELECTED EOF MODEL.
!
    call comp_ortho_rot_eof( eof(:nvar,:nrot), rot_eof(:nvar,:nrot), rot(:nrot,:nrot), &
                             std_rot_pc(:nrot), failure, knorm=knorm, maxiter=maxiter, &
                             w=w, delta=delta )
!
!   NOW ROTATE THE SELECTED PC TIME SERIES.
!
#ifdef _BLAS
    call gemm('N', 'N', time3, nrot, nrot, one, pc, time3, rot, nrot, zero, rot_pc, time3 )
#else
    rot_pc(:time3,:nrot) = matmul( pc(:time3,:nrot), rot(:nrot,:nrot) )
#endif
!
    if ( distance==1 .and. .not.knorm ) then
!
!       UNSCALE THE ROTATED SPATIAL LOADINGS IF REQUIRED.
!
        do i = 1, nrot
!
            rot_eof(:nvar,i) = rot_eof(:nvar,i)/scal(:nvar)
!
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( scal )
!
    end if
!
!   COMPUTE EXPLAINED VARIANCES BY THE ROTATED PC TIME SERIES.
!
    var_rot_pc(:nrot) = (std_rot_pc(:nrot)/ssqmat)**2
!
!   CREATE THE NEW NETCDF DATASET FOR THE ROTATED EOF MODEL.
!
    if ( create_hdf5 ) then
        status = nf90_create( rot_name, cmode=output_mode, ncid=ncid_rot )
    else
        i = nf90_sizehint_default
        status = nf90_create( rot_name, cmode=output_mode, ncid=ncid_rot, chunksize=i )
    end if
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_rot, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = nlon
    dimlens(2)   = nlat
    dimlens(3)   = nrot
!
    if ( use_tlimited ) then
        dimlens(4)   = time3
    else
        dimlens(4)   = nf90_unlimited
    end if
!
    dim_names(3) = "pc_number"
!
    coor_type(3) = nf90_int
!
    do i = 1, 4
!
        status = nf90_def_dim( ncid_rot, name=trim(dim_names(i)), len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
!   DEFINE COORDINATE VARIABLES.
!
    do i = 1, 4
!
        status = nf90_def_var( ncid_rot, trim(dim_names(i)), coor_type(i),                &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEFINE VARIABLES RELATED TO THE ROTATED EOF MODEL.
!
    status = nf90_def_var( ncid_rot, trim(eof_name)//"_ortho_rot_eof", output_type,       &
                           dimids=dimids(1:3), varid=roteofid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_rot, trim(eof_name)//"_ortho_rot_pc", output_type,        &
                           dimids=dimids(3:4), varid=pcid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_rot, trim(eof_name)//"_ortho_rot_std", nf90_real8,        &
                           dimids=dimids(3), varid=stdid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_rot, trim(eof_name)//"_ortho_rot_var", output_type,       &
                           dimids=dimids(3), varid=varid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
!   ASSIGN ATTRIBUTES FOR COORDINATE VARIABLES.
!
    do j = 1, 4
!
        if ( .not.use_coor_variable(j) ) cycle
!
        do i = 1, coor_natts(j)
!
            status = nf90_inq_attname( ncid_eof, coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid_eof, coor_id(j), att_name, ncid_rot, coor_id2(j) )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end do
!
!   ASSIGN ATTRIBUTES FOR VARIABLES OF THE ROTATED EOF MODEL.
!
    if ( .not.use_coor_variable(3) ) then
!
        status = nf90_put_att( ncid_rot, coor_id2(3), "long_name", "Empirical Orthogonal Function numbers" )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_rot, coor_id2(3), "short_name", "EOF numbers" )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    do i = 1, var_natts
!
        status = nf90_inq_attname( ncid_eof, eofid, i, att_name )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        select case ( att_name )
!
            case ( '_FillValue', 'missing_value', 'long_name', 'short_name'  )
!
!
            case default
!
                status = nf90_copy_att(  ncid_eof, eofid, att_name, ncid_rot, roteofid )
                if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end select
!
    end do
!
    status = nf90_put_att( ncid_rot, roteofid, "long_name", "Rotated Empirical Orthogonal Functions" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, roteofid, "short_name", "Rotated EOFs" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_rot, roteofid, "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_rot, roteofid, "_FillValue", real( missing, dp ) )
!
    else
!
        status = nf90_put_att( ncid_rot, roteofid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_rot, roteofid, "_FillValue", real( missing, sp ) )
!
    end if
!
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, pcid, "long_name", "Rotated Standardized Principal Components" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, pcid, "short_name", "Rotated PCs" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, stdid,     &
             "long_name", "Standard-Deviations of Rotated Principal Components" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, stdid, "short_name", "STDs of Rotated PCs" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, varid, "long_name", "Explained Variances" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, varid, "short_name", "ExpVars" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   SET GLOBAL ATTRIBUTES.
!
    name = " Rotated EOF analysis"
!
    status = nf90_put_att( ncid_rot, nf90_global, "title", trim(eof_name)//trim(name) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "distance", cdistance(distance) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( knorm ) then
        i = 1
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_rot, nf90_global, "Kaiser_row_normalization", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "W_rotation_factor", w )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "sqrt_ssqmat", ssqmat )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_rot, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_rot )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW  NETCDF DATASET.
!
    do j = 1, 4 
!
        if ( j==3 ) cycle
!
        select case ( j )
            case ( 1 )
                j1 = 1
                j2 = nlon
            case ( 2 )
                j1 = 1
                j2 = nlat
            case ( 4 )
                j1 = time1
                j2 = time3
        end select
!
        if ( use_coor_variable(j) ) then
!
            allocate( record(j2), stat=status )
            if ( status/=0 ) call merror( name_proc//allocate_error )
!
            status = nf90_get_var( ncid_eof, coor_id(j), record, start=(/ j1 /), count=(/ j2 /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_rot, coor_id2(j), record )
!
            deallocate( record )
!
        else
!
            status = nf90_put_var( ncid_rot, coor_id2(j), arth(j1,1_i4b,j2 ) )
!
        end if
!
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   WRITE THE LIST OF ROTATED PC TIME SERIES.
!
    status = nf90_put_var( ncid_rot, coor_id2(3), eof_list(1:nrot) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE STANDARD-DEVIATIONS OF THE ROTATED PC TIME SERIES.
!
    status = nf90_put_var( ncid_rot, stdid, std_rot_pc(:nrot) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE EXPLAINED VARIANCES BY THE ROTATED PC TIME SERIES.
!
    status = nf90_put_var( ncid_rot, varid, var_rot_pc(:nrot) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE THE ROTATED EOFS.
!
    startlec(1:2) = 1
!
    do i = 1, nrot
!
        temp(:nlon,:nlat) = unpack( rot_eof(:nvar,i), mask, missing )
!
        if ( grid_type/='n' .and. nlon==nlon_def ) then
!
            if ( lat2>=nlat_def-1 .and. nlat==nlat_def ) then
!
!               ADD DUPLICATE POINTS TO THE ORCA GRID IF THE WHOLE GLOBE IS USED.
!
                call duplicate_orca_points_2d( grid_type, mask, temp )
!
            else
!
!               ADD DUPLICATE LONGITUDES LINES TO THE ORCA GRID IF THE WHOLE LONGITUDE
!               DOMAIN IS USED.
!
                temp(1,:nlat) = temp(nlon-1,:nlat)
                temp(nlon,:nlat) = temp(2,:nlat)
!
            end if
!
        end if
!
        startlec(3) = i
!
        status = nf90_put_var( ncid_rot, roteofid, temp, start=startlec )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( temp, mask, eof_list, eof, rot_eof, pc, rot, std_rot_pc, var_rot_pc )
!
!   ALLOCATE WORK ARRAY.
!
    allocate( temp(nrot,time3),  stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   WRITE ROTATED STANDARDIZED PC TIME SERIES.
!
    temp(:nrot,:time3) = transpose( rot_pc(:time3,:nrot) )
!
    status = nf90_put_var( ncid_rot, pcid, temp(:nrot,:time3) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( rot_pc, temp )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid_eof )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_mask )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_rot )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_ortho_rot_eof_3d
! ====================================
!
end program comp_ortho_rot_eof_3d
