program comp_spectrum_diff_3d
!! #############################################################################
!!
!! Copyright 2024 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_spectrum_diff_3d:
!
! =====================
! comp_spectrum_diff_3d
! =====================
!
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 30/05/2024
!
! Purpose
! =======
!
! Test if two multi-channel power spectra, computed from two real multi-channel time series with
! the help of :ref:`comp_cross_spectrum_3d` or :ref:`comp_cross_spectrum2_3d`, are significantly different or have a different
! shape using various nonparametric statistical tests based on the average of the squares of the (Log) spectral ratios of the
! two estimated multi-channel power spectra (e.g., a CHI2 statistic), the minimum and maximum values of these spectral ratios or
! the differences between the two (normalized) cumulative periodograms derived from the two estimated multi-channel power spectra  [Diggle]_ 
! [Priestley]_  [Jenkins_Watts]_  [Diggle_Fisher]_  [Coates_Diggle]_  [Potscher_Reschenhofera]_  [Potscher_Reschenhoferb]_ .
!
! The critical probabilities associated with these different spectral statistics are estimated or approximated under
! the assumptions (e.g., the null hpothesis) that the two "true" underlying multi-channel power spectra are the same or have
! the same shape depending on the selected test statistic.
!
! See below for more details on the different test statistics available in comp_spectrum_diff_3d and their associated null hypothesis,
! e.g., same underlying  "true" power spectrum or same shape of the two "true" power spectra.
!
! The two input multi-channel power spectra are read from two tridimensional variables (as specified by the **-nv=** and **-dv=** arguments) 
! readed from one or two NetCDF datasets (as specified by the **-nf=** and **-df=** arguments).
!
! It is assumed that the two estimated multi-channel power spectra have been computed by the Welch's method  [Welch]_ by calls to :ref:`comp_cross_spectrum_3d`
! or :ref:`comp_cross_spectrum2_3d` with the **-conflim** argument. This optional argument is required for computing the DOFs of the Power Spectral Density (PSD) estimates,
! which are used by comp_spectrum_diff_3d, excepted when **-a=**\ ``ks_perm`` or ``chi2_perm`` is specified in the call to comp_spectrum_diff_3d. Note also that the methods
! used to estimate the two multi-channel power spectra must be identical and that the two multi-channel power spectra must not have been smoothed in the frequency domain.
!
! The multi-channel test statistics and their associated critical probabilities are stored in an output NetCDF dataset.
!
! If you want to get a pointwise tolerance interval of the given (Log) spectral ratios for two estimated multi-channel power spectra, use :ref:`comp_spectrum_ratio_3d`
! instead of comp_spectrum_diff_3d.
!
! For additional details on these spectral computations, see  [Welch]_  [Diggle]_  [Priestley]_  [Jenkins_Watts]_  [Coates_Diggle]_   [Diggle_Fisher]_  [Potscher_Reschenhoferb]_ .
!                                                                              
!                                                                              
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_spectrum_diff_3d \
!              -nf=numerator_netcdf_file \
!              -nv=numerator_netcdf_variable \
!              -dv=denominator_netcdf_variable \
!              -df=denominator_netcdf_file       (optional) \
!              -m=input_mesh_mask_netcdf_file    (optional) \
!              -g=grid_type                      (optional : n, t, u, v, w, f) \
!              -x=lon1,lon2                      (optional) \
!              -y=lat1,lat2                      (optional) \
!              -np=numerator_period              (optional) \
!              -dp=denominator_period            (optional) \
!              -t=freq1,freq2                    (optional) \
!              -a=type_of_statistic              (optional : minmax, range, chi2, chi2_perm, ks_perm) \
!              -nb=number_of_shuffles            (optional) \
!              -o=output_netcdf_file             (optional) \
!              -mi=missing_value                 (optional) \
!              -ksnonorm                         (optional) \
!              -double                           (optional) \
!              -bigfile                          (optional) \
!              -hdf5                             (optional) \
!              -tlimited                         (optional)
!
! By default
! ----------
!
!        *-df=*
!         the *denominator_netcdf_file* is the same as the *numerator_netcdf_file*
!        *-m=*
!         an *input_mesh_mask_netcdf_file* is not used
!        *-g=*
!         the *grid_type* is set to ``n``, which means that the 2-D grid-mesh associated with the input
!         *numerator_netcdf_variable* and *denominator_netcdf_variable* is assumed to be regular or Gaussian
!        *-x=*
!         the whole longitude domain associated with the *numerator_netcdf_variable* and *denominator_netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *numerator_netcdf_variable* and *denominator_netcdf_variable*
!        *-np=*
!         the *numerator_period* argument is not used
!        *-dp=*
!         the *denominator_period* argument is not used
!        *-t=*
!         the whole set of frequencies included in the two input NetCDF files, but excluding the zero and Nyquist frequencies
!        *-a=*
!         the type_of_statistic is set to ``chi2``, e.g., a CHI2 statistic is used as the spectral test statistic
!        *-nb=*
!         the number_of_shuffles is set to  ``999``, meaning that ``999`` random interchanges are performed if a randomization test is
!         requested (e.g., if  **-a=**\ ``chi2_perm`` or  **-a=**\ ``ks_perm``). For other values of the **-a=** argument, the **-nb=**
!         argument has no effect
!        *-o=*
!         *output_netcdf_file* name is set to ``spectrum_diff_``\ *numerator_netcdf_variable*\ ``.``\ *denominator_netcdf_variable*\ ``.nc``
!         where *numerator_netcdf_variable* and *denominator_netcdf_variable* are specified with the **-nv=** and **-dv=** arguments.
!        *-mi=*
!         the *missing_value* for the output NetCDF variables is set to ``1.e+20``
!        *-ksnonorm*
!         the Kolmogorov-Smirnov statistic is computed from the normalized cumulative periodograms if **-a=**\ ``ks_perm`` is specified.
!         However, if **-ksnonorm** is also specified, the cumulative periodograms are not normalized before computing the Kolmogorov-Smirnov statistic.
!         For other values of the **-a=** argument, the **-ksnonorm** argument has no effect
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. However, if **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created . However, if **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension (e.g., the frequency axis) is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the frequency dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-nf=**\ *numerator_netcdf_file* argument specifies the NetCDF file containing the multi-channel PSD
!      estimates and their associated degrees of freedom for the numerator of the spectral ratio or for the first real
!      multi-channel time series if a Kolmogorov-Smirnov statistic is used.
!
!      It is assumed that this NetCDF file has been created by NCSTAT procedures performing spectral computations
!      like :ref:`comp_cross_spectrum_3d` or :ref:`comp_cross_spectrum2_3d` with the **-conflim** argument activated (e.g.,
!      this NetCDF file must contain two NetCDF variables named edof and freq).
!
!   #) The **-df=**\ *denominator_netcdf_file* argument specifies the NetCDF file containing the multi-channel PSD
!      estimates and their associated degrees of freedom for the denominator of the spectral ratio or for the second real
!      multi-channel time series if a Kolmogorov-Smirnov statistic is used.
!
!      If this argument is not specified, it is assumed that the *denominator_netcdf_file* is 
!      the same as the *numerator_netcdf_file*.
!
!      It is also assumed that this NetCDF file has been created by NCSTAT procedures performing spectral computations
!      like :ref:`comp_cross_spectrum_3d` or :ref:`comp_cross_spectrum2_3d` with the **-conflim** argument activated (e.g.,
!      this NetCDF file must contain two NetCDF variables named edof and freq).
!
!   #) The **-nv=**\ *numerator_netcdf_variable* and **-dv=**\ *denominator_netcdf_variable* arguments specify
!      the NetCDF variables containing the multi-channel PSD estimates for the numerator and denominator
!      of the spectral ratio, respectively (or the first and second PSD estimates when a Komogorov-Smirnov statistic is used).
!
!      The geographical shapes of the numerator_netcdf_variable and denominator_netcdf_variable must agree.
!
!      In addition, the multi-channel PSD estimates stored in *numerator_netcdf_variable* and *denominator_netcdf_variable*
!      must have been normalized in the same way and not been smoothed in the frequency domain (for details, see the arguments
!      **-nsm=** and **-nonormpsd** in the NCSTAT procedures :ref:`comp_cross_spectrum_3d` or :ref:`comp_cross_spectrum2_3d`
!      in order to verify the conditions of use of the statistical tests available in comp_spectrum_diff_3d.
!
!      Finally, if **-a=**\ ``chi2_perm`` or  **-a=**\ ``ks_perm``, the two estimated multi-channel power spectra must have been obtained by
!      averaging exactly the same number of (modified) periodograms.
!
!   #) The optional argument **-m=**\ *input_mesh_mask_netcdf_file* specifies the land-sea mask
!      to apply to the *numerator_netcdf_variable* and *denominator_netcdf_variable* for transforming
!      these tridimensional NetCDF variables as rectangular matrices before computing the spectral ratio of the multi-channel
!      PSD estimates or the differences between their cumulative periodograms.
!
!      By default, it is assumed that each cell in the 2-D grid-mesh associated with the input tridimensional
!      NetCDF variables is a valid power spectrum (e.g., missing values are not present in the PSD estimates).
!
!      The geographical shapes of the *numerator_netcdf_variable* (in the *numerator_netcdf_file*), the
!      *denominator_netcdf_variable* (in the *denominator_netcdf_file*) and the mask
!      (in the *input_mesh_mask_netcdf_file*) must agree if an *input_mesh_mask_netcdf_file* is used.
!
!      Refer to :ref:`comp_clim_3d` or :ref:`comp_mask_3d` for creating a valid
!      *input_mesh_mask_netcdf_file* NetCDF file for regular or gaussian grids before
!      using comp_spectrum_diff_3d.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that the input NetCDF variables
!      are from an experiment with the NEMO model (ORCA configuration and R2, R4 or R05 resolutions).
!      This argument is also used to determine the name of the mesh_mask variable if
!      an *input_mesh_mask_netcdf_file* is used.
!
!   #) If the **-x=**\ *lon1,lon2* and **-y=**\ *lat1,lat2* arguments are missing the whole geographical
!      domain associated with the the *numerator_netcdf_variable* and *denominator_netcdf_variable*  is used.
!
!      The longitude or latitude range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_3d` for transforming geographical coordinates as indices or
!      generating an appropriate mesh-mask before using comp_spectrum_diff_3d.
!
!   #) If the **-np=**\ *numerator_period* argument is specified, the multi-channel PSD estimates in the *numerator_netcdf_variable*
!      have been computed for different periods separately (as determined by the value of the **-p=**\ *period_length* argument
!      in :ref:`comp_cross_spectrum_3d`). In that case, the **-np=** argument selects the period in the 
!      *numerator_netcdf_variable* which must be used for computing the spectral ratio or the difference between the cumulative
!      periodograms of the two multi-channel PSD estimates.
!
!   #) If the **-dp=**\ *denominator_period* argument is specified, the multi-channel PSD estimates in the *denominator_netcdf_variable* 
!      have been computed for different periods separately (as determined by the value of the **-p=**\ *period_length* argument
!      in :ref:`comp_cross_spectrum_3d`. In that case, the **-dp=** argument selects the period in the 
!      *denominator_netcdf_variable* which must be used for computing the spectral ratio or the difference between the cumulative
!      periodograms of the two multi-channel PSD estimates.
!
!   #) If the **-t=**\ *freq1,freq2*  argument is missing, the spectral ratios or differences between the cumulative
!      periodograms of the two multi-channel PSD estimates are computed for the whole set of frequencies stored in the 
!      *numerator_netcdf_variable* and *denominator_netcdf_variable* (e.g., the multi-channel power
!      spectra must have been estimated for the same set of frequencies).
!
!      The selected frequency interval is a vector of two integers specifying the first and last 
!      frequencies for which the spectral ratios  or differences between the cumulative periodograms must be computed.
!      The indices are relative to ``1``.
!
!      However, in all cases, the zero and Nyquist frequencies are excluded from the computations of the statistics.
!
!   #) The **-a=** argument selects the type of test statistic and/or the method for computing critical probabilities associated with
!      the test statistic: 
!                                                                              
!      * If **-a=**\ ``minmax``, the minimum and maximum values of the spectral ratios of the two estimated multi-channel spectra
!        are used as a test statistic. The critical probabilities are derived from the associated probabilities
!        of obtaining, respectively, a value less (for the minimum ratio) and higher (for
!        the maximum ratio) than attained under the null hypothesis of a common underlying population spectrum
!        for the two multi-channel time series. The computed critical probabilities are only approximate and are conservative.
!        The null hypothesis is that the two "true" underlying multi-channel spectra are the same.
!        For more details, see Chapter 4 of  [Diggle]_ .
!      * If **-a=**\ ``range``, the range of the Log spectral ratios of the two estimated multi-channel spectra is used as a test statistic.
!        The null hypothesis is that of a common shape of the two multi-channel power spectra of the two multi-channel time series.
!        For more details and theory, see  [Coates_Diggle]_  [Potscher_Reschenhofera]_  [Potscher_Reschenhoferb]_ .
!      * If **-a=**\ ``chi2``, a standard CHI2 statistic computed from the Log spectral ratios of the two multi-channel power spectra is used
!        as a test statistic. The null hypothesis is that of a common spectrum of the two multi-channel series and asymptotic theory is used
!        to derive the associated critical probabilities.
!        For more details and theory, see  [Priestley]_  [Jenkins_Watts]_ .
!      * If **-a=**\ ``chi2_perm``, a CHI2 statistic computed from the Log spectral ratios is also used, but the distribution of this
!        statistic under the null hypothesis of a common power spectrum for the two multi-channel series 
!        is approximated by recalculating the CHI2 statistic for some large number (e.g., see the desciption of the **-nb=**
!        argument below) of random interchanges of periodogram or PSD ordinates at each frequency for the two 
!        estimated multi-channel power spectra.
!        This statistical randomization test relies on the assumptions that the different spectral ordinates have the same
!        sampling distribution and are independent of each other  [Priestley]_ . This means, in particular, that the PSD
!        ordinates corresponding to the zero and Nyquist frequencies must be excluded and that the two estimated multi-channel
!        power spectra have not been obtained by smoothing the periodograms in the frequency domain.
!        For more details and theory, see  [Priestley]_  [Diggle_Fisher]_ .
!      * If **-a=**\ ``ks_perm``, a Kolmogorov-Smirnov statistic is used, but the distribution of this statistic under 
!        the null hypothesis of a common shape for the two "true" power spectra of the two multi-channel series 
!        is approximated by calculating the Kolmogorov-Smirnov statistic for some large number
!        (e.g., as determined by the **-nb=** argument) of random interchanges of periodogram or PSD ordinates at each frequency
!        for the two estimated multi-channel spectra.
!        For more details and theory, see [Priestley]_  [Diggle_Fisher]_ .
!                                                                              
!      If **-a=**\ ``chi2_perm`` or ``ks_perm``, the number of frequencies used (e.g., as determined by the **-t=** argument) 
!       must be greater or equal to ``10``.                                   
!                                                                              
!      The default is **-a=**\ ``chi2``.                                                       
!
!   #) The **-nb=**\ *number_of_shuffles* argument specifies the number of random interchanges of periodogram or PSD ordinates
!      at each frequency in order to approximate the randomization distribution of the Kolmogorov-Smirnov or CHI2
!      statistics under the null hypothesis if **-a=**\ ``chi2_perm`` or ``ks_perm``.
!
!      The default value is ``999``.
!
!   #) The **-ksnonorm** argument specifies that the Kolmogorov-Smirnov statistics must be
!      computed from non-normalized cumulative periodograms if **-a=**\ ``ks_perm``.
!      In that case the null hypothesis which is tested is that the "true" underlying population power spectra
!      of the two multi-channel time series are the same instead of having the same shape.
!
!      By default, the Kolmogorov-Smirnov statistics are calculated from normalized
!      cumulative periodograms and the null hypothesis which is tested is that of
!      a common shape for the two "true" underlying multi-channel spectra (see the description
!      of the **-a=** argument for more details).
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator for
!      the output NetCDF variables in the *output_netcdf_file*. 
!
!      If the **-mi=** argument is not specified, the *missing_value* in the output NetCDF variables is set
!      to ``1.e+20``.
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!
!      By default, the results are stored as single-precision floating point numbers 
!      in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g., ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g., ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!                                                                              
!   #) For more details on power spectrum analysis, theory for the statistical tests used in 
!      comp_spectrum_diff_3d and examples of use in the climate literature, see
!
!      * "The use of Fast Fourier Transform for the estimation of power spectra: A method based on time averaging over short, modified periodograms",
!        by P.D. Welch, IEEE trans. on audio and electroacoustics, 15:2, 70-73, 1967. DOI: https://doi.org/10.1109/TAU.1967.1161901
!
!      * "Time series: a biostatistical introduction",
!        by P.J. Diggle, Clarendon Press, Oxford, Chap. 4, 1990. http://www.oupcanada.com/catalog/9780198522263.html
!
!      * "Tests for comparing two estimated spectral densities",
!        by D.S. Coates and P.J. Diggle, Journal of Time series Analysis, 7:1, 7-20, 1986. DOI: https://doi.org/10.1111/j.1467-9892.1986.tb00482.x
!
!      * "Nonparametric comparison of cumulative periodograms",
!        by P.J. Diggle and N.I. Fisher, Applied Statistics, 40:3, 423-434, 1991. DOI: https://doi.org/10.2307/2347522
!
!      * "Spectral Analysis and its Applications",
!        by G.M.Jenkins and D.G., San Francisco: Holden-Day. ISBN-10: 0816244642, 1968.
!
!      * "Spectral Analysis and Time Series",
!        by M.B. Priestley, London: Academic Press, 1981. ISBN-10: 0125649223
!
!      * "Discriminating between two spectral densities in case of replicated observations",
!        by B.M. Potscher and E. Reschenhofer, Journal of Time series Analysis, 9:3, 221-224, 1988. DOI: https://doi.org/10.1111/j.1467-9892.1988.tb00466.x
!
!      * "Distribution of the Coates-Diggle test statistic in case of replicated observations",
!        by B.M. Potscher and E. Reschenhofer, Statistics, 20:3, 417-421, 1989. DOI: https://doi.org/10.1080/02331888908802190
!
!      * "Impact of intra-daily SST variability on ENSO characteristics in a coupled model"
!        by S. Masson et al., 2012, Climate Dynamics, Vol. 39:681-707, doi: `10.1007/s00382-011-1247-2`_
!
!      * "The role of the intra-daily SST variability in the Indian monsoon variability in a global coupled model"
!        by P. Terray et al., 2012, Climate Dynamics, Vol. 39:729-754, doi: `10.1007/s00382-011-1240-9`_
!
!      * "Southern Hemisphere extra-tropical forcing: A new paradigm for El Nino-Southern Oscillation"
!        by P. Terray, 2011, Climate Dynamics, Vol. 36:2171-2199, doi: `10.1007/s00382-010-0825-z`_
!
! .. _`10.1007/s00382-011-1247-2` : http://dx.doi.org/10.1007/s00382-011-1247-2
! .. _`10.1007/s00382-011-1240-9` : http://dx.doi.org/10.1007/s00382-011-1240-9
! .. _`10.1007/s00382-010-0825-z` : http://dx.doi.org/10.1007/s00382-010-0825-z
!
! Outputs
! -------
!
!   comp_spectrum_diff_3d creates an output NetCDF file that contains the selected test statistics for the two input multi-channel power spectra
!   and the associated critical probabilities under the assumption that the two "true" underlying multi-channel power spectra are the
!   same or have the same shape depending on the selected test statistic (see the description of the **-a=** argument above for more details
!   on the available test statistics and the null hypothesis used to derive the critical probabilities for each type_of_statistic.
!
!   The output NetCDF dataset contains the following NetCDF variables (in the description below, nlat and nlon are
!   the lengths of the spatial dimensions of the two input NetCDF variables *numerator_netcdf_variable* and
!   *denominator_netcdf_variable*, and nfreq is the number of frequencies at which the two input multi-channel
!   power spectra have been estimated) :
!
!   #) frequency\ ``(nfreq)`` : the frequencies at which the input PSD estimates have been computed.
!
!   #) periodicity\ ``(nfreq)`` : the corresponding periods in number of time observations.
!
!   #) *numerator_netcdf_variable*\ _\ *denominator_netcdf_variable*\ _stat\ ``(nlat,nlon)`` : the multi-channel test statistics
!      computed from the two input multi-channel power spectra stored in the two input NetCDF variables *numerator_netcdf_variable*
!      and *denominator_netcdf_variable*. The type of the computed statistic, as determined by the **-a=** argument, is stored as a
!      a global attribute in the output NetCDF dataset.
!
!   #) *numerator_netcdf_variable*\ _\ *denominator_netcdf_variable*\ _prob\ ``(nlat,nlon)`` : the multi-channel critical probabilities
!      associated with the computed multi-channel test statistics. These critical probabilities are computed or estimated (depending on
!      the type of statistic) under the null hypothesis that the two "true" multi-channel power spectra are identical or have the
!      same shape.
!
!      If the **-a=**\ ``minmax`` argument is used, these critical probabilities are conservative and are simply computed as two times the
!      minimum of the critical probabilities associated with the minimum and maximum values of the spectral ratios computed from the two input
!      multi-channel power spectra. The minimum and maximum values of the spectral ratios and their associated critical probabilities are also
!      stored in the output NetCDF dataset.
!
!   #) *numerator_netcdf_variable*\ _\ *denominator_netcdf_variable*\_minratio _stat\ ``(nlat,nlon)`` : the multi-channel minimum values of the
!      spectral ratios computed from the two input multi-channel power spectra stored in the two input NetCDF variables *numerator_netcdf_variable*
!      and *denominator_netcdf_variable*.
!
!      This variable is stored only if the **-a=**\ ``minmax`` argument has been specified when calling comp_spectrum_diff_3d.
!
!   #) *numerator_netcdf_variable*\ _\ *denominator_netcdf_variable*\ _minratio_prob\ ``(nlat,nlon)`` : the multi-channel critical probabilities
!      associated with the minimum values of the multi-channel spectral ratios. These critical probabilities give the probabilities of obtaining
!      minimum values less than attained under the null hypothesis of a common power spectra for the two (multi-channel) time series.
!
!      This variable is stored only if the **-a=**\ ``minmax`` argument has been specified when calling comp_spectrum_diff_3d.
!
!   #) *numerator_netcdf_variable*\ _\ *denominator_netcdf_variable*\_maxratio _stat\ ``(nlat,nlon)`` : the multi-channel maximum values of the
!      spectral ratios computed from the two input multi-channel power spectra stored in the two input NetCDF variables *numerator_netcdf_variable*
!      and *denominator_netcdf_variable*.
!
!      This variable is stored only if the **-a=**\ ``minmax`` argument has been specified when calling comp_spectrum_diff_3d and is used to derive
!      the conservative critical probabilities stored in the NetCDF variable *numerator_netcdf_variable*\ _\ *denominator_netcdf_variable*\ _prob .
!
!   #) *numerator_netcdf_variable*\ _\ *denominator_netcdf_variable*\ _maxratio_prob\ ``(nlat,nlon)`` : the multi-channel critical probabilities
!      associated with the maximum values of the multi-channel spectral ratios. These critical probabilities give the probabilities of obtaining
!      maximum values higher than attained under the null hypothesis of a common power spectra for the two (multi-channel) time series.
!
!      This variable is stored only if the **-a=**\ ``minmax`` argument has been specified when calling comp_spectrum_diff_3d and is used to derive
!      the conservative critical probabilities stored in the NetCDF variable *numerator_netcdf_variable*\ _\ *denominator_netcdf_variable*\ _prob .
!
!   All the multi-channel spectral statistics and their associated critical probabilities involving the two input NetCDF
!   variables *numerator_netcdf_variable* and *denominator_netcdf_variable* are packed in tridimensional variables
!   whose first and second dimensions are exactly the same as those associated with these input NetCDF variables even
!   if you restrict the geographical domain with the **-x=** and **-y=** arguments. However, outside the selected domain,
!   the output NetCDF variables are filled with missing values.
!
! Examples
! --------
!
!   #)   For testing that the two multi-channel power spectra stored in a NetCDF variable called ``sst`` extracted, respectively,
!        from the files ``spectrum_sst.monthly.1979_2005.nc`` and ``spectrum_sst.monthly.1950_1978.nc`` are the same (more precisely
!        that the two "true" underlying power spectra are identical) using a standard CHI2 statistic, use the following command :
!
!        ::
!
!          $ comp_spectrum_diff_3d  -fn=spectrum_sst.monthly.1979_2005.nc \
!                                   -fd=spectrum_sst.monthly.1950_1978.nc \
!                                   -nv=sst \
!                                   -dv=sst \
!                                   -o=spectrum_diff_sst_1979_2005.1950_1978.nc
!
!-
! ================================================================================================
!
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,   only : nf90_hdf5
#endif
!   
    use netcdf,   only : nf90_max_name, nf90_max_var_dims, nf90_clobber, nf90_real4, nf90_real8,        &
                         nf90_sizehint_default, nf90_open, nf90_clobber,                                &
                         nf90_nowrite, nf90_noerr, nf90_get_att, nf90_get_var, nf90_strerror,           &
                         nf90_create, nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_unlimited,         &
                         nf90_def_var, nf90_inq_attname, nf90_put_att, nf90_copy_att, nf90_enddef,      &
                         nf90_global, nf90_put_var, nf90_close
!   
    use Statpack, only : i1b, i4b, stnd, extd, lgl, true, false, zero, one, kfix, kexp, kint, tolower,    & 
                         case_change, string_to_val, arth, allocate_error, merror, replace_nan,           &
                         spctrm_ratio2, spctrm_ratio3, spctrm_ratio4, spctrm_diff, spctrm_diff2
!   
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing,                    &
                                 get_listarg, get_variable_dims, get_variable_dimcoords
!   
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_spectrum_diff_3d : ',           &
!      
                  cmethod(5)=(/ "chi2     ", "minmax   ", "range    ", "chi2_perm", "ks_perm  " /)
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid_mask, maskid, ncid_stat, periodid,     &
                     stat_id, prob_id, minratio_id, maxratio_id,               &
                     probminratio_id, probmaxratio_id, nrep, method,           &
                     i, j, j1, j2, k, nfic, ndim, freq1, freq2, freq3, nf, h0, &
                     lat1, lat2, lat3, nlat, lon1, lon2, lon3, nlon, nvar,     &
                     rec, offset_rec, output_type, output_mode
    integer(i4b), dimension(2)                 :: ncid, freqid, dofid, varid, nfreq, period, startwri, countwri
    integer(i4b), dimension(3)                 :: coor_id2
    integer(i4b), dimension(6)                 :: vecid
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i4b), dimension(4,2)               :: startlec, countlec
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),   allocatable    :: stat, prob, minratio, maxratio, probmin, probmax
    real(stnd), dimension(:,:),   allocatable  :: freq, edof, temp
    real(stnd), dimension(:,:,:), allocatable  :: spect
    real(stnd)                                 :: missing
!
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                                :: use_tlimited, ksnorm, use_x, use_y, use_mask
    logical(lgl), dimension(2)                  :: use_period
    logical(lgl), dimension(nf90_max_var_dims)  :: use_coor_variable
    logical(lgl), dimension(:,:),   allocatable :: mask
!
    character(len=nchar)                                       :: name, name2, name3
    character(len=nf90_max_name)                               :: stat_name, mask_name, att_name
    character(len=nf90_max_name), dimension(2)                 :: fic_name, var_name
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                          :: fmt
    character                                  :: chr, grid_type
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
! 
    if ( narg==0 ) then
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Test if two estimated multi-channel power spectra are significantly different or have different shape'
        write (prtunit,*) ' using various nonparametric statistical tests based on the (Log) ratios of the two estimated'
        write (prtunit,*) ' multi-channel spectra or the differences between (normalized) cumulative periodograms derived'
        write (prtunit,*) ' from the two estimated multi-channel power spectra.'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_spectrum_diff_3d  '
        write (prtunit,*) '        -nf=numerator_netcdf_file'
        write (prtunit,*) '        -nv=numerator_netcdf_variable'
        write (prtunit,*) '        -dv=denominator_netcdf_variable'
        write (prtunit,*) '        -df=denominator_netcdf_file       (optional)'
        write (prtunit,*) '        -m=input_mesh_mask_netcdf_file    (optional)'
        write (prtunit,*) '        -g=grid_type                      (optional : n, t, u, v, w, f)'
        write (prtunit,*) '        -x=lon1,lon2                      (optional)'
        write (prtunit,*) '        -y=lat1,lat2                      (optional)'
        write (prtunit,*) '        -np=numerator_period              (optional)'
        write (prtunit,*) '        -dp=denominator_period            (optional)'
        write (prtunit,*) '        -t=freq1,freq2                    (optional)'
        write (prtunit,*) '        -a=type_of_statistic              (optional : minmax, range, chi2, chi2_perm, ks_perm)'
        write (prtunit,*) '        -nb=number_of_shuffles            (optional)'
        write (prtunit,*) '        -o=output_netcdf_file             (optional)'
        write (prtunit,*) '        -mi=missing_value                 (optional)'
        write (prtunit,*) '        -ksnonorm                         (optional)'
        write (prtunit,*) '        -double                           (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '        -bigfile                          (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                             (optional)'
#endif
        write (prtunit,*) '        -tlimited                         (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-df=         : by default, the denominator_netcdf_file is the same as the numerator_netcdf_file'
        write (prtunit,*) '-m=          : by default, an input_mesh_mask_netcdf_file is not used'
        write (prtunit,*) '-g=n         : by default, a regular or gaussian grid is assumed'
        write (prtunit,*) '-x=          : the whole longitude domain associated with the two input NetCDF variables'
        write (prtunit,*) '-y=          : the whole latitude domain associated with the two input NetCDF variables'
        write (prtunit,*) '-np=         : by default, the numerator_period argument is not used'
        write (prtunit,*) '-dp=         : by default, the denominator_period argument is not used'
        write (prtunit,*) '-t=          : by default, the whole set of frequencies in the two input NetCDF files,'
        write (prtunit,*) '               but excluding the zero and Nyquist frequencies'
        write (prtunit,*) '-a=chi2      : by default,  a CHI2 statistic is used'
        write (prtunit,*) '-nb=999      : by default, 999 random interchanges are performed if a randomization test is'
        write (prtunit,*) '               requested (e.g., if -a=chi2_perm or ks_perm is used)'
        write (prtunit,*) '-o=spectrum_diff_numerator_netcdf_variable.denominator_netcdf_variable.nc'
        write (prtunit,*) '-mi=1.e+20   : by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-ksnonorm    : by default, the Kolmogorov-Smirnov statistic is computed from normalized'
        write (prtunit,*) '               cumulative periodograms if -a=ks_perm is specified'
        write (prtunit,*) '-double      : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile     : by default, a NetCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5        : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited    : by default, the time (e.g., frequency) dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<3 .or. narg>20 ) then
        call merror( name_proc//"Wrong number of line arguments" )
    end if
!
    fic_name(:) = " "
    var_name(:) = " "
    mask_name   = " "
    stat_name   = " "
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode  = nf90_clobber
    output_type  = nf90_real4
    method       = 1
    nrep         = 999
!
    missing = def_missing
!
    grid_type     = "n"
!
    lon1  = 1
    lon2  = huge( lon2 )
    lat1  = 1
    lat2  = huge( lat2 )
    freq1 = 1
    freq2 = huge( freq2 )
!
    period(:2)     = -1
    use_period(:2) = false
!
    h0 = 1
!
    ksnorm       = true
    use_tlimited = false
    use_mask     = false
    use_x        = false
    use_y        = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-t=')
!
!               GET THE FREQUENCY INTERVAL.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "freq1 and freq2 are integer values",   &
                                  "Wrong frequency interval",             &
                                  freq1, freq2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE NETCDF MESH-MASK DATASET.
!
                mask_name = name(4:)
!
                use_mask = true
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                stat_name = name(4:)
!
            case ('-a=')
!
!               DETERMINE THE TYPE OF TEST STATISTICS AND THE NULL HYPOTHESIS USED IN EACH CASE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( cmethod(1) )
                        method = 1
                        h0 = 1
                    case ( cmethod(2) )
                        method = 2
                        h0 = 1
                   case ( cmethod(3) )
                        method = 3
                        h0 = 0
                    case ( cmethod(4) )
                        method = 4
                        h0 = 1
                    case ( cmethod(5) )
                        method = 5
                        h0 = 0
                    case default
                        call merror( name_proc//"Unknown test statistic option" )
                end select
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-nf=')
!
!                       GET THE NAME OF THE NUMERATOR NETCDF DATASET.
!
                        fic_name(1) = name(5:)
!
                    case ('-nv=')
!
!                       GET THE NAME OF THE NUMERATOR NETCDF VARIABLE.
!
                        var_name(1) = name(5:)
!
                    case ('-np=')
!
!                       GET THE NUMERATOR PERIOD.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kint ) &
                        call merror( name_proc//"The numerator_period is an integer value" )
!
                        read( name(5:), fmt ) period(1)
!
                        if ( period(1)<=0 ) &
                        call merror( name_proc//"The -np= parameter is a positive integer" )
!
                        use_period(1) = true
!
                    case ('-df=')
!
!                       GET THE NAME OF THE DENOMINATOR NETCDF DATASET.
!
                        fic_name(2) = name(5:)
!
                    case ('-dv=')
!
!                       GET THE NAME OF THE DENOMINATOR NETCDF VARIABLE.
!
                        var_name(2) = name(5:)
!
                    case ('-dp=')
!
!                       GET THE DENOMINATOR PERIOD.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kint ) &
                        call merror( name_proc//"The denominator_period is an integer value" )
!
                        read( name(5:), fmt ) period(2)
!
                        if ( period(2)<=0 ) &
                        call merror( name_proc//"The -dp= parameter is a positive integer" )
!
                        use_period(2) = true
!
                    case ('-nb=')
!
!                       GET THE NUMBER OF RANDOM INTERCHANGES OF PERIODOGRAM ORDINATES FOR THE PERMUTATION TESTS.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) nrep
                        end if
                        if ( kcode/=kint .or. nrep<=0  )                                      &
                        call merror( name_proc//"The number of shuffles is a positive integer value" )
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                        call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:) )
!
                            case ('-ksnonorm')
!
!                               COMPUTE KS STATISTICS FROM NON-NORMALIZED CUMULATIVE PERIODOGRAMS.
!
                                ksnorm = false
!
#ifdef _USE_NETCDF36
                            case ('-bigfile')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                            case ('-hdf5')
!
!                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                output_mode = nf90_hdf5
#endif
!
                            case ('-tlimited')
!
!                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                use_tlimited = true
!
                            case ('-double')
!
!                               STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                output_type = nf90_real8
!
                            case default
!
!                               WRONG ARGUMENT.
!
                                call merror( name_proc//"Wrong argument" )
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name(1)==" " )                                                       &
    call merror( name_proc//"Input numerator NETCDF file is missing" )
!
    if ( fic_name(2)==" " )   fic_name(2) = fic_name(1)
!
    if ( var_name(1)==" " )                                                       &
    call merror( name_proc//"Input numerator NETCDF variable is missing" )
!
    if ( var_name(2)==" " )                                                       &
    call merror( name_proc//"Input denominator NETCDF variable is missing" )
!
    if ( stat_name==" " )                                                         &
    stat_name="spectrum_diff_"//trim(var_name(1))//"."//trim(var_name(2))//".nc"
!
!   CORRECT THE NULL HYPOTHESIS IF REQUIRED.
!
    if ( method==5 .and. .not.ksnorm ) h0 = 1
!
!   OPEN THE NETCDF DATASETS.
!
    if ( use_mask ) then
        status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    nfic = merge( 1, 2, fic_name(1)==fic_name(2) )
!
    name = "(numerator)"
!
    do j = 1, 2
!
        if ( j<=nfic ) then
!
            i = nf90_sizehint_default
            status = nf90_open( trim(fic_name(j)), nf90_nowrite, ncid(j), chunksize=i )
!
            if ( status/=nf90_noerr )                                                &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           GET THE VARIABLES ID FROM THEIR NAME.
!
!           FIND THE DIMENSION LENGTH AND ID FOR THE NETCDF VARIABLES edof
!           AND frequency IN THE NETCDF FILES.
!
            call get_variable_dims(  name_proc, "frequency", ncid(j), freqid(j),        &
                                     dimids_out=dimids, dimlens_out=dimlens  )
!
            ndim = count(dimlens/=0)
            if ( ndim/=1  )     &
            call merror( name_proc//"Input NETCDF variable frequency "//trim(name)//" is not unidimensional" )
!
            nfreq(j) = dimlens(1)
            i        = dimids(1)
!
            call get_variable_dims(  name_proc, "edof", ncid(j), dofid(j),      &
                                     dimids_out=dimids, dimlens_out=dimlens     )
!
            ndim = count(dimlens/=0)
            if ( ndim/=1  )     &
            call merror( name_proc//"Input NETCDF variable edof "//trim(name)//" is not unidimensional" )
!
            if ( dimids(1)/=i  )     &
            call merror( name_proc//"Incorrect input "//trim(name)//" NETCDF file" )
!
        else
!
!           THE TWO INPUT NETCDF DATASETS ARE IDENTICAL.
!
            ncid(2)  = ncid(1)
            nfreq(2) = nfreq(1)
!
        end if
!
        if ( j==1 ) then
!
!           GET THE VARIABLE ID FROM ITS NAME.
!           FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!           GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!           FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!           FOR THE COORDINATE VARIABLES.
!
            call get_variable_dimcoords(  name_proc, trim(var_name(j)), ncid(j),                       &
                                          varid(j),  dimids_out=dimids, dim_names_out=dim_names,       &
                                          dimlens_out=dimlens, coor_id_out=coor_id,                    &
                                          coor_natts_out=coor_natts, coor_type_out=coor_type,          &
                                          use_coor_variable_out=use_coor_variable                      )
!
            nlon   = dimlens(1)
            nlat   = dimlens(2)
!
        else
!
!           GET THE VARIABLE IDS FROM THEIR NAME.
!           FIND THE DIMENSION LENGTHS AND IDS FOR THE NETCDF VARIABLES.
!
            call get_variable_dims(  name_proc, trim(var_name(j)), ncid(j),                 &
                                     varid(j),  dimids_out=dimids, dimlens_out=dimlens      )
!
            if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
            call merror( name_proc//"Shape disagreement between the two input NETCDF files" )
!
        end if
!
        ndim = count(dimlens/=0)
!
        if ( use_period(j) ) then
!
            if ( ndim/=4  )     &
            call merror( name_proc//"Input NETCDF variable "//trim(name)//" is not fourdimensional" )
!
!           CHECK THE PERIODS.
!
            if ( period(j)>dimlens(4) )      &
            call merror( name_proc//"period "//trim(name)//" is out of bounds" )
!
        else
!
            if ( ndim/=3  )     &
            call merror( name_proc//"Input NETCDF variable "//trim(name)//" is not tridimensional" )
!
        end if
!
        if ( dimids(3)/=i )     &
        call merror( name_proc//"Incorrect "//trim(name)//" NETCDF file" )
!
        name = "(denominator)"
!
    end do
!
    if ( nfreq(1)/=nfreq(2) )        &
    call merror( name_proc//"Frequencies are not the same in the two input NETCDF files" )
!
    nf     = nfreq(1)
!
    if ( use_mask ) then
!
        chr = merge( 't', grid_type, grid_type=='w' )
!
!       GET THE MASK ID FROM ITS NAME.
!       VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
        call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
        call merror( name_proc//"Shape disagreement with the mask" )
!
        if ( .not.use_x ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
            if ( status==nf90_noerr ) lon1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
            if ( status==nf90_noerr ) lon2 = i
!
        end if
!
        if ( .not.use_y ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
            if ( status==nf90_noerr ) lat1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
            if ( status==nf90_noerr ) lat2 = i
!
        end if
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    if ( lon1<0 ) then
        lon2 = max( lon2, 1 )
        lon3 = nlon + lon1 - lon2 + 1
    else
        lon1 = max( lon1, 1 )
        lon2 = min( lon2, nlon )
        lon3 = lon2 - lon1 + 1
    end if
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    if ( lon1<0 ) then
       lon3 = nlon
       j1   = 1
       j2   = nlon
    else
       j1 = lon1
       j2 = lon2
    end if
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE FREQUENCY INTERVAL AND ELIMINATE THE ZERO AND NYQUIST FREQUENCIES.
!
    freq1 = max( freq1, 2 )
    freq2 = min( freq2, nf-1 )
    freq3 = freq2 - freq1 + 1
!
    if ( method>=4 .and. freq3<10 )     &
    call merror( name_proc//"The number of frequencies must be greater or equal to 10 with permutation tests" )
!
!   COMPUTE THE GEOGRAPHICAL MASK.
!
    if ( use_mask ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( temp(nlon,nlat), mask(nlon,nlat), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET THE MASK.
!
        status = nf90_get_var( ncid_mask, maskid, temp )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       DETERMINE THE GRID POINTS USED IN THE CORRELATION ANALYSIS FROM THE MASK.
!
        mask(:nlon,:nlat) = temp(:nlon,:nlat)==one
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
!
    else
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( mask(nlon,nlat), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       CONSTRUCT A DUMMY MASK.
!
        mask(:nlon,:nlat) = true
!
    end if
!
!   SELECT THE DOMAIN.
!
    if ( lon1>0 ) then
        mask(1:lon1-1,:nlat)    = false
        mask(lon2+1:nlon,:nlat) = false
    else
        mask(lon2+1:nlon+lon1,:nlat) = false
    end if
!
    mask(:nlon,1:lat1-1)    = false
    mask(:nlon,lat2+1:nlat) = false
!   
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   ALLOCATE WORK ARRAYS FOR SPECTRUM, STATISTIC, PROBABILITY, DEGREES OF FREEDOM AND FREQUENCY AXIS.
!
    if ( method/=2 ) then
        allocate( spect(nvar,freq3,2), stat(nvar), prob(nvar), temp( lon3,lat3),   &
                  freq(freq3,2), edof(freq3,2), stat=status                        )
    else
        allocate( spect(nvar,freq3,2), minratio(nvar), maxratio(nvar), temp(lon3,lat3),  &
                  prob(nvar),  probmin(nvar), probmax(nvar),                             &
                  freq(freq3,2), edof(freq3,2), stat=status                              )
    end if
!
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE DATA.
!
    do j = 1, 2
!
        if ( j<=nfic ) then
!
!           GET THE FREQUENCY AXES.
!
            status = nf90_get_var( ncid(j), freqid(j), freq(:freq3,j), start=(/ freq1 /), count=(/ freq3 /)   )
!
            if ( status/=nf90_noerr )                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           GET THE DEGREES OF FREEDOM FOR THE TWO SPECTRA.
!
            status = nf90_get_var( ncid(j), dofid(j), edof(:freq3,j), start=(/ freq1 /), count=(/ freq3 /)   )
!
            if ( status/=nf90_noerr )                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        else
!
            freq(:freq3,2) = freq(:freq3,1)
            edof(:freq3,2) = edof(:freq3,1)
!
        end if
!
    end do
!
!   CHECK THE ASSUMPTIONS OF THE STATISTICAL TESTS.
!
    status = nf90_get_att( ncid(1), nf90_global, "smoothing", i )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( i==1 .or. any( edof(1,1)/=edof(:freq3,1) ) )   &
    call merror( name_proc//"Spectral estimates have been smoothed in the frequency domain (numerator)" )
!
    if ( nfic/=1 ) then
!
        if ( any( freq(:freq3,1)/=freq(:freq3,2) )   )       &
        call merror( name_proc//"Frequencies are not the same in the two input NETCDF files" )
!
        status = nf90_get_att( ncid(2), nf90_global, "smoothing", i )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( i==1 .or. any( edof(1,2)/=edof(:freq3,2) ) )   &
        call merror( name_proc//"Spectral estimates have been smoothed in the frequency domain (denominator)" )
!
        status = nf90_get_att( ncid(1), nf90_global, "normalization", i )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_get_att( ncid(2), nf90_global, "normalization", j )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( i/=j )   &
        call merror( name_proc//"Normalization of spectral estimates is different in the two input netcdf files" )
!
        if ( method>=4 ) then
            if ( any( edof(:freq3,1)/=edof(:freq3,2) )   )       &
            call merror( name_proc//"Spectral estimates have not been computed as the average of the same number of periodograms")
        end if
!
    end if
!
!   GET THE DATA.
!
!   COMPUTE THE OFFSET FOR THE RECORD NUMBER.
!
    offset_rec = freq1 - 1
!
    do j = 1, 2
!
        startlec(1:4,j)  = (/ j1,   lat1, 1_i4b, 1_i4b  /)
        countlec(1:4,j)  = (/ lon3, lat3, 1_i4b, 1_i4b   /)
!
        if ( use_period(j) ) then
            startlec(4,j)  = period(j)
            ndim = 4
        else
            ndim = 3
        end if
!
        do k = freq1, freq2
!
            startlec(3,j) = k
!
!           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
            status = nf90_get_var( ncid(j), varid(j), temp(:lon3,:lat3),              &   
                                   start=startlec(:ndim,j), count=countlec(:ndim,j)   )     
!
            if ( status/=nf90_noerr )                                       &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            rec = k - offset_rec
!
!           PACK THE DATA.
!
            spect(:nvar,rec,j) = pack( temp(:lon3,:lat3), mask(j1:j2,lat1:lat2) )
!
        end do
!
    end do
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( temp )
!
!   COMPUTE THE STATISTICS AND THE CRITICAL PROBABILITIES.
!
    select case ( method )
!
        case ( 1_i4b )
!
!           CHI2 STATISTIC.
!
            call spctrm_ratio3( spect(:nvar,:freq3,1), spect(:nvar,:freq3,2),   &
                                edof(1,1), edof(1,2), stat(:nvar),              &
                                prob(:nvar) )
        case ( 2_i4b )
!
!           MIN AND MAX RATIO STATISTICS.
!
            call spctrm_ratio2( spect(:nvar,:freq3,1), spect(:nvar,:freq3,2),                  &
                                edof(1,1), edof(1,2), prob(:nvar),                             &
                                min_ratio=minratio(:nvar), max_ratio=probmax(:nvar),           &
                                prob_min_ratio=probmin(:nvar), prob_max_ratio=probmax(:nvar)   )
!
        case ( 3_i4b )
!
!           RANGE STATISTIC.
!
            call spctrm_ratio4( spect(:nvar,:freq3,1), spect(:nvar,:freq3,2),   &
                                edof(1,1), edof(1,2), stat(:nvar),    &
                                prob(:nvar) )
!
        case ( 4_i4b )
!
!           CHI2 STATISTIC AND RANDOMIZED TEST.
!
            call spctrm_diff2( spect(:nvar,:freq3,1), spect(:nvar,:freq3,2),       &
                               stat(:nvar), prob(:nvar), nrep=nrep, initseed=true  )
!
        case ( 5_i4b )
!
!           KOLMOGOROV-SMIRNOV STATISTIC AND RANDOMIZED TEST.
!
            call spctrm_diff( spect(:nvar,:freq3,1), spect(:nvar,:freq3,2),     &
                              stat(:nvar), prob(:nvar), nrep=nrep, norm=ksnorm, &
                              initseed=true  )
!
        case default
!
            call merror( name_proc//"Unknown test analysis option" )
!
    end select
!
!   CREATE THE NEW NETCDF DATASET FOR THE FREQUENCY ANALYSIS.
!
    k = nf90_sizehint_default
    status = nf90_create( stat_name, cmode=output_mode, ncid=ncid_stat, chunksize=k )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_stat, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = nlon
    dimlens(2)   = nlat
    if ( use_tlimited ) then
        dimlens(3)   = freq3
    else
        dimlens(3)   = nf90_unlimited
    end if
!
    coor_type(3) = output_type
    dim_names(3) = "frequency"
!
    do i = 1, 3
        status = nf90_def_dim( ncid_stat, name=trim(dim_names(i)),                       &
                               len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 3
        status = nf90_def_var( ncid_stat, trim(dim_names(i)), coor_type(i),              &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                        &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end do
!
    status = nf90_def_var( ncid_stat, "periodicity",  output_type,                   &
                           dimids=dimids(3), varid=periodid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    name = trim(var_name(1))//"_"//trim(var_name(2))
!
    if ( method/=2 ) then
!
        name2 = trim(name)//"_stat"
        status = nf90_def_var( ncid_stat, trim(name2),  output_type,   &
                               dimids=dimids(1:2), varid=stat_id )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        do i = 1, 4
!
            select case ( i )
                case ( 1_i4b)
                    name2 = "_minratio_stat"
                case ( 2_i4b )
                    name2 = "_maxratio_stat"
                case ( 3_i4b )
                    name2 = "_minratio_prob"
                case ( 4_i4b )
                    name2 = "_maxratio_prob"
            end select
!
            name3 = trim(name)//trim(name2)
!
            status = nf90_def_var( ncid_stat, trim(name3),  output_type,   &
                                   dimids=dimids(1:2), varid=vecid(i) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        minratio_id     = vecid(1)
        maxratio_id     = vecid(2)
        probminratio_id = vecid(3)
        probmaxratio_id = vecid(4)
!
    end if
!
    name2 = trim(name)//"_prob"
!
    status = nf90_def_var( ncid_stat, trim(name2),  output_type,   &
                           dimids=dimids(1:2), varid=prob_id )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 3
        if ( .not.use_coor_variable(j) ) cycle
        do i = 1, coor_natts(j)
            status = nf90_inq_attname( ncid(1), coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid(1), coor_id(j), att_name, ncid_stat, coor_id2(j) )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
        end do
    end do
!
    status = nf90_put_att( ncid_stat, periodid, "long_name", "Periodicity axis" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, periodid, "short_name", "Periodicity" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, periodid, "units", "Number of observations" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    name = trim(var_name(1))//" and "//trim(var_name(2))
!
    if ( method/=2 ) then
!
        name2 = trim(cmethod(method))// " statistic of "//trim(name)
!
        status = nf90_put_att( ncid_stat, stat_id, "long_name", trim(name2)  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_stat, stat_id, "short_name", trim(cmethod(method))//" statistic"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_stat, stat_id, "units", "none"  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        do i = 1, 4
!
            select case ( i )
                case ( 1_i4b)
                    name2 = "minratio statistic"
                case ( 2_i4b )
                    name2 = "maxratio statistic"
                case ( 3_i4b )
                    name2 = "minratio probability"
                case ( 4_i4b )
                    name2 = "maxratio probability"
            end select
!
            name3 = trim(name2)//" of "//trim(name)
!
            status = nf90_put_att( ncid_stat, vecid(i), "long_name", trim(name3)  )
           if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, vecid(i), "short_name", trim(name2)  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, vecid(i), "units", "none"  )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end if
!
    name2 = "Probability under H0 of "//trim(cmethod(method))//" statistic"
!
    status = nf90_put_att( ncid_stat, prob_id, "long_name", trim(name2)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, prob_id, "short_name", "Probability"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method/=2 ) then
        vecid(:3) = (/ prob_id, coor_id2(3), stat_id  /)
        j1 = 3
        j2 = 1
    else
        vecid(:6) = (/ prob_id, probminratio_id, probmaxratio_id, coor_id2(3), minratio_id, maxratio_id /)
        j1 = 6
        j2 = 3
    end if
!
!   SET missing_value, _FillValue, valid_min AND valid_max ATTRIBUTES.
!
    if (  output_type==nf90_real8 ) then
!
        do j=1, j1
!
            status = nf90_put_att( ncid_stat, vecid(j), "missing_value", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, vecid(j), "_FillValue", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        do j=1, j2
!
            status = nf90_put_att( ncid_stat, vecid(j),  "valid_min", real( zero, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, vecid(j), "valid_max", real( one, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    else
!
        do j=1, j1
!
            status = nf90_put_att( ncid_stat, vecid(j), "missing_value", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, vecid(j), "_FillValue", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        do j=1, j2
!
            status = nf90_put_att( ncid_stat, vecid(j),  "valid_min", real( zero, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_stat, vecid(j), "valid_max", real( one, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    name2 = "Power spectrum ratio or difference analysis of "//trim(var_name(1))//" (numerator) and "//   &
            trim(var_name(2))//" (denominator)"
!
    status = nf90_put_att( ncid_stat, nf90_global, "title", trim(name2) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_stat, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    name = "numerator"
!
    do j = 1, 2
!
        if ( use_period(j) ) then
!
            status = nf90_put_att( ncid_stat, nf90_global, trim(name)//"_period", period(j) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if            
!
        name = "denominator"
!
    end do
!
    status = nf90_put_att( ncid_stat, nf90_global, "test_statistic", cmethod(method) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( h0==1 ) then
        name = "Identical underlying population power spectra"
    else
        name = "Same shape of the two underlying population power spectra"
    end if
!
    status = nf90_put_att( ncid_stat, nf90_global, "H0", name )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( method>=4 ) then
!
        status = nf90_put_att( ncid_stat, nf90_global, "number_of_shuffles", nrep )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        if ( method==5 ) then
!
            if ( ksnorm ) then
                i = 1
            else
                i = 0
            end if
!
            status = nf90_put_att( ncid_stat, nf90_global, "normalized_cumulative_periodograms", i )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if
!
    end if
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_stat )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
    i = max( nlon, nlat )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( record(i), temp(nlon,nlat), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    do j = 1, 2
!
        i = dimlens(j)
!
        if ( use_coor_variable(j) ) then
!
            status = nf90_get_var( ncid(1), coor_id(j), record(:i), count=(/ i /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_stat, coor_id2(j), record(:i), count=(/ i /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        else
            status = nf90_put_var( ncid_stat, coor_id2(j), arth(1_i4b,1_i4b,i) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end do
!
!   WRITE FREQUENCY AND PERIOD AXES.
!
    status = nf90_put_var( ncid_stat, coor_id2(3), freq(:freq3,1) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   BUILD UP PERIODICITY VECTOR.
!
    where( freq(:freq3,1)/=zero )
        freq(:freq3,2) = one/freq(:freq3,1)
    elsewhere
        freq(:freq3,2) = missing
    end where
!
    status = nf90_put_var( ncid_stat, periodid, freq(:freq3,2) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE STATISTICS AND CRITICAL PROBABILITIES.
!
    startwri(1:2) = 1
    countwri(1:2) = (/ nlon, nlat /)
!
    if ( method/=2 ) then
!
        call replace_nan( stat(:nvar), missing )
        temp(:nlon,:nlat) = unpack( stat(:nvar), mask(:nlon,:nlat), missing )
!
        status = nf90_put_var( ncid_stat, stat_id, temp, start=startwri, count=countwri )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    else
!
        call replace_nan( minratio(:nvar), missing )
        temp(:nlon,:nlat) = unpack( minratio(:nvar), mask(:nlon,:nlat), missing )
!
        status = nf90_put_var( ncid_stat, minratio_id, temp, start=startwri, count=countwri )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        call replace_nan( maxratio(:nvar), missing )
        temp(:nlon,:nlat) = unpack( maxratio(:nvar), mask(:nlon,:nlat), missing )
!
        status = nf90_put_var( ncid_stat, maxratio_id, temp, start=startwri, count=countwri )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        call replace_nan( probmin(:nvar), missing )
        temp(:nlon,:nlat) = unpack( probmin(:nvar), mask(:nlon,:nlat), missing )
!
        status = nf90_put_var( ncid_stat, probminratio_id, temp, start=startwri, count=countwri )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        call replace_nan( probmax(:nvar), missing )
        temp(:nlon,:nlat) = unpack( probmax(:nvar), mask(:nlon,:nlat), missing )
!
        status = nf90_put_var( ncid_stat, probmaxratio_id, temp, start=startwri, count=countwri )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    call replace_nan( prob(:nvar), missing )
    temp(:nlon,:nlat) = unpack( prob(:nvar), mask(:nlon,:nlat), missing )
!
    status = nf90_put_var( ncid_stat, prob_id, temp, start=startwri, count=countwri )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( method/=2 ) then
        deallocate( mask, temp, record, spect, stat, prob, freq, edof )
    else
        deallocate( mask, temp, record, spect, minratio, maxratio, prob,   &
                    probmin, probmax, freq, edof                           )
    end if
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid(1) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( ncid(1)/=ncid(2) ) then
        status = nf90_close( ncid(2) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( use_mask ) then
        status = nf90_close( ncid_mask )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_stat )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!
! END OF PROGRAM comp_spectrum_diff_3d
! ====================================
!
end program comp_spectrum_diff_3d
