program comp_spectrum_ratio_3d
!! #############################################################################
!!
!! Copyright 2024 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_spectrum_ratio_3d:
!
! ======================
! comp_spectrum_ratio_3d
! ======================
!
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 30/05/2024
!
! Purpose
! =======
!
! Compute the ratio of two multi-channel Power Spectral Density (PSD) estimates (computed from two real multi-channel time series with
! the help of :ref:`comp_cross_spectrum_3d` or :ref:`comp_cross_spectrum2_3d`) and a pointwise tolerance interval for
! this spectral ratio under the assumption that the two "true" underlying multi-channel power spectra are the same. The two multi-channel PSD
! estimates are readed from two tridimensional variables (as specified by the **-nv=** and **-dv=** arguments) readed from one or two 
! estimates NetCDF datasets (as specified by the **-nf=** and **-df=** arguments).
!
! How to compare quantitatively two estimated power spectra is an important question solved by comp_spectrum_ratio_3d.
! Suppose that *f1*\ (\ ``w``\ ) and *f2*\ (\ ``w``\ ) are ( multi-channel) PSD estimates (where ``w`` is a Fourier frequency)
! with ``df1`` and ``df2`` Degrees Of Freedom (DOF), respectively. To answer this question, spectral analysis theory [Diggle]_
! suggests to examine the spectral ratio
!
! *R*\ (\ ``w``\ ) = *f1*\ (\ ``w``\ ) / *f2*\ (\ ``w``\ )
!
! , which may be assumed to follow an F-distribution with numerator and denominator degrees of
! freedom ``df1`` and ``df2``, respectively. This result can be used to calculate pointwise confidence (or tolerance) 
! intervals for the spectral ratios at the Fourier frequencies [Diggle]_ [Masson_etal]_ [Terray_etalb]_.
!
! It is assumed that the two estimated multi-channel power spectra have been computed by the Welch's method  [Welch]_ by calls to :ref:`comp_cross_spectrum_3d`
! or :ref:`comp_cross_spectrum2_3d` with the **-conflim** argument (this argument is required for computing the DOFs of the PSD estimates, which are required
! by comp_spectrum_ratio_3d).
!
! The pointwise tolerance interval and the spectral ratio of the two estimated multi-channel power spectra are stored in an output NetCDF dataset.
!
! If you want to test if two estimated (multi-channel) power spectra have the same underlying population spectrum or if the underlying population
! spectra have the same shape for a given set of frequencies, use :ref:`comp_spectrum_diff_3d` instead of comp_spectrum_ratio_3d.
!
! For additional details on these spectral computations, see  [Welch]_  [Bloomfield]_  [Diggle]_ .
!
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_spectrum_ratio_3d \
!              -nf=numerator_netcdf_file \
!              -nv=numerator_netcdf_variable \
!              -dv=denominator_netcdf_variable \
!              -df=denominator_netcdf_file       (optional) \
!              -m=input_mesh_mask_netcdf_file    (optional) \
!              -g=grid_type                      (optional : n, t, u, v, w, f) \
!              -x=lon1,lon2                      (optional) \
!              -y=lat1,lat2                      (optional) \
!              -np=numerator_period              (optional) \
!              -dp=denominator_period            (optional) \
!              -t=freq1,freq2                    (optional) \
!              -o=output_netcdf_file             (optional) \
!              -pro=probability_interval         (optional : 0.0 > 1.) \
!              -mi=missing_value                 (optional) \
!              -logratio                         (optional) \
!              -double                           (optional) \
!              -bigfile                          (optional) \
!              -hdf5                             (optional) \
!              -tlimited                         (optional)
!
! By default
! ----------
!
!        *-df=*
!         the *denominator_netcdf_file* is the same as the *numerator_netcdf_file*
!        *-m=*
!         an *input_mesh_mask_netcdf_file* is not used
!        *-g=*
!         the *grid_type* is set to ``n``, which means that the 2-D grid-mesh associated with the input
!         *numerator_netcdf_variable* and *denominator_netcdf_variable* is assumed to be regular or Gaussian
!        *-x=*
!         the whole longitude domain associated with the *numerator_netcdf_variable* and *denominator_netcdf_variable*
!        *-y=*
!         the whole latitude domain associated with the *numerator_netcdf_variable* and *denominator_netcdf_variable*
!        *-np=*
!         the *numerator_period* argument is not used
!        *-dp=*
!         the *denominator_period* argument is not used
!        *-t=*
!         the whole set of frequencies included in the two input NetCDF files
!        *-o=*
!         *output_netcdf_file* name is set to ``spectrum_ratio_``\ *numerator_netcdf_variable*\ ``.``\ *denominator_netcdf_variable*\ ``.nc``
!         where *numerator_netcdf_variable* and *denominator_netcdf_variable* are specified with the **-nv=** and **-dv=** arguments.
!        *-pro=*
!         the *probability_interval* is set to ``0.90``, which means that a ``90``\ % tolerance interval
!         for the spectral ratio of the estimated multi-channel power spectra is computed
!        *-mi=*
!         the *missing_value* for the output NetCDF variables is set to ``1.e+20``
!        *-logratio*
!         the spectral ratios and their associated upper and lower critical spectral ratios are not Log transformed.
!         However, if **-logratio** is specified, the spectral ratios, upper and lower critical spectral ratios are Log transformed
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-bigfile*
!         a NetCDF classical format file is created. However, if **-bigfile** is activated,
!         the output NetCDF file is a 64-bit offset format file
!        *-hdf5*
!         a NetCDF classical format file is created . However, if **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension (e.g., the frequency axis) is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the frequency dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-nf=**\ *numerator_netcdf_file* argument specifies the NetCDF file containing the multi-channel power spectrum
!      estimates and their associated degrees of freedom for the numerator of the spectral ratio.
!
!      It is assumed that this NetCDF file has been created by NCSTAT procedures performing spectral computations
!      like :ref:`comp_cross_spectrum_3d` or :ref:`comp_cross_spectrum2_3d` with the **-conflim** argument activated (e.g.,
!      this NetCDF file must contain two NetCDF variables named edof and freq).
!
!   #) The **-df=**\ *denominator_netcdf_file* argument specifies the NetCDF file containing the multi-channel power spectrum
!      estimates and their associated degrees of freedom for the denominator of the spectral ratio.
!      If this argument is not specified, it is assumed that the *denominator_netcdf_file* is 
!      the same as the *numerator_netcdf_file*.
!
!      It is assumed that this NetCDF file has been created by NCSTAT procedures performing spectral computations
!      like :ref:`comp_cross_spectrum_3d` or :ref:`comp_cross_spectrum2_3d` with the **-conflim** argument activated (e.g.,
!      this NetCDF file must contain two NetCDF variables named edof and freq).
!
!   #) The **-nv=**\ *numerator_netcdf_variable* and **-dv=**\ *denominator_netcdf_variable* arguments specify
!      the NetCDF variables containing the multi-channel PSD estimates for the numerator and denominator
!      of the spectral ratio, respectively.
!
!      The geographical shapes of the numerator_netcdf_variable and denominator_netcdf_variable
!      must agree.
!
!   #) The optional argument **-m=**\ *input_mesh_mask_netcdf_file* specifies the land-sea mask
!      to apply to the *numerator_netcdf_variable* and *denominator_netcdf_variable* for transforming
!      these tridimensional NetCDF variables as rectangular matrices before computing the spectral ratio
!      of the multi-channel PSD estimates.
!
!      By default, it is assumed that each cell in the 2-D grid-mesh associated with the input tridimensional
!      NetCDF variables is a valid power spectrum (e.g., missing values are not present).
!
!      The geographical shapes of the *numerator_netcdf_variable* (in the *numerator_netcdf_file*), the
!      *denominator_netcdf_variable* (in the *denominator_netcdf_file*) and the mask
!      (in the *input_mesh_mask_netcdf_file*) must agree if an *input_mesh_mask_netcdf_file* is used.
!
!      Refer to :ref:`comp_clim_3d` or :ref:`comp_mask_3d` for creating a valid
!      *input_mesh_mask_netcdf_file* NetCDF file for regular or gaussian grids before
!      using comp_spectrum_ratio_3d.
!
!   #) If **-g=** is set to ``t``, ``u``, ``v``, ``w`` or ``f`` it is assumed that
!      the input NetCDF variables are from an experiment with the NEMO model (ORCA configuration and R2, R4 or R05 resolutions).
!      This argument is also used to determine the name of the mesh_mask variable if
!      an *input_mesh_mask_netcdf_file* is used.
!
!   #) If the **-x=**\ *lon1,lon2* and **-y=**\ *lat1,lat2* arguments are missing the whole geographical
!      domain associated with the the *numerator_netcdf_variable* and *denominator_netcdf_variable*  is used.
!
!      The longitude or latitude range must be a vector of two integers specifying the first
!      and last selected indices along each dimension. The indices are relative to ``1``\ .
!      Negative values are allowed for *lon1*. In this case the longitude domain
!      is from ``nlon``\ +\ *lon1*\ +\ ``1`` to *lon2* where ``nlon`` is the number of longitude
!      points in the grid associated with the NetCDF variable and it is assumed that the grid
!      is periodic.
!
!      Refer to :ref:`comp_mask_3d` for transforming geographical coordinates as indices or
!      generating an appropriate mesh-mask before using comp_spectrum_ratio_3d.
!
!   #) If the **-np=**\ *numerator_period* argument is specified, the multi-channel PSD estimates in the *numerator_netcdf_variable*
!      have been computed for different periods separately (as determined by the value of the **-p=**\ *period_length* argument
!      in :ref:`comp_cross_spectrum_3d`). In that case, the **-np=** argument selects the period in the 
!      *numerator_netcdf_variable* which must be used for computing the spectral ratio of the two multi-channel PSD estimates.
!
!   #) If the **-dp=**\ *denominator_period* argument is specified, the multi-channel PSD estimates in the *denominator_netcdf_variable* 
!      have been computed for different periods separately (as determined by the value of the **-p=**\ *period_length* argument
!      in :ref:`comp_cross_spectrum_3d`). In that case, the **-dp=** argument selects the period in the 
!      *denominator_netcdf_variable* which must be used for computing the spectral ratio of the two multi-channel PSD estimates.
!
!   #) If the **-t=**\ *freq1,freq2*  argument is missing, the spectral ratios are computed for the whole set of
!      frequencies stored in the *numerator_netcdf_variable* and *denominator_netcdf_variable* (e.g., the multi-channel power
!      spectra must have been estimated for the same set of frequencies).
!
!      The selected frequency interval is a vector of two integers specifying the first and last 
!      frequencies for which the spectral ratios and the associated tolerance intervals must be computed.
!      The indices are relative to ``1``.
!
!   #) The **-pro=**\ *probability_interval* argument specifies that a ( *probability_interval*\  * ``100`` )\ %
!      tolerance interval must be computed. This argument is used to determine the upper and lower critical spectral ratios
!      of the computed tolerance interval.
!      *probability_interval* must be greater than ``0.`` and less than ``1.``.
!
!      The default value is ``0.90``, e.g., a ``90`` )\ % tolerance interval is estimated.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator for
!      the output NetCDF variables in the *output_netcdf_file*. 
!
!      If the **-mi=** argument is not specified, the *missing_value* in the output NetCDF variables is set
!      to ``1.e+20``.
!
!   #) The **-logratio** argument specifies that the logarithms of the spectral ratios must be computed and stored
!      in the output NetCDF file.
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!
!      By default, the results are stored as single-precision floating point numbers 
!      in the output NetCDF file.
!
!   #) The **-bigfile** argument is allowed only if the NCSTAT software has been compiled with the
!      _USE_NETCDF36 or _USE_NETCDF4 CPP macros (e.g., ``-D_USE_NETCDF36`` or ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 3.6 library or higher.
!
!      If this argument is specified, the *output_netcdf_file* will be a 64-bit offset format file
!      instead of a NetCDF classic format file. However, this argument is recognized in the procedure
!      only if the NCSTAT software has been built with the _USE_NETCDF36 or _USE_NETCDF4 CPP macros.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g., ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence of a parameter
!      which will be used for the computations. Moreover, the number of specified parameters
!      must not be greater than the total number of allowed parameters.
!
!   #) For more details on power spectrum analysis, theory for computing a pointwise tolerance
!      interval for the ratio of two estimated power spectra and examples in the climate literature, see
!
!      * "The use of Fast Fourier Transform for the estimation of power spectra: A method based on time averaging over short, modified periodograms",
!        by P.D. Welch, IEEE trans. on audio and electroacoustics, 15:2, 70-73, 1967. DOI: https://doi.org/10.1109/TAU.1967.1161901
!
!      * "Fourier analysis of time series- An introduction",
!        by Bloomfield, P., John Wiley and Sons, New York, Chapter 6, 1976. http://as.wiley.com/WileyCDA/WileyTitle/productCd-0471889482.html
!
!      * "Time series: a biostatistical introduction",
!        by P.J. Diggle, Clarendon Press, Oxford, Chap. 4, 1990. http://www.oupcanada.com/catalog/9780198522263.html
!
!      * "Impact of intra-daily SST variability on ENSO characteristics in a coupled model"
!        by S. Masson et al., 2012, Climate Dynamics, Vol. 39:681-707, doi: `10.1007/s00382-011-1247-2`_
!
!      * "The role of the intra-daily SST variability in the Indian monsoon variability in a global coupled model"
!        by P. Terray et al., 2012, Climate Dynamics, Vol. 39:729-754, doi: `10.1007/s00382-011-1240-9`_
!
!      * "Southern Hemisphere extra-tropical forcing: A new paradigm for El Nino-Southern Oscillation"
!        by P. Terray, 2011, Climate Dynamics, Vol. 36:2171-2199, doi: `10.1007/s00382-010-0825-z`_
!
! .. _`10.1007/s00382-011-1247-2` : http://dx.doi.org/10.1007/s00382-011-1247-2
! .. _`10.1007/s00382-011-1240-9` : http://dx.doi.org/10.1007/s00382-011-1240-9
! .. _`10.1007/s00382-010-0825-z` : http://dx.doi.org/10.1007/s00382-010-0825-z
!
! Outputs
! -------
!
!   comp_spectrum_ratio_3d creates an output NetCDF file that contains the upper and lower critical spectral
!   ratios associated with the specified pointwise tolerance interval and the multi-channel spectral ratios of the
!   two input estimated multi-channel power spectra.
!
!   The output NetCDF dataset contains the following NetCDF variables (in the description below, nlat and nlon are
!   the lengths of the spatial dimensions of the input NetCDF variables *numerator_netcdf_variable* and
!   *denominator_netcdf_variable*, nfreq is the number of frequencies for which the multi-channel spectral ratios and, 
!   eventually, the upper and lower critical spectral ratios are computed as determined by the **-t=** and **-p=** arguments) :
!
!   #) frequency\ ``(nfreq)`` : the frequencies at which the spectral ratios are computed.
!
!   #) periodicity\ ``(nfreq)`` : the corresponding periods in number of time observations.
!
!   #) lower_critical_ratio\ ``(nfreq)`` : the lower critical spectral ratios associated with the
!      specified pointwise tolerance interval.
!
!      This variable is stored only if the **-logratio** argument has not been specified when calling
!      comp_spectrum_ratio_3d.
!
!   #) upper_critical_ratio\ ``(nfreq)`` : the upper critical spectral ratios associated with the
!      specified pointwise tolerance interval.
!
!      This variable is stored only if the **-logratio** argument has not been specified when calling
!      comp_spectrum_ratio_3d.
!
!   #) *numerator_netcdf_variable*\ _\ *denominator_netcdf_variable*\ _ratio\ ``(nfreq,nlat,nlon)`` : the multi-channel spectral ratios
!      computed from the two multi-channel power spectra stored in the two input NetCDF variables *numerator_netcdf_variable*
!      and *denominator_netcdf_variable*.
!
!      This variable is stored only if the **-logratio** argument has not been specified
!      when calling comp_spectrum_ratio_3d.
!
!   #) lower_critical_logratio\ ``(nfreq)`` : the lower critical spectral ratios associated with the
!      specified pointwise tolerance interval.
!
!      This variable is stored only if the **-logratio** argument has been specified when calling
!      comp_spectrum_ratio_3d.
!
!   #) upper_critical_logratio\ ``(nfreq)`` : the upper critical spectral ratios associated with the
!      specified pointwise tolerance interval.
!
!      This variable is stored only if the **-logratio** argument has been specified when calling
!      comp_spectrum_ratio_3d.
!
!   #) *numerator_netcdf_variable*\ _\ *denominator_netcdf_variable*\ _logratio\ ``(nfreq)`` : the multi-channel spectral ratios
!      computed from the two multi-channel power spectra stored the two input NetCDF variables *numerator_netcdf_variable*
!      and *denominator_netcdf_variable*.
!
!      This variable is stored only if the **-logratio** argument has been specified when calling comp_spectrum_ratio_3d.
!
!   All the multi-channel spectral ratios involving the two input NetCDF variables *numerator_netcdf_variable* and *denominator_netcdf_variable*
!   are packed in tridimensional variables whose first and second dimensions are exactly the same as those associated with these input NetCDF
!   variables even if you restrict the geographical domain with the **-x=** and **-y=** arguments. However, outside the selected domain,
!   the output NetCDF variables are filled with missing values.
!
! Examples
! --------
!
!   #)   For computing the multi-channel spectral ratios and the associated ``90``\ % confidence interval
!        for two multi-channel power spectra stored in a NetCDF variable called ``sst`` extracted, respectively,
!        from the files ``spectrum_sst.monthly.1979_2005.nc`` and ``spectrum_sst.monthly.1950_1978.nc``, use the following command :
!
!        ::
!
!          $ comp_spectrum_ratio_3d  -fn=spectrum_sst.monthly.1979_2005.nc \
!                                    -fd=spectrum_sst.monthly.1950_1978.nc \
!                                    -nv=sst \
!                                    -dv=sst \
!                                    -o=spectrum_ratio_sst_1979_2005.1950_1978.nc
!
!-
! ================================================================================================
!
!
#ifdef _USE_NETCDF4
#define _USE_NETCDF36
#endif
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF36
    use netcdf,    only : nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,   only : nf90_hdf5
#endif
!   
    use netcdf,   only : nf90_max_name, nf90_max_var_dims, nf90_clobber, nf90_real4, nf90_real8,        &
                         nf90_sizehint_default, nf90_open, nf90_clobber,                                &
                         nf90_nowrite, nf90_noerr, nf90_get_att, nf90_get_var, nf90_strerror,           &
                         nf90_create, nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_unlimited,         &
                         nf90_def_var, nf90_inq_attname, nf90_put_att, nf90_copy_att, nf90_enddef,      &
                         nf90_global, nf90_put_var, nf90_close
!   
    use Statpack, only : i1b, i4b, stnd, extd, lgl, true, false, zero, one, kfix, kexp, kint, tolower,  & 
                         case_change, string_to_val, arth, allocate_error, merror, spctrm_ratio
!   
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing,                    &
                                 get_listarg, get_variable_dims, get_variable_dimcoords
!   
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_spectrum_ratio_3d : '
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid_mask, maskid, ncid_ratio, periodid,   &
                     ratio_id, upratio_id, loratio_id,                        &
                     i, j, j1, j2, k, nfic, ndim, freq1, freq2, freq3, nf,    &
                     lat1, lat2, lat3, nlat, lon1, lon2, lon3, nlon, nvar,    &
                     rec, offset_rec, output_type, output_mode
    integer(i4b), dimension(2)                 :: ncid, freqid, dofid, varid, nfreq, period, ndimlec
    integer(i4b), dimension(3)                 :: coor_id2, startwri, countwri
    integer(i4b), dimension(4)                 :: vecid
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens, coor_natts, coor_id, coor_type
    integer(i4b), dimension(4,2)               :: startlec, countlec
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),     allocatable    :: lwr_ratio, upr_ratio
    real(stnd), dimension(:,:),   allocatable    :: freq, edof, temp
    real(stnd), dimension(:,:,:), allocatable    :: spect, spect_ratio
    real(stnd)                                   :: missing, probtol
!
    real(extd), dimension(:),     allocatable  :: record
!
    logical(lgl)                                :: use_tlimited, logratio, use_x, use_y, use_mask
    logical(lgl), dimension(2)                  :: use_period
    logical(lgl), dimension(nf90_max_var_dims)  :: use_coor_variable
    logical(lgl), dimension(:,:),   allocatable :: mask
!
    character(len=nchar)                                       :: name, name2
    character(len=nf90_max_name)                               :: ratio_name, mask_name, att_name
    character(len=nf90_max_name), dimension(2)                 :: fic_name, var_name
    character(len=nf90_max_name), dimension(nf90_max_var_dims) :: dim_names
    character(len=14)                          :: fmt
    character                                  :: chr, grid_type
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
! 
    if ( narg==0 ) then
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute the multi-channel spectral ratio of two estimated multi-channel power spectra and a pointwise'
        write (prtunit,*) ' tolerance interval for this multi-channel spectral ratio under the assumption that the two "true"'
        write (prtunit,*) ' underlying multi-channel power spectra are the same'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_spectrum_ratio_3d  '
        write (prtunit,*) '        -nf=numerator_netcdf_file'
        write (prtunit,*) '        -nv=numerator_netcdf_variable'
        write (prtunit,*) '        -dv=denominator_netcdf_variable'
        write (prtunit,*) '        -df=denominator_netcdf_file       (optional)'
        write (prtunit,*) '        -m=input_mesh_mask_netcdf_file    (optional)'
        write (prtunit,*) '        -g=grid_type                      (optional : n, t, u, v, w, f)'
        write (prtunit,*) '        -x=lon1,lon2                      (optional)'
        write (prtunit,*) '        -y=lat1,lat2                      (optional)'
        write (prtunit,*) '        -np=numerator_period              (optional)'
        write (prtunit,*) '        -dp=denominator_period            (optional)'
        write (prtunit,*) '        -t=freq1,freq2                    (optional)'
        write (prtunit,*) '        -o=output_netcdf_file             (optional)'
        write (prtunit,*) '        -pro=probability_interval         (optional : 0. > 1.)'
        write (prtunit,*) '        -mi=missing_value                 (optional)'
        write (prtunit,*) '        -logratio                         (optional)'
        write (prtunit,*) '        -double                           (optional)'
#ifdef _USE_NETCDF36
        write (prtunit,*) '        -bigfile                          (optional)'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                             (optional)'
#endif
        write (prtunit,*) '        -tlimited                         (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-df=         : by default, the denominator_netcdf_file is the same as the numerator_netcdf_file'
        write (prtunit,*) '-m=          : by default, an input_mesh_mask_netcdf_file is not used'
        write (prtunit,*) '-g=n         : by default, a regular or gaussian grid is assumed'
        write (prtunit,*) '-x=          : the whole longitude domain associated with the two input NetCDF variables'
        write (prtunit,*) '-y=          : the whole latitude domain associated with the two input NetCDF variables'
        write (prtunit,*) '-np=         : by default, the numerator_period argument is not used'
        write (prtunit,*) '-dp=         : by default, the denominator_period argument is not used'
        write (prtunit,*) '-t=the whole set of frequencies included in the two input NetCDF files'
        write (prtunit,*) '-o=spectrum_ratio_numerator_netcdf_variable.denominator_netcdf_variable.nc'
        write (prtunit,*) '-pro=0.90    : by default, a 90% tolerance interval is computed'
        write (prtunit,*) '-mi=1.e+20   : by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-logratio    : by default, the ratios are not Log transformed'
        write (prtunit,*) '-double      : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF36
        write (prtunit,*) '-bigfile     : by default, a NetCDF classical format file is created'
#endif
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5        : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited    : by default, the time (e.g., frequency) dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<3 .or. narg>19 ) then
        call merror( name_proc//"Wrong number of line arguments" )
    end if
!
    fic_name(:) = " "
    var_name(:) = " "
    mask_name   = " "
    ratio_name  = " "
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode  = nf90_clobber
    output_type  = nf90_real4
!
    probtol = 0.90_stnd
!
    missing = def_missing
!
    grid_type = "n"
!
    lon1  = 1
    lon2  = huge( lon2 )
    lat1  = 1
    lat2  = huge( lat2 )
    freq1 = 1
    freq2 = huge( freq2 )
!
    period(:2)     = -1
    use_period(:2) = false
!
    logratio     = false
    use_tlimited = false
    use_mask     = false
    use_x        = false
    use_y        = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-g=')
!
!               GET THE GRID-TYPE.
!
                call case_change( name(4:), tolower )
!
                select case ( name(4:) )
                    case ( 'n', 'u', 'v', 'w', 't', 'f' )
                        grid_type = name(4:4)
                    case default
                        call merror( name_proc//"Unknown grid type" )
                end select
!
            case ('-x=')
!
!               GET THE DOMAIN DEFINITION (Longitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lon1 and lon2 are integer values",     &
                                  "Wrong domain definition (longitude)",  &
                                  lon1, lon2 )
!
                use_x = true
!
            case ('-y=')
!
!               GET THE DOMAIN DEFINITION (latitude).
!
                call get_listarg( name_proc, name(4:),                    &
                                  "lat1 and lat2 are integer values",     &
                                  "Wrong domain definition (latitude)",   &
                                  lat1, lat2 )
!
                use_y = true
!
            case ('-t=')
!
!               GET THE FREQUENCY INTERVAL.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "freq1 and freq2 are integer values",   &
                                  "Wrong frequency interval",             &
                                  freq1, freq2 )
!
            case ('-m=')
!
!               GET THE NAME OF THE MESH-MASK NETCDF DATASET.
!
                mask_name = name(4:)
!
                use_mask = true
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                ratio_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-nf=')
!
!                       GET THE NAME OF THE NUMERATOR NETCDF DATASET.
!
                        fic_name(1) = name(5:)
!
                    case ('-nv=')
!
!                       GET THE NAME OF THE NUMERATOR NETCDF VARIABLE.
!
                        var_name(1) = name(5:)
!
                    case ('-np=')
!
!                       GET THE NUMERATOR PERIOD.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kint ) &
                        call merror( name_proc//"The numerator_period is an integer value" )
!
                        read( name(5:), fmt ) period(1)
!
                        if ( period(1)<=0 ) &
                        call merror( name_proc//"The -np= parameter is a positive integer" )
!
                        use_period(1) = true
!
                    case ('-df=')
!
!                       GET THE NAME OF THE DENOMINATOR NETCDF DATASET.
!
                        fic_name(2) = name(5:)
!
                    case ('-dv=')
!
!                       GET THE NAME OF THE DENOMINATOR NETCDF VARIABLE.
!
                        var_name(2) = name(5:)
!
                    case ('-dp=')
!
!                       GET THE DENOMINATOR PERIOD.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kint ) &
                        call merror( name_proc//"The denominator_period is an integer value" )
!
                        read( name(5:), fmt ) period(2)
!
                        if ( period(2)<=0 ) &
                        call merror( name_proc//"The -dp= parameter is a positive integer" )
!
                        use_period(2) = true
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                        call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
                        select case ( name(1:5) )
!
                            case ('-pro=')
!
!                               DETERMINE THE PROBABILITY FOR THE TOLERANCE INTERVAL.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kfix .or. kcode==kexp ) then
                                    read( name(6:), fmt ) probtol
                                end if
                                if ( (kcode/=kfix .and. kcode/=kexp) .or.     &
                                     probtol<=zero                   .or.     &
                                     probtol>=one )                           &
      call merror( name_proc//"The -pro= argument must be a real number and verify 0.<pro<1. " )
!
                            case default
!
                                select case ( name(1:) )
!
                                    case ('-logratio')
!
!                                       COMPUTE THE LOG RATIOS.
!
                                        logratio = true
!
#ifdef _USE_NETCDF36
                                    case ('-bigfile')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_64bit_offset
#endif
!
#ifdef _USE_NETCDF4
                                    case ('-hdf5')
!
!                                       REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                        output_mode = nf90_hdf5
#endif
!
                                    case ('-tlimited')
!
!                                       DEFINE THE TIME (E.G., FREQUENCY) DIMENSION AS LIMITED.
!
                                        use_tlimited = true
!
                                    case ('-double')
!
!                                       STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                        output_type = nf90_real8
!
                                    case default
!
!                                       WRONG ARGUMENT.
!
                                        call merror( name_proc//"Wrong argument" )
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name(1)==" " )                                                        &
    call merror( name_proc//"Input numerator NETCDF file is missing" )
!
    if ( fic_name(2)==" " )   fic_name(2) = fic_name(1)
!
    if ( var_name(1)==" " )                                                        &
    call merror( name_proc//"Input numerator NETCDF variable is missing" )
!
    if ( var_name(2)==" " )                                                        &
    call merror( name_proc//"Input denominator NETCDF variable is missing" )
!
    if ( ratio_name==" " )                                                         &
    ratio_name="spectrum_ratio_"//trim(var_name(1))//"."//trim(var_name(2))//".nc"
!
!   OPEN THE NETCDF DATASETS.
!
    if ( use_mask ) then
        status = nf90_open( trim(mask_name), nf90_nowrite, ncid_mask )
        if ( status/=nf90_noerr )                                            &
        call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    nfic = merge( 1, 2, fic_name(1)==fic_name(2) )
!
    name = "(numerator)"
!
    do j = 1, 2
!
        if ( j<=nfic ) then
!
            i = nf90_sizehint_default
            status = nf90_open( trim(fic_name(j)), nf90_nowrite, ncid(j), chunksize=i )
!
            if ( status/=nf90_noerr )                                                &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           GET THE VARIABLES ID FROM THEIR NAME.
!
!           FIND THE DIMENSION LENGTH AND ID FOR THE NETCDF VARIABLES edof
!           AND frequency IN THE NETCDF FILES.
!
            call get_variable_dims(  name_proc, "frequency", ncid(j), freqid(j),        &
                                     dimids_out=dimids, dimlens_out=dimlens  )
!
            ndim = count(dimlens/=0)
            if ( ndim/=1  )     &
            call merror( name_proc//"Input NETCDF variable frequency "//trim(name)//" is not unidimensional" )
!
            nfreq(j) = dimlens(1)
            i        = dimids(1)
!
            call get_variable_dims(  name_proc, "edof", ncid(j), dofid(j),      &
                                     dimids_out=dimids, dimlens_out=dimlens     )
!
            ndim = count(dimlens/=0)
            if ( ndim/=1  )     &
            call merror( name_proc//"Input NETCDF variable edof "//trim(name)//" is not unidimensional" )
!
            if ( dimids(1)/=i  )     &
            call merror( name_proc//"Incorrect "//trim(name)//" NETCDF file" )
!
        else
!
!           THE TWO INPUT NETCDF DATASETS ARE IDENTICAL.
!
            ncid(2)  = ncid(1)
            nfreq(2) = nfreq(1)
!
        end if
!
        if ( j==1 ) then
!
!           GET THE VARIABLE ID FROM ITS NAME.
!           FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!           GET THE COORDINATE VARIABLES IDS FROM THEIR NAMES, IF ANY.
!           FIND THE EXTERNAL TYPE, THE DIMENSION ID AND THE NUMBER OF ATTRIBUTES 
!           FOR THE COORDINATE VARIABLES.
!
            call get_variable_dimcoords(  name_proc, trim(var_name(j)), ncid(j),                       &
                                          varid(j),  dimids_out=dimids, dim_names_out=dim_names,       &
                                          dimlens_out=dimlens, coor_id_out=coor_id,                    &
                                          coor_natts_out=coor_natts, coor_type_out=coor_type,          &
                                          use_coor_variable_out=use_coor_variable                      )
!
            nlon   = dimlens(1)
            nlat   = dimlens(2)
!
        else
!
!           GET THE VARIABLE IDS FROM THEIR NAME.
!           FIND THE DIMENSION LENGTHS AND IDS FOR THE NETCDF VARIABLES.
!
            call get_variable_dims(  name_proc, trim(var_name(j)), ncid(j),                 &
                                     varid(j),  dimids_out=dimids, dimlens_out=dimlens      )
!
            if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
            call merror( name_proc//"Shape disagreement between the two input NETCDF variables" )
!
        end if
!
        ndim = count(dimlens/=0)
!
        if ( use_period(j) ) then
!
            if ( ndim/=4  )     &
            call merror( name_proc//"Input NETCDF variable "//trim(name)//" is not fourdimensional" )
!
!           CHECK THE PERIODS.
!
            if ( period(j)>dimlens(4) )      &
            call merror( name_proc//"period "//trim(name)//" is out of bounds" )
!
        else
!
            if ( ndim/=3  )     &
            call merror( name_proc//"Input NETCDF variable "//trim(name)//" is not tridimensional" )
!
        end if
!
        if ( dimids(3)/=i )     &
        call merror( name_proc//"Incorrect "//trim(name)//" NETCDF file" )
!
        name = "(denominator)"
!
    end do
!
    if ( nfreq(1)/=nfreq(2) )        &
    call merror( name_proc//"Frequencies are not the same in the two input NETCDF files" )
!
    nf = nfreq(1)
!
    if ( use_mask ) then
!
        chr = merge( 't', grid_type, grid_type=='w' )
!
!       GET THE MASK ID FROM ITS NAME.
!       VERIFY THE DIMENSION LENGTHS FOR THE NETCDF MASK VARIABLE.
!
        call get_variable_dims( name_proc, chr//"mask", ncid_mask, maskid, dimids, dimlens )
!
        if ( dimlens(1)/=nlon .or. dimlens(2)/=nlat )                      &
        call merror( name_proc//"Shape disagreement with the mask" )
!
        if ( .not.use_x ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lon1_Eastern_limit", i )
            if ( status==nf90_noerr ) lon1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lon2_Western_limit", i )
            if ( status==nf90_noerr ) lon2 = i
!
        end if
!
        if ( .not.use_y ) then
!
            status = nf90_get_att( ncid_mask, maskid, "lat1_Southern_limit", i )
            if ( status==nf90_noerr ) lat1 = i
!
            status = nf90_get_att( ncid_mask, maskid, "lat2_Northern_limit", i )
            if ( status==nf90_noerr ) lat2 = i
!
        end if
!
    end if
!
!   CHECK THE BOUNDARIES OF THE GEOGRAPHICAL DOMAIN.
!
    if ( lon1<0 ) then
        lon2 = max( lon2, 1 )
        lon3 = nlon + lon1 - lon2 + 1
    else
        lon1 = max( lon1, 1 )
        lon2 = min( lon2, nlon )
        lon3 = lon2 - lon1 + 1
    end if
!
    if ( lon3<=0  )  call merror( name_proc//"Wrong domain definition (longitude)" )
!
    if ( lon1<0 ) then
       lon3 = nlon
       j1   = 1
       j2   = nlon
    else
       j1 = lon1
       j2 = lon2
    end if
!
    lat1 = max( lat1, 1 )
    lat2 = min( lat2, nlat )
    lat3 = lat2 - lat1 + 1
    if ( lat3<=0  )  call merror( name_proc//"Wrong domain definition (latitude)" )
!
!   CHECK THE FREQUENCY INTERVAL.
!
    freq1 = max( freq1, 1 )
    freq2 = min( freq2, nf )
    freq3 = freq2 - freq1 + 1
!
!   COMPUTE THE GEOGRAPHICAL MASK.
!
    if ( use_mask ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( temp(nlon,nlat), mask(nlon,nlat), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       GET THE MASK.
!
        status = nf90_get_var( ncid_mask, maskid, temp )     
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!       DETERMINE THE GRID POINTS USED IN THE CORRELATION ANALYSIS FROM THE MASK.
!
        mask(:nlon,:nlat) = temp(:nlon,:nlat)==one
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( temp )
!
    else
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( mask(nlon,nlat), stat=status )
        if ( status/=0 ) call merror( name_proc//allocate_error )
!
!       CONSTRUCT A DUMMY MASK.
!
        mask(:nlon,:nlat) = true
!
    end if
!
!   SELECT THE DOMAIN.
!
    if ( lon1>0 ) then
        mask(1:lon1-1,:nlat)    = false
        mask(lon2+1:nlon,:nlat) = false
    else
        mask(lon2+1:nlon+lon1,:nlat) = false
    end if
!
    mask(:nlon,1:lat1-1)    = false
    mask(:nlon,lat2+1:nlat) = false
!   
    nvar = count( mask )
    if ( nvar<=0  )  call merror( name_proc//"Wrong domain definition : domain is empty" )
!
!   ALLOCATE WORK ARRAYS FOR THE POWER SPECTRUM, DEGREES OF FREEDOM AND FREQUENCY AXIS.
!
    allocate( spect(lon3,lat3,2), spect_ratio(lon3,lat3,freq3),                                &
              lwr_ratio(freq3), upr_ratio(freq3), freq(freq3,2), edof(freq3,2), stat=status    )
!
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE DATA.
!
    do j = 1, 2
!
        if ( j<=nfic ) then
!
!           GET THE FREQUENCY AXES.
!
            status = nf90_get_var( ncid(j), freqid(j), freq(:freq3,j), start=(/ freq1 /), count=(/ freq3 /)   )
!
            if ( status/=nf90_noerr )                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           GET THE DEGREES OF FREEDOM FOR THE TWO POWER SPECTRA.
!
            status = nf90_get_var( ncid(j), dofid(j), edof(:freq3,j), start=(/ freq1 /), count=(/ freq3 /)   )
!
            if ( status/=nf90_noerr )                                     &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        else
!
            freq(:freq3,2) = freq(:freq3,1)
            edof(:freq3,2) = edof(:freq3,1)
!
        end if
!
    end do
!
    if ( nfic/=1 ) then
!
        if ( any( freq(:freq3,1)/=freq(:freq3,2) )   )       &
        call merror( name_proc//"Frequencies are not the same in the two input NETCDF files" )
!
    end if
!
!   COMPUTE THE UPPER AND LOWER CRITICAL SPECTRAL RATIOS.
!
    call spctrm_ratio( edof(:freq3,1), edof(:freq3,2), lwr_ratio(:freq3),    &
                       upr_ratio(:freq3), pinterval=probtol                  )
!
    if ( logratio ) then
!
!       COMPUTE THE LOG OF THE TWO CRITICAL SPECTRAL RATIOS IF REQUIRED.
!
        where( lwr_ratio(:freq3)>zero )
            lwr_ratio( :freq3) = log( lwr_ratio( :freq3) )
        elsewhere
            lwr_ratio( :freq3) = missing
        end where
!
        where( upr_ratio(:freq3)>zero )
            upr_ratio( :freq3) = log( upr_ratio( :freq3) )
        elsewhere
            upr_ratio( :freq3) = missing
        end where        
!
    end if
!
!   GET THE DATA.
!
!   COMPUTE THE OFFSET FOR THE RECORD NUMBER.
!
    offset_rec = freq1 - 1
!
    do j = 1, 2
!
        startlec(1:4,j)  = (/ j1,   lat1, 1_i4b, 1_i4b  /)
        countlec(1:4,j)  = (/ lon3, lat3, 1_i4b, 1_i4b   /)
!
        if ( use_period(j) ) then
            startlec(4,j)  = period(j)
            ndimlec(j)     = 4
        else
            ndimlec(j)     = 3
        end if
!
    end do
!
    do k = freq1, freq2
!
        do j = 1, 2
!
            startlec(3,j) = k
            ndim          = ndimlec(j)
!
!           FILL THE WORK ARRAY WITH A RECORD FROM THE NETCDF VARIABLE.
!
            status = nf90_get_var( ncid(j), varid(j), spect(:lon3,:lat3,j),           &   
                                   start=startlec(:ndim,j), count=countlec(:ndim,j)   )     
!
            if ( status/=nf90_noerr )                                       &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
        rec = k - offset_rec
!
        if ( logratio ) then
!
!           COMPUTE THE LOG SPECTRAL RATIO OF THE TWO MULTI-CHANNEL PSD ESTIMATES.
!
            where( mask(j1:j2,lat1:lat2) .and. spect(:lon3,:lat3,2)>zero .and. spect(:lon3,:lat3,1)>zero )
                spect_ratio(:lon3,:lat3,rec) = log( spect(:lon3,:lat3,1)/spect(:lon3,:lat3,2) )
            elsewhere
                spect_ratio(:lon3,:lat3,rec) = missing
            end where
!
        else
!
!           COMPUTE THE SPECTRAL RATIO OF THE TWO MULTI-CHANNEL PSD ESTIMATES.
!
            where( mask(j1:j2,lat1:lat2) .and. spect(:lon3,:lat3,2)/=zero )
                 spect_ratio(:lon3,:lat3,rec) = spect(:lon3,:lat3,1)/spect(:lon3,:lat3,2)
            elsewhere
                 spect_ratio(:lon3,:lat3,rec) = missing
            end where
!
        end if
!
    end do
!
!   CREATE THE NEW NETCDF DATASET FOR THE FREQUENCY ANALYSIS.
!
    status = nf90_create( ratio_name, cmode=output_mode, ncid=ncid_ratio )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_ratio, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    dimlens(1)   = nlon
    dimlens(2)   = nlat
    if ( use_tlimited ) then
        dimlens(3)   = freq3
    else
        dimlens(3)   = nf90_unlimited
    end if
!
    coor_type(3) = output_type
    dim_names(3) = "frequency"
!
    do i = 1, 3
        status = nf90_def_dim( ncid_ratio, name=trim(dim_names(i)),                         &
                               len=dimlens(i), dimid=dimids(i) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
    end do
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    do i = 1, 3
        status = nf90_def_var( ncid_ratio, trim(dim_names(i)), coor_type(i),              &
                               dimids=dimids(i), varid=coor_id2(i) )
        if ( status/=nf90_noerr )                                                         &
        call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end do
!
    status = nf90_def_var( ncid_ratio, "periodicity",  output_type,                   &
                           dimids=dimids(3), varid=periodid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( logratio ) then
        name = "logratio"
    else
        name = "ratio"
    end if
!
    status = nf90_def_var( ncid_ratio, "lower_critical_"//trim(name),  output_type,      &
                           dimids=dimids(3), varid=loratio_id     )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_ratio, "upper_critical_"//trim(name),  output_type,      &
                           dimids=dimids(3), varid=upratio_id  )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    name2 = trim(var_name(1))//"_"//trim(var_name(2))//"_"//trim(name)
!
    status = nf90_def_var( ncid_ratio, trim(name2),  output_type,   &
                           dimids=dimids(1:3), varid=ratio_id )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    do j = 1, 3
        if ( .not.use_coor_variable(j) ) cycle
        do i = 1, coor_natts(j)
            status = nf90_inq_attname( ncid(1), coor_id(j), i, att_name )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!           COPY ATTRIBUTES FOR THE COORDINATE VARIABLES.
!
            status = nf90_copy_att(  ncid(1), coor_id(j), att_name, ncid_ratio, coor_id2(j) )
            if ( status/=nf90_noerr )                                                      &
            call merror( name_proc//trim( nf90_strerror( status ) ) )
        end do
    end do
!
    status = nf90_put_att( ncid_ratio, periodid, "long_name", "Periodicity axis" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, periodid, "short_name", "Periodicity" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, periodid, "units", "Number of observations" )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, loratio_id, "long_name",                           &
                               "Lower probability_interval*100% critical "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, loratio_id, "short_name", "Lower critical "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, loratio_id, "units", "none"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, upratio_id, "long_name",                           &
                               "Upper probability_interval*100% critical "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, upratio_id, "short_name", "Upper critical "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, upratio_id, "units", "none"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    if ( logratio ) then
        name2 = "Log ratio of Power Spectral Density estimates of"
    else
        name2 = "Ratio of Power Spectral Density estimates of"
    end if
!
    name2 = trim(name2)//" "//trim(var_name(1))//" and "//trim(var_name(2))
!
    status = nf90_put_att( ncid_ratio, ratio_id, "long_name", trim(name2)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, ratio_id, "short_name", "PSD "//trim(name)  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, ratio_id, "units", "none"  )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    vecid(:4) = (/ coor_id2(3), loratio_id, upratio_id, ratio_id  /)
!
!   SET missing_value AND _FillValue ATTRIBUTES.
!
    if (  output_type==nf90_real8 ) then
!
        do j=1, 4
!
            status = nf90_put_att( ncid_ratio, vecid(j), "missing_value", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_ratio, vecid(j), "_FillValue", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    else
!
        do j=1, 4
!
            status = nf90_put_att( ncid_ratio, vecid(j), "missing_value", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_ratio, vecid(j), "_FillValue", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        end do
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    name2 = "Power spectrum ratio analysis of "//trim(var_name(1))//" (numerator) and "//   &
            trim(var_name(2))//" (denominator)"
!
    status = nf90_put_att( ncid_ratio, nf90_global, "title", trim(name2) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, nf90_global, "grid_type", grid_type )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_ratio, nf90_global, "probability_interval", probtol )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( logratio ) then
        i = 1
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_ratio, nf90_global, "log_transform", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    name = "numerator"
!
    do j = 1, 2
!
        if ( use_period(j) ) then
!
            status = nf90_put_att( ncid_ratio, nf90_global, trim(name)//"_period", period(j) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        end if            
!
        name = "denominator"
!
    end do
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_ratio )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
    i = max( nlon, nlat )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( record(i), temp(nlon,nlat), stat=status )
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
    do j = 1, 2
!
        i = dimlens(j)
!
        if ( use_coor_variable(j) ) then
!
            status = nf90_get_var( ncid(1), coor_id(j), record(:i), count=(/ i /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_var( ncid_ratio, coor_id2(j), record(:i), count=(/ i /) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        else
            status = nf90_put_var( ncid_ratio, coor_id2(j), arth(1_i4b,1_i4b,i) )
            if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end do
!
!   WRITE FREQUENCY AND PERIOD AXES.
!
    status = nf90_put_var( ncid_ratio, coor_id2(3), freq(:freq3,1) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   BUILD UP PERIODICITY VECTOR.
!
    where( freq(:freq3,1)/=zero )
        freq(:freq3,2) = one/freq(:freq3,1)
    elsewhere
        freq(:freq3,2) = missing
    end where
!
    status = nf90_put_var( ncid_ratio, periodid, freq(:freq3,2) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE THE CRITICAL SPECTRAL RATIOS.
!
    status = nf90_put_var( ncid_ratio, loratio_id, lwr_ratio( :freq3) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_var( ncid_ratio, upratio_id, upr_ratio( :freq3) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE THE SPECTRAL RATIO OF THE TWO PSD ESTIMATES.
!
    startwri(1:2) = 1
    countwri(1:3) = (/ nlon, nlat, 1_i4b /)
!
    temp(:nlon,:nlat ) = missing
!
    do i = 1, freq3
!
        startwri(3) = i
!
        temp(j1:j2,lat1:lat2) = spect_ratio(:lon3,:lat3,i)
!
        status = nf90_put_var( ncid_ratio, ratio_id, temp(:nlon,:nlat), start=startwri, count=countwri )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( mask, temp, record, spect, spect_ratio, lwr_ratio, upr_ratio, freq, edof  )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid(1) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( ncid(1)/=ncid(2) ) then
        status = nf90_close( ncid(2) )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( use_mask ) then
        status = nf90_close( ncid_mask )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    status = nf90_close( ncid_ratio )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!
! END OF PROGRAM comp_spectrum_ratio_3d
! =====================================
!
end program comp_spectrum_ratio_3d
