program comp_spectvar_1d
!! #############################################################################
!!
!! Copyright 2024 IRD
!!
!! This file is part of NCSTAT.
!!
!! NCSTAT is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!
!! NCSTAT is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the NCSTAT/doc directory.
!!
!! #############################################################################
!
!+
!
! .. _comp_spectvar_1d:
!
! ================
! comp_spectvar_1d
! ================
!
!
! Authors
! =======
!
! Pascal Terray (LOCEAN/IPSL)
!
! Latest revision
! ===============
!
! 28/05/2024
!
! Purpose
! =======
!
! Compute variance estimates in a selected frequency band for a real time series
! by windowed filtering in the frequency domain. These variance estimates correspond exactly
! to the variance of the filtered time series output by :ref:`comp_fftfilter_1d`
! when similar parameters are used in both procedures. Alternatively, comp_spectvar_1d can
! also compute spectral variance estimates in a given frequency band from a spectrum analysis of a real
! time series as output by :ref:`comp_spectrum_1d` or :ref:`comp_cross_spectrum_1d`.
! The input time series is extracted from a uni- or bidimensional variable readed from
! a NetCDF dataset.
!
! At the user option, the time series can be detrended before computing the power spectrum and the
! variance estimate in the selected frequency band (see the **-tr=** and **-tr2=** arguments). The variance estimate
! and power spectrum are computed with the help of the Fast Fourier Transform (FFT) of the real time series and
! windowed filtering in the frequency domain.
!
! The user can select different data windows to apply to the time series for the computations of
! the Power Spectral Density (PSD) estimates (see the **-win2=** argument) and also  different frequency  windows to convolve
! with the selected filter response in the frequency domain (see the **-win=** argument) in order to get more accurate
! variance estimates.
!
! Furthermore, in order to obtain stable PSD and variance estimates, the real time series can be segmented and PSD estimates computed
! on, eventually overlapping, segments can be averaged at the user option (see the **-sl=** and **-nooverlap=** arguments).
! The stability of the PSD and final variance estimates is increased by this averaging process.
! That is, the greater the number of segments, the more stable, the resulting PSD and final variance estimates. See the documentation
! of :ref:`comp_spectrum_1d` or :ref:`comp_cross_spectrum_1d` for more details on this method known as the Welch's method  [Welch]_ .
!
! Optionally, theses PSD estimates may also been smoothed in the frequency
! domain by Daniell weights (e.g., a simple moving average; see the **-nsm=** argument) as in a conventional
! spectrum analysis.
!
! By default, the variance estimate is returned in units, which are the square of the data, and, at the user option, as
! percentage of the total variance in the selected frequency band if the **-normvar** argument is specified
! in the call to the procedure.
!
! Finally, the variances estimates can be computed separately for different times intervals (see the **-p=** argument).
!
! The variance estimates are stored in an output NetCDF dataset.
!
! For more information on spectral analysis and filtering in the frequency domain, see the references
! below  [Welch]_  [Bloomfield]_  [Diggle]_  [Iacobucci_Noullez]_ .
!                                                                              
! Further Details
! ===============
!
! Usage
! -----
!
! ::
!
!       $ comp_spectvar_1d \
!              -f=input_netcdf_file \
!              -v=netcdf_variable \
!              -t=time1,time2                           (optional) \
!              -o=output_netcdf_file                    (optional) \
!              -ni=index_for_2d_netcdf_variable         (optional) \
!              -p=period_length                         (optional) \
!              -sl=segment_length                       (optional : 16,32,64,128,...) \
!              -pl=minimum_period                       (optional) \
!              -ph=maximum_period                       (optional) \
!              -tr=trend_removal                        (optional : 0,1,2,3) \
!              -tr2=trend_removal_on_segment            (optional : 0,1,2,3) \
!              -win=window_parameter                    (optional : 0.5 > 1.) \
!              -win2=window_choice                      (optional : 1,2,3,4,5,6) \
!              -tap=tapered_percentage                  (optional : 0.0 > 1.) \
!              -nsm=Daniell_filter_length               (optional) \
!              -mi=missing_value                        (optional) \
!              -nooverlap                               (optional) \
!              -normvar                                 (optional) \
!              -double                                  (optional) \
!              -hdf5                                    (optional) \
!              -tlimited                                (optional)
!
! By default
! ----------
!
!        *-t=*
!         the whole time period associated with the *netcdf_variable*
!        *-o=*
!         the *output_netcdf_file* is named ``spectvar_``\ *netcdf_variable*\ ``.nc``
!        *-ni=*
!         if the *netcdf_variable* is bidimensional, the first time series is used
!        *-p=*
!         the *period_length* is set to ``time2 - time1 + 1`` , which means
!         that the time series is considered as continuous with only one time period
!        *-sl=*
!         the *segment_length* is set to ``0``, which means that the time series (or each
!         time period if the **-p=** argument is used) is not segmented to stabilize the
!         the PSD and final variance estimates
!        *-pl=*
!         the *minimum_period* is set to ``0``, which means that no filtering is done
!         for the shorter periods
!        *-ph=*
!         the *maximum_period* is set to ``0``, which means that no filtering is done
!         for the longer periods
!        *-tr=*
!         the *trend_removal* is set to ``1``, which means that the mean of the whole time series
!         is removed before the spectral computations
!        *-tr2=*
!         the *trend_removal_on_segment* is set to ``0``, which means that no trend processing is
!         done on each time segment before the spectral computations
!        *-win=*
!         the *window_parameter* is set to ``1.``, which means that the rectangular window is convolved
!         with the ideal filter response in the frequency domain in order to compute the final variance estimates
!        *-win2=*
!         the *window_choice* is set to ``2``, which means that the square window is applied to the time series
!         and used in the spectral computations
!        *-tap=*
!         the *tapered_percentage* is set to ``0.2``, which means that if a split-cosine-bell window
!         is used for the spectral computations (e.g., if the *window_choice* is set to ``6`` with the
!         **-win2=** argument), ``20``\ % of the data are tapered
!        *-nsm=*
!         the *Daniell_filter_length* is set to ``0``, which means that the PSD estimates are not
!         smoothed in the frequency domain by a moving average before computing the variance estimates
!         in the selected frequency band
!        *-mi=*
!         the *missing_value* for the output NetCDF variable is set to ``1.e+20``
!        *-nooverlap*
!         normally, if the **-sl=** argument is used, the time segments are overlapped. However,
!         if **-nooverlap** is specified, the time segments are not overlapped
!        *-normvar*
!         normally, the raw variance in the selected frequency band is output. However,
!         if **-normvar** is specified, the percentage of total variance in the selected
!         frequency band for the real multichannel time series is also computed and stored
!         in the output NetCDF file.
!        *-double*
!         the results are stored as single-precision floating
!         point numbers in the output NetCDF file. If **-double** is activated, the
!         results are stored as double-precision floating point numbers 
!        *-hdf5*
!         a NetCDF classical format file is created. If **-hdf5** is activated,
!         the output NetCDF file is a NetCDF-4/HDF5 format file
!        *-tlimited*
!         the time dimension is defined as unlimited in the output NetCDF file.
!         However, if **-tlimited** is activated, the time dimension is defined
!         as limited in the output NetCDF file
!
! Remarks
! -------
!
!   #) The **-v=**\ *netcdf_variable* argument specifies the NetCDF variable for which a
!      frequency variance analysis must be performed and the **-f=**\ *input_netcdf_file* 
!      argument specifies that this NetCDF variable must be extracted from the NetCDF file
!      *input_netcdf_file*.
!
!   #) If the **-t=**\ *time1,time2* argument is missing the whole time period
!      associated with the *netcdf_variable* is used to compute the spectral variance
!      estimates in the selected frequency band.
!
!      The selected time period is a vector of two integers specifying the first
!      and last time observations. The indices are relative to ``1``\ .
!
!      *time2* - *time1* + ``1`` must be an even integer if the **-sl=** and **-p=** arguments are not used
!      and of any length if one of these arguments is used.
!
!   #) The **-ni=**\ *index_for_2d_netcdf_variable* argument specifies the index for selecting the
!      time series if the *netcdf_variable* is a 2D NetCDF variable. By default, the first time series
!      is used, which is equivalent to set *index_for_2d_netcdf_variable* to ``1``\ .
!
!   #) If the **-p=** argument is specified, the frequency variance estimates are computed separately for
!      each time period of length *period_length* (as determined by the value of the **-p=** argument).
!
!      The whole selected time period (e.g., *time2* - *time1* + ``1`` ) must be a multiple of 
!      the *period_length*.
!      The *period_length* must also be of even length if the **-sl=** argument is not used.
!
!   #) The **-sl=**\ *segment_length* argument specifies an integer used to segment the time series.
!      The *segment_length* must be a positive even integer less or equal to the length of the time series
!      (e.g., *time2* - *time1* + ``1`` ) or the *period_length* if the **-p=** argument is used.
!
!      Spectral computations are at ( *segment_length*\ /\ ``2``\  )  + ``1`` Fourier frequencies and suggested
!      values for *segment_length* are ``16``, ``32``, ``64``, ``128``, ...  (e.g., the spectral estimates are taken
!      at frequencies ``(i-1)``\  / *segment_length* for i=1,2, ... , (*segment_length*\ /\ ``2``\ ) + ``1``.
!
!      By default, *segment_length* is set to ``0``, e.g., the time series is not segmented and the segment length is
!      equal to *time2* - *time1* + ``1`` or *period_length* if the **-p=** argument is used.
!
!   #) The **-pl=** argument specifies the minimum period of oscillation of the filtered time series. 
!      The *minimum_period* is expressed in number of time observations. 
!
!      Do not use the **-pl=** argument or use **-pl=**\ ``0`` for high-pass filtering 
!      frequencies corresponding to periods shorter than **-ph=**\ ``PH`` .
!
!      The **-pl=** argument is a positive integer equal to ``0`` or greater than ``2``.
!
!   #) The **-ph=** argument specifies the maximum period of oscillation of the filtered time series. 
!      The *maximum_period* is expressed in number of time observations. 
!      Do not use the **-ph=** argument or use **-ph=**\ ``0`` for low-pass filtering 
!      frequencies corresponding to periods longer than **-pl=**\ ``PL`` .
!      For example, **-pl=**\ ``6``  (or ``18``) and **-ph=**\ ``32``  (or ``96``) select periods
!      between ``1.5`` and ``8``  years for quarterly (monthly) time series.
!
!      The **-ph=** argument is a positive integer equal to ``0`` or greater than ``2`` and 
!      less than the length of the time series or the *period_length* if the **-p=**
!      argument is used.
!
!      The **-ph=** argument must also be greater or equal to the **-pl=** argument if 
!      both are specified.
!                                                                              
!   #) Setting **-pl=**\ ``PL``, **-ph=**\ ``PH`` and ``PH``<``PL`` is also allowed and
!      performs band rejection of periods between ``PH`` and ``PL``. In that case, the meaning
!      of the **-pl=** and **-ph=** arguments reversed.                  
!
!   #) Setting **-pl=** and **-ph=** to the same value ``P`` is allowed. 
!      In this case, an -ideal- band-pass filter with peak response near one
!      at the single period ``P`` is applied to the time series to get the variance estimates at this particular fequency.
!
!   #) Setting both **-pl=**\ ``0`` and **-ph=**\ ``0`` is not allowed since in that case,
!      no frequency filtering is done.
!
!   #) The **-tr=** argument specifies pre-filtering processing of the real time series before
!      estimating the power spectrum of the time series. If:
!
!      * **-tr=**\ ``+1``, the mean of the time series is removed before the frequency analysis
!
!      * **-tr=**\ ``+2``, the drift from the time series is removed before the frequency analysis.
!        The drift for the time series is estimated using the formula : ``drift = ( tseries(ntime) - tseries(1) )/( ntime - 1 )``
!
!      * **-tr=**\ ``+3``, the least-squares line from the time series is removed before the frequency analysis.
!
!      For other values of the **-tr=** argument, nothing is done before estimating the spectrum.
!
!      If the **-p=** argument is present, the pre-filtering processing is applied to 
!      each time period, separately.
!
!      The **-tr=** argument must be an integer and the default value for the **-tr=** argument is ``0``,
!      e.g., nothing is done before the spectrum computations.
!
!   #) The **-tr2=** argument specifies pre-filtering processing of the real time series before
!      estimating the power spectrum on each time segment (as specified by the **-sl=** argument). If:
!
!      * **-tr2=**\ ``+1``, the mean of the time segment is removed before computing the power spectrum
!        on the segment
!
!      * **-tr2=**\ ``+2``, the drift from the time segment is removed before computing the power spectrum
!        on the segment
!
!      * **-tr2=**\ ``+3``, the least-squares line from the time segment is removed before computing the
!        power spectrum on the segment.
!
!      For other values of the **-tr2=** argument, nothing is done before estimating the spectrum on each
!      time segment.
!
!      The **-tr2=** argument must be an integer and the default value for the **-tr2=** argument is ``0``,
!      e.g., nothing is done before the spectral computations on each time segment.
!
!   #) The **-win=**\ *window_parameter* argument specifies the form of the window which will be convolved
!      with the ideal filter response as in the :ref:`comp_fftfilter_1d` procedure.
!
!      Set **-win=**\ ``0.5`` for using a Hanning window, **-win=**\ ``0.54`` for using a Hamming window or
!      **-win=**\ ``1.`` for rectangular window.
!                                                        
!      The **-win=**\ *window_parameter* argument is a real number greater or equal to ``O.5`` and less or equal to ``1.``.
!
!      By default, rectangular window filtering is used (i.e., **-win=**\ ``1.``).
!
!   #) The **-win2=**\ *window_choice* argument specifies the form of the data window used in the computations
!      of the power spectrum as the **-win=**\ *window_choice* argument in the :ref:`comp_spectrum_1d` procedure. If:
!
!      * **-win2=**\ ``1`` The Bartlett window is used
!
!      * **-win2=**\ ``2`` The square window is used
!
!      * **-win2=**\ ``3`` The Welch window is used
!
!      * **-win2=**\ ``4`` The Hann window is used
!
!      * **-win2=**\ ``5`` The Hamming window is used
!
!      * **-win2=**\ ``6`` A split-cosine-bell window is used.
!
!      The *window_choice* must be an integer between ``1`` and ``6``.
!
!      The default value for the *window_choice* is ``2``, e.g., a square window window is used for the spectral
!      computations.
!
!   #) The **-tap=**\ *tapered_percentage* argument specifies the total percentage of the data to be tapered if
!      the *window_choice* is set to ``6`` with the **-win2=** argument.
!      The *tapered_percentage* is a real number which must be greater than ``0.`` and less or equal to ``1.``.
!
!      The default value for the *tapered_percentage* is ``0.2`` (e.g., ``20``\ % of the data are tapered, ``10``\ % at
!      the start of the series and ``10``\ % at the end).
!
!   #) The **-nsm=**\ *Daniell_filter_length* argument specifies that the PSD estimates must be computed by smoothing
!      the periodogram with Daniell weights (e.g., a simple moving average) as in the :ref:`comp_spectrum_1d` procedure.
!
!      On entry, the **-nsm=** argument gives the length of the Daniell filter to be applied.
!      The *Daniell_filter_length* must be an odd integer, greater than ``2`` and less or equal to ( *segment_length*\ /\ ``2``\  ) + ``1``
!      if the  **-sl=** argument is specified.
!
!      By default, the PSD estimates are not smoothed in the frequency domain.
!
!   #) If the **-nooverlap** argument is specified and the **-sl=** argument is also specified,
!      comp_spectvar_1d segments the time series without any overlapping (in each time period 
!      if the **-p=** argument is used).
!
!      By default, comp_spectvar_1d overlaps the segments by one half of their length (which is
!      equal to the value of the **-sl=** argument).
!
!      In both cases, each segment will be FFT'd, and the resulting periodograms
!      will be averaged together to obtain PSD estimates at the
!      ( *segment_length*\ /\ ``2``\  ) + ``1`` frequencies.
!
!      If the **-sl=** argument is not specified, the **-nooverlap** argument has no effect.           
!
!   #) The **-normvar** argument specifies that the percentage of variance in the selected frequency
!       band must be also computed and output in addition of the raw variance estimates.
!
!   #) The **-mi=**\ *missing_value* argument specifies the missing value indicator for
!      the output variables in the *output_netcdf_file*. 
!
!      If the **-mi=** argument is not specified, the *missing_value* is set
!      to ``1.e+20``.
!
!   #) The **-double** argument specifies that the results are stored as double-precision floating
!      point numbers in the output NetCDF file.
!
!      By default, the results are stored as single-precision floating point numbers
!      in the output NetCDF file.
!
!   #) The **-hdf5** argument is allowed only if the NCSTAT software has been
!      compiled with the _USE_NETCDF4 CPP macro (e.g., ``-D_USE_NETCDF4``) and linked
!      to the NetCDF 4 library or higher. 
!
!      If this argument is specified, the *output_netcdf_file* will be a NetCDF-4/HDF5
!      format file instead of a NetCDF classic format file. However, this
!      argument is recognized in the procedure only if the NCSTAT software has
!      been built with the _USE_NETCDF4 CPP macro.
!
!   #) It is assumed that the data has no missing values.
!
!   #) Duplicate parameters are allowed, but this is always the last occurrence
!      of a parameter which will be used for the computations. Moreover, the
!      number of specified parameters must not be greater than the total number
!      of allowed parameters.
!
!   #) For more details on power spectrum analysis and on windowed filtering in the frequency domain, see
!
!      * "The use of Fast Fourier Transform for the estimation of power spectra: A method based on time averaging over short, modified periodograms",
!        by P.D. Welch, IEEE trans. on audio and electroacoustics, 15:2, 70-73, 1967. DOI: https://doi.org/10.1109/TAU.1967.1161901
!
!      * "Fourier analysis of time series- An introduction",
!        by Bloomfield, P., John Wiley and Sons, New York, Chapter 6, 1976. http://as.wiley.com/WileyCDA/WileyTitle/productCd-0471889482.html
!
!      * "Time series: a biostatistical introduction",
!        by P.J. Diggle, Clarendon Press, Oxford, Chap. 4, 1990. http://www.oupcanada.com/catalog/9780198522263.html
!                                                                              
!      * "A Frequency Selective Filter for Short-Length Time Series", 
!        by Iacobucci, A., and Noullez, A., Computational Economics, Vol. 25, 75-102, 2005. https://link.springer.com/article/10.1007/s10614-005-6276-7
!
! Outputs
! -------
!
!   comp_spectvar_1d creates an output NetCDF file that contains the variance estimates in the selected frequency band
!   for the time series associated with the input NetCDF variable. The output NetCDF dataset contains the
!   following NetCDF variables (in the description below, nperiod is ``1`` if the **-p=** argument is not used or is equal
!   to ``ntime``\  / *period_length* if the **-p=** argument is used, where ``ntime``\  = *time2* - *time1* + ``1`` ) :
!
!   #) *netcdf_variable*\ _spectvar\ ``(nperiod)`` : the corresponding variance estimates
!      for each time period associated with the input NetCDF variable.
!
!   #) *netcdf_variable*\ _spectvar_percent\ ``(nperiod)`` : the corresponding percentages of total variance
!      for each time period associated with the input NetCDF variable.
!
!      This variable is stored only if the **-normvar** argument has been specified when calling
!      comp_spectvar_1d.
!                                                                              
! Examples
! --------
!
!   #)   For estimating variance in the biennial time scale (e.g., between ``18`` and ``30``months)
!        for a real monthly time series readed from a unidimensional NetCDF variable called ``sst``
!        extracted from the file ``sst.monthly.nino34.nc``, which includes a monthly time
!        series, and store the results in the NetCDF file ``var_tbo_sst_nino34.nc``, use the following
!        command :
!                                                                              
!        ::
!
!          $ comp_spectvar_1d  \
!            -f=sst.monthly.nino34.nc \
!            -v=sst \
!            -pl=18 \
!            -ph=30 \
!            -o=var_tbo_sst_nino34.nc \
!            -normvar
!-
! ================================================================================================
!
!
! USED MODULES 
! ============
!
#ifdef _USE_NAGWARE
    use f90_unix_env,     only : iargc, getarg
#endif
!
#ifdef _USE_NETCDF4
    use netcdf,           only : nf90_hdf5
#endif
!   
    use netcdf,   only : nf90_max_name, nf90_max_var_dims, nf90_clobber, nf90_real4, nf90_real8,   &
                         nf90_int, nf90_sizehint_default, nf90_open,                               &
                         nf90_nowrite, nf90_noerr, nf90_get_att, nf90_get_var, nf90_strerror,      &
                         nf90_create, nf90_set_fill, nf90_nofill, nf90_def_dim, nf90_unlimited,    &
                         nf90_def_var, nf90_inq_attname, nf90_copy_att, nf90_put_att, nf90_enddef, &
                         nf90_global, nf90_put_var, nf90_close
!   
    use Statpack, only : i1b, i4b, stnd, lgl, true, false, zero, one, half, c100,               & 
                         kfix, kexp, kint, tolower, tseries_error13, percent,                   &
                         case_change, string_to_val, val_to_string,                             &
                         allocate_error, merror, arth, power_spctrm2, init_fft, end_fft
!   
    use NCSTAT_Utilities, only : prtunit, nchar, sp, dp, def_missing,      &
                                 get_listarg, get_variable_dims
!   
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_spectvar_1d : '
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
#ifndef _USE_NAGWARE
    integer(i4b)  :: iargc
#endif
    integer(i4b)  :: narg, status, ncid, ncid_filt,                              &
                     varid, filtid, filtid2, coor_id,                            &
                     i, i1, i2, k, l, l1, l2, m, iindex, ndim, nt, var_natts,    &
                     time1, time2, time3, nf, sl, nsmooth,                       &
                     pl, ph, trend, trend2, win2, output_type, output_mode,      &
                     period, nperiod
    integer(i4b), dimension(nf90_max_var_dims) :: dimids, dimlens
    integer(i1b)                               :: kcode
!
    real(stnd), dimension(:),   allocatable    :: vec, psvec, h, spectvar, spectvar2
    real(stnd)                                 :: missing, win, hwin, a, d, tmp, comp, taperp
!
    logical(lgl)                               :: use_tlimited, use_period, use_miss, normvar, overlap
!
    character(len=nchar)                       :: name, name2, history
    character(len=nf90_max_name)               :: fic_name, var_name, filt_name, att_name 
    character(len=14)                          :: fmt
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   GET THE NUMBER OF COMMAND LINE ARGUMENTS.
!
    narg = iargc()
!
!   CHECK THE NUMBER OF COMMAND LINE ARGUMENTS.
! 
    if ( narg==0 ) then
        write (prtunit,*)
        write (prtunit,*) ' Purpose : '
        write (prtunit,*)
        write (prtunit,*) ' Compute variance estimates of a real time series in a selected frequency band.'
        write (prtunit,*) ' The time series is extracted from a uni- or bidimensional variable readed from'
        write (prtunit,*) ' a NetCDF dataset'
        write (prtunit,*)
        write (prtunit,*) ' Usage : '
        write (prtunit,*)
        write (prtunit,*) ' comp_spectvar_1d  '
        write (prtunit,*) '        -f=input_netcdf_file'
        write (prtunit,*) '        -v=netcdf_variable'
        write (prtunit,*) '        -t=time1,time2                   (optional)'
        write (prtunit,*) '        -o=output_netcdf_file            (optional)'
        write (prtunit,*) '        -ni=index_for_2d_netcdf_variable (optional)'
        write (prtunit,*) '        -p=period_length                 (optional)'
        write (prtunit,*) '        -sl=segment_length               (optional : 16,32,64,128,...)'
        write (prtunit,*) '        -pl=minimum_period               (optional)'
        write (prtunit,*) '        -ph=maximum_period               (optional)'
        write (prtunit,*) '        -tr=trend_removal                (optional : 0,1,2,3)'
        write (prtunit,*) '        -tr2=trend_removal_on_segment    (optional : 0,1,2,3)'
        write (prtunit,*) '        -win=window_parameter            (optional : 0.5 > 1.)'
        write (prtunit,*) '        -win2=window_choice              (optional : 1,2,3,4,5,6)'
        write (prtunit,*) '        -tap=tapered_percentage          (optional : 0.0 > 1.)'
        write (prtunit,*) '        -nsm=Daniell_filter_length       (optional)'
        write (prtunit,*) '        -mi=missing_value                (optional)'
        write (prtunit,*) '        -nooverlap                       (optional)'
        write (prtunit,*) '        -normvar                         (optional)'
        write (prtunit,*) '        -double                          (optional)'
#ifdef _USE_NETCDF4
        write (prtunit,*) '        -hdf5                            (optional)'
#endif
        write (prtunit,*) '        -tlimited                        (optional)'
        write (prtunit,*)
        write (prtunit,*) ' By default : '
        write (prtunit,*)
        write (prtunit,*) '-t=the whole time period  associated with the netcdf_variable'
        write (prtunit,*) '-o=spectvar_netcdf_variable.nc'
        write (prtunit,*) '-ni=1            : by default if the netcdf_variable is bidimensional,'
        write (prtunit,*) '                   the first time series is used'
        write (prtunit,*) '-p=time2-time1+1 : by default, the period_length argument is not used'
        write (prtunit,*) '-sl=0            : by default, the time series is not segmented'
        write (prtunit,*) '-pl=0            : by default no filtering is done'
        write (prtunit,*) '-ph=0            : by default no filtering is done'
        write (prtunit,*) '-tr=0            : by default, no trend removal is done'
        write (prtunit,*) '-tr2=0           : by default, no preprocessing is done on the segments'
        write (prtunit,*) '-win=1.          : by default, the rectangular window is convolved with the ideal filter response'
        write (prtunit,*) '-win2=2          : by default, the square window is applied to the time series'
        write (prtunit,*) '-tap=0.2         : by default, if a split-cosine-bell window is used, 20% of the data are tapered'
        write (prtunit,*) '-nsm=0           : by default, the spectral estimates are not smoothed in the frequency domain'
        write (prtunit,*) '-mi=1.e+20       : by default, the missing_value is equal to 1.e+20'
        write (prtunit,*) '-nooverlap       : by default, the segments are overlapped'
        write (prtunit,*) '-normvar         : by default, the raw variance in the selected frequency band is output'
        write (prtunit,*) '-double          : by default, the results are stored as single floating numbers'
#ifdef _USE_NETCDF4
        write (prtunit,*) '-hdf5            : by default, a NetCDF classical format file is created'
#endif
        write (prtunit,*) '-tlimited        : by default, the time dimension is defined as unlimited'
        write (prtunit,*)
!
        stop
!
    else if ( narg<2 .or. narg>21 ) then
        call merror( name_proc//"Wrong number of line arguments" )
    end if
!
    fic_name   = " "
    var_name   = " "
    filt_name  = " "
!
    history = "comp_spectvar_1d"
!
!   DEFAULT VALUES FOR COMMAND LINE ARGUMENTS.
!
    output_mode   = nf90_clobber
    output_type   = nf90_real4
!
    iindex        = 1
    pl            = 0
    ph            = 0
    sl            = 0
    trend         = 0
    trend2        = 0
    win           = one
    win2          = 2
    taperp        = 0.2_stnd
    nsmooth       = 0
!
    missing = def_missing
!
    time1 = 1
    time2 = huge( time2 )
!
    period    = huge( period )
    nperiod   = 1
!
    overlap      = true
    normvar      = false
    use_period   = false
    use_tlimited = false
!
!   GET THE COMMAND LINE ARGUMENTS.
!
    do i = 1, narg
!
        call getarg( i, name )
!
        history = trim(history)//" "//trim(name)
!
        call case_change( name(:3), tolower )
!
        select case ( name(1:3 ) )
!
            case ('-f=')
!
!               GET THE NAME OF THE NETCDF DATASET.
!
                fic_name = name(4:)
!
            case ('-v=')
!
!               GET THE NAME OF THE NETCDF VARIABLE IN THE DATASET.
!
                var_name = name(4:)
!
            case ('-t=')
!
!               GET THE TIME DEFINITION.
!
                call get_listarg( name_proc, name(4:),                    &
                                  "time1 and time2 are integer values",   &
                                  "Wrong period definition",              &
                                  time1, time2 )
!
            case ('-p=')
!
!               GET THE PERIOD LENGTH.
!
                call string_to_val( name(4:), kcode, fmt )
                if ( kcode/=kint ) &
                call merror( name_proc//"The period length is an integer value" )
!
                read( name(4:), fmt ) period
!
                if ( period<=0 ) &
                call merror( name_proc//"The -p= parameter is a positive integer" )
!
                use_period = true
!
            case ('-o=')
!
!               GET THE NAME OF THE NEW NETCDF DATASET FOR STORING THE RESULTS.
!
                filt_name = name(4:)
!
            case default
!
                select case ( name(1:4 ) )
!
                    case ('-ni=')
!
!                       GET THE INDEX FOR A 2D VARIABLE.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) iindex
                        end if
                        if ( kcode/=kint .or. iindex<=0  )                           &
                        call merror( name_proc//"The index for a 2D variable is a positive integer value" )
!
                    case ('-pl=')
!
!                       GET THE MINIMUM PERIOD OF OSCILLATION.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) pl
                        end if
                        if ( kcode/=kint .or. pl<0 .or. pl==1  )                           &
                        call merror( name_proc//"The -pl= argument is an integer equal to 0 or greater than 1" )
!
                    case ('-ph=')
!
!                       GET THE MAXIMUM PERIOD OF OSCILLATION.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) ph
                        end if
                        if ( kcode/=kint .or. ph<0 .or. ph==1  )                           &
                        call merror( name_proc//"The -ph= argument is an integer equal to 0 or greater than 1" )
!
                    case ('-sl=')
!
!                       GET THE LENGTH OF THE SEGMENTS.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) sl
                        end if
!
                        if ( kcode/=kint .or. sl<0  )                           &
                        call merror( name_proc//"The -sl= argument is a positive integer value" )
!
!                       CHECK IF sl IS A POSITIVE EVEN INTEGER.
!
                        if ( sl/=2*(sl/2) )     &
                        call merror( name_proc//'The -sl= argument is an even integer !' )
!
                    case ('-tr=')
!
!                       DETERMINE PRE-FILTERING COMPUTATIONS.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode==kint ) then
                            read( name(5:), fmt ) trend
                        else
                            call merror( name_proc//"The -tr= argument is an integer value" )
                        end if
!
                    case ('-mi=')
!
!                       GET THE MISSING VALUE INDICATOR.
!
                        call string_to_val( name(5:), kcode, fmt )
                        if ( kcode/=kfix .and. kcode/=kexp )                &
                        call merror( name_proc//"The -mi= argument must be a real number " )
                        read( name(5:), fmt ) missing
!
                    case default
!
                        call case_change( name(4:), tolower )
!
!
                        select case ( name(1:5) )
!
                            case ('-tr2=')
!
!                               DETERMINE PRE-FILTERING COMPUTATIONS ON EACH SEGMENT.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) trend2
                                else
                                    call merror( name_proc//"The -tr2= argument is an integer value" )
                                end if
!
                            case ('-win=')
!
!                               DETERMINE THE FORM OF THE WINDOW.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kfix .or. kcode==kexp ) then
                                    read( name(6:), fmt ) win
                                end if
                                if ( (kcode/=kfix .and. kcode/=kexp) .or.     &
                                     win<half                        .or.     &
                                     win>one )                                &
      call merror( name_proc//"The -win= argument must be a real number and verify 0.5<=win<=1. " )
!
                            case ('-nsm=')
!
!                               DETERMINE THE LENGTH OF THE DANIELL FILTER TO BE APPLIED.
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kint ) then
                                    read( name(6:), fmt ) nsmooth
                                end if
!
                                if ( kcode/=kint .or.             &
                                     nsmooth<3   .or.             &
                                     nsmooth==(nsmooth/2)*2   )   &
      call merror( name_proc//"The -nsm= argument must be an odd integer and verify 3<=nsm " )
!
                            case ('-tap=')
!
!                               DETERMINE THE TOTAL PERCENTAGE OF THE DATA TO BE TAPERED IF -WIN2=6 .
!
                                call string_to_val( name(6:), kcode, fmt )
                                if ( kcode==kfix .or. kcode==kexp ) then
                                    read( name(6:), fmt ) taperp
                                end if
                                if ( (kcode/=kfix .and. kcode/=kexp) .or.     &
                                     taperp<=zero                    .or.     &
                                     taperp>one )                             &
      call merror( name_proc//"The -tap= argument must be a real number and verify 0.<tap<=1 " )
!
                            case default
!
                                select case ( name(1:6) )
!
                                    case ('-win2=')
!
!                                       DETERMINE THE FORM OF THE DATA WINDOW.
!
                                        call string_to_val( name(6:), kcode, fmt )
                                        if ( kcode==kint ) then
                                            read( name(6:), fmt ) win2
                                        end if
                                        if ( kcode/=kint .or.      &
                                             win2<1      .or.      &
                                             win2>6 )              &
      call merror( name_proc//"The -win2= argument must be an integer number and verify 1<=win2<=6 " )
!
                                    case default
!
                                        select case ( name(1:) )
!
                                            case ('-nooverlap')
!
!                                               DO NOT OVERLAP THE SEGMENTS.
!
                                                overlap = false
!
                                            case ('-normvar')
!
!                                               COMPUTE PERCENTAGE OF VARIANCE.
!
                                                normvar = true
!
                                            case ('-double')
!
!                                               STORE RESULTS AS DOUBLE PRECISION NUMBERS.
!
                                                output_type = nf90_real8
!
#ifdef _USE_NETCDF4
                                            case ('-hdf5')
!
!                                               REDEFINE THE CREATION MODE FLAG FOR THE OUTPUT FILE.
!
                                                output_mode = nf90_hdf5
#endif
!
                                            case ('-tlimited')
!
!                                               DEFINE THE TIME DIMENSION AS LIMITED.
!
                                                use_tlimited = true
!
                                            case default
!
!                                               WRONG ARGUMENT.
!
                                                call merror( name_proc//"Wrong argument" )
!
                                        end select
!
                                end select
!
                        end select
!
                end select
!
        end select
!
    end do
!
!   CHECK THE COMMAND LINE ARGUMENTS.
!
    if ( fic_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF file is missing" )
!
    if ( var_name==" " )                                                 &
    call merror( name_proc//"Input NETCDF variable is missing" )
!
    if ( filt_name==" " ) filt_name="spectvar_"//trim(var_name)//".nc"
!
!   CHECK PERIODS ORDERING TO SELECT FREQUENCIES.
!
    if ( ph<pl ) then
        i1   = ph
        i2   = pl
        comp = one
    else
        i1    = pl
        i2    = ph
        comp  = zero
    end if
!
    if ( i2==0 )                                          &
    call merror( name_proc//"Setting both -pl=0 and -ph=0  is not allowed" )
!
!   OPEN THE NETCDF DATASET.
!
    i = nf90_sizehint_default
    status = nf90_open( trim(fic_name), nf90_nowrite, ncid, chunksize=i )
    if ( status/=nf90_noerr )                                                &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   GET THE VARIABLE ID FROM ITS NAME.
!   FIND THE DIMENSION LENGTHS AND NAMES FOR THE NETCDF VARIABLE.
!
    call get_variable_dims(  name_proc, trim(var_name), ncid,                       &
                             varid,  dimids_out=dimids, dimlens_out=dimlens,        &
                             natts=var_natts  )
!
    ndim = count(dimlens/=0)
    if ( ndim/=1 .and. ndim/=2 )     &
    call merror( name_proc//"Input NETCDF variable is not uni or bidimensional" )
!
    if ( ndim==2 ) then
        if ( iindex>dimlens(1) )     &
        call merror( name_proc//"Selected index for 2D input NETCDF variable is out of range " )
        nt = dimlens(2)
    else 
        nt = dimlens(1)
    end if
!
!   CHECK THE TIME AND PERIOD DEFINITION.
!
    time1 = max( time1, 1 )
    time2 = min( time2, nt )
    time3 = time2 - time1 + 1
!
    if ( time3<4  ) call merror( name_proc//"Wrong time definition or not enough time observations" )
!
    if ( use_period ) then
!
        if ( period<4  ) call merror( name_proc//"Not enough time observations in the periods" )
!
        nperiod = time3/period
        if ( nperiod*period/=time3 )    &
        call merror( name_proc//"The selected time period is not a multiple of the periodicity" )
!
    else
!
        period = time3
!
    end if
!
!   CHECK SEGMENT LENGTH.
!
    if ( sl==0 ) then
!
        i = 2*(period/2)
        if ( period/=i )       &
    call merror( name_proc//"The period or the number of observations is not an even integer" )
!
        sl = period
!
    else
!
        if ( period<sl )       &
    call merror( name_proc//"Length of segments cannot be greater than the number of observations or the period")
!
    end if
!
    if ( sl<i2 )   call merror( name_proc//tseries_error13 )
!
!   DETERMINE THE NUMBER OF FREQUENCIES.
!
    nf = ( sl/2 ) + 1
!
    if ( nsmooth>nf )    &
    call merror( name_proc//"The length of the Daniell filter must be less or equal to the number of frequencies")
!
!   ALLOCATE WORK ARRAYS FOR THE TIME SERIES AND THE PSD ESTIMATES.
!
!    nf = ( period/2 ) + 1
!
    if ( normvar ) then
!
        allocate( vec(time3), psvec(nf), h(nf), spectvar(nperiod), spectvar2(nperiod), stat=status )
!
    else
!
        allocate( vec(time3), psvec(nf), h(nf), spectvar(nperiod), stat=status )
!
    end if
!
    if ( status/=0 ) call merror( name_proc//allocate_error )
!
!   GET THE DATA.
!
    if ( ndim==1 ) then
        status = nf90_get_var( ncid, varid, vec(:time3),              &   
                               start=(/ time1 /), count=(/ time3 /)   )
    else 
        status = nf90_get_var( ncid, varid, vec(:time3),                          &   
                               start=(/ iindex, time1 /), count=(/ 1, time3 /)    )
    end if
!
    if ( status/=nf90_noerr )                                     &
    call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   FIND THE MISSING VALUE ATTRIBUTE ASSOCIATED WITH THE NETCDF VARIABLE, IF ANY.
!
    status = nf90_get_att( ncid, varid, "missing_value", tmp )
    use_miss = status==nf90_noerr
!
    if ( .not.use_miss ) then
        status = nf90_get_att( ncid, varid, "_FillValue", tmp )
        use_miss = status==nf90_noerr
    end if
!
    if ( use_miss ) then
        if ( any( vec(:time3)==tmp ) )      &
        call merror( name_proc//"Missing values are not allowed" )
    end if
!
!   MULTIPLY THE NETCDF VARIABLE BY THE SCALE FACTOR IF REQUIRED.
!
    status = nf90_get_att( ncid, varid, "scale_factor", tmp )
    if ( status==nf90_noerr ) vec(:time3) = vec(:time3)*tmp
!
!   ADD OFFSET TO THE NETCDF VARIABLE IF REQUIRED.
!
    status = nf90_get_att( ncid, varid, "add_offset", tmp )
    if ( status==nf90_noerr ) vec(:time3) = vec(:time3) + tmp
!
!   BUILD UP IDEAL FILTER GAIN VECTOR.
!
    k = nf - 1
!
    do i=0, k
!
!       SELECT FREQUENCY BAND.
!
        if ( i*i2>=period .and. i*i1<=period ) then
            h(i+1) = one - comp
        else
            h(i+1) = comp
        end if
!
    end do
!
!   CONVOLVE WITH WINDOW TRANSFORM TO GET FILTER FREQUENCY RESPONSE IF NEEDED.
!
    if ( win/=one ) then
!
        hwin = (one-win)*half
        a    = h(2)
!
        do i = 1, k
            d    = h(i)
            h(i) = win*d + hwin*( a+h(i+1) )
            a    = d
        end do
!
        h(k+1) = win*h(k+1) + hwin*(a+a)
!
    end if
!
    if ( normvar ) then
!
!       DETERMINE IF THE MEAN IS INCLUDED WHEN COMPUTING EXPLAINED POWER.
!
        if ( ph<pl ) then
            m = 1
        else
            m = 2
        end if
!
    end if
!
!   INITIALIZE THE FFT SUBROUTINE.
!
    l = nf - 1
    call init_fft( l  )
!
!   FREQUENCY ANALYSIS OF THE REAL TIME SERIES.
!
    l2 = 0
!
    do l=1, nperiod
!
        l1 = l2 + 1
        l2 = l2 + period
!
!       FREQUENCY ANALYSIS OF THE REAL TIME SERIES.
!
        call power_spctrm2( vec(l1:l2), sl, psvec(:nf), overlap=overlap, trend=trend, trend2=trend2,  &
                            win=win2, taperp=taperp, nsmooth=nsmooth, normpsd=false, initfft=false    )
!
        spectvar(l) = dot_product( h(:nf), psvec(:nf) )
!
        if ( normvar ) then
!
            spectvar2(l) = c100*(spectvar(l)/sum(psvec(m:nf)))
!
        end if
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    call end_fft( )
!
!   CREATE THE NEW NETCDF DATASET FOR THE FREQUENCY ANALYSIS.
!
    status = nf90_create( filt_name, cmode=output_mode, ncid=ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_set_fill( ncid_filt, nf90_nofill, i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   DEFINE THE DIMENSIONS FOR THE NEW NETCDF DATASET.
!
    if ( use_tlimited ) then
        dimlens(1)   = nperiod
    else
        dimlens(1)   = nf90_unlimited
    end if
    name       = "time"
!
    status = nf90_def_dim( ncid_filt, name=trim(name), len=dimlens(1), dimid=dimids(1) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim(nf90_strerror(status)) )
!
!   DEFINE THE VARIABLES FOR THE NEW NETCDF DATASET.
!
    status = nf90_def_var( ncid_filt, trim(name),  nf90_int,                &
                           dimids=dimids(1), varid=coor_id )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_def_var( ncid_filt, trim(var_name)//"_spectvar",  output_type,            &
                           dimids=dimids(1), varid=filtid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( normvar ) then
!
        status = nf90_def_var( ncid_filt, trim(var_name)//"_spectvar_percent",  output_type, &
                               dimids=dimids(1), varid=filtid2 )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   ASSIGN ATTRIBUTES FOR THE NEW NETCDF DATASET.
!
    status = nf90_put_att( ncid_filt, coor_id, "long_name", "Time axis" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, coor_id, "short_name", "Time" )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   COPY ATTRIBUTES FROM THE  VARIABLE .
!
    do i = 1, var_natts
!
        status = nf90_inq_attname( ncid, varid, i, att_name )
        if ( status/=nf90_noerr )  call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( att_name=="add_offset"         .or.  &
             att_name=="scale_factor"       .or.  &
             att_name=="units"              .or.  &
             att_name=="missing_value"      .or.  &
             att_name=="_FillValue"               )   cycle
!
!       COPY USEFUL ATTRIBUTES.
!
        status = nf90_copy_att(  ncid, varid, att_name, ncid_filt, filtid )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        if ( normvar ) then
            status = nf90_copy_att(  ncid, varid, att_name, ncid_filt, filtid2 )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
        end if
!
    end do
!
    if (  output_type==nf90_real8 ) then
!
        status = nf90_put_att( ncid_filt, filtid, "missing_value", real( missing, dp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, filtid, "_FillValue", real( missing, dp ) )
!
    else
!
        status = nf90_put_att( ncid_filt, filtid, "missing_value", real( missing, sp ) )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, filtid, "_FillValue", real( missing, sp ) )
!
    end if
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    name = " "
    status = nf90_get_att( ncid, varid, "units", name )
    if ( status==nf90_noerr ) then
        status = nf90_put_att( ncid_filt, filtid, "units", "("//trim(name)//")**2 "  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
    end if
!
    if ( normvar ) then
!
        if (  output_type==nf90_real8 ) then
!
            status = nf90_put_att( ncid_filt, filtid2, "missing_value", real( missing, dp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_filt, filtid2, "_FillValue", real( missing, dp ) )
!
        else
!
            status = nf90_put_att( ncid_filt, filtid2, "missing_value", real( missing, sp ) )
            if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
            status = nf90_put_att( ncid_filt, filtid2, "_FillValue", real( missing, sp ) )
!
        end if
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
        status = nf90_put_att( ncid_filt, filtid2 , "units", percent  )
        if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
!   SET GLOBAL ATTRIBUTES.
!
    if ( ndim==1 ) then
        name = trim(var_name)
    else
        call val_to_string( iindex, fmt, i )
        name = trim(var_name)//"("//fmt(1:i)//")"
    end if
!
    name2 = "Filtered variance of "//trim(name)
!
    status = nf90_put_att( ncid_filt, nf90_global, "title", trim(name2) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "minimum_period", pl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "maximum_period", ph )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "period_length", period )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "segment_length", sl )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "trend_removal", trend )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "trend_removal_segment", trend2 )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "window_parameter", win )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "window_choice", win2 )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( win2==6 ) then
!
        status = nf90_put_att( ncid_filt, nf90_global, "tapered_percentage", taperp )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    end if
!
    if ( nsmooth/=0 ) then
!
        status = nf90_put_att( ncid_filt, nf90_global, "Daniell_filter_length", nsmooth )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
        i = 1
!
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_filt, nf90_global, "smoothing", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( overlap ) then
        i = 1
    else
        i = 0
    end if
!
    status = nf90_put_att( ncid_filt, nf90_global, "overlapping", i )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_put_att( ncid_filt, nf90_global, "history", trim(history) )
    if ( status/=nf90_noerr ) call merror( name_proc//trim( nf90_strerror( status ) ) )
!
!   LEAVE DEFINE MODE FOR THE NEW NETCDF DATASET.
!
    status = nf90_enddef( ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE VARIABLES FOR THE NEW NETCDF DATASET.
!
    status = nf90_put_var( ncid_filt, coor_id, arth(1_i4b,1_i4b,nperiod) )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!   WRITE FILTERED VARIANCE ESTIMATES.
!
    status = nf90_put_var( ncid_filt, filtid, spectvar )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    if ( normvar ) then
!
        status = nf90_put_var( ncid_filt, filtid2, spectvar2 )
        if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( spectvar2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( vec, psvec, h, spectvar )
!
!   CLOSE THE NETCDF DATASETS.
!
    status = nf90_close( ncid )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
    status = nf90_close( ncid_filt )
    if ( status/=nf90_noerr ) call merror(  name_proc//trim( nf90_strerror( status ) ) )
!
! END OF PROGRAM comp_spectvar_1d
! ===============================
!
end program comp_spectvar_1d
